#include "chess.h"
#include "data.h"
/* last modified 04/23/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   Evaluate() is used to evaluate the chess board.  broadly, it addresses    *
 *   four (4) distinct areas:  (1) material score which is simply a summing of *
 *   piece types multiplied by piece values;  (2) pawn scoring which considers *
 *   placement of pawns and also evaluates passed pawns, particularly in end-  *
 *   game situations;  (3) piece scoring which evaluates the placement of each *
 *   piece as well as things like piece mobility;  (4) king safety which       *
 *   considers the pawn shelter around the king along with material present to *
 *   facilitate an attack.                                                     *
 *                                                                             *
 *******************************************************************************
 */
int Evaluate(TREE * RESTRICT tree, int ply, int wtm, int alpha, int beta)
{
  register PAWN_HASH_ENTRY *ptable;
  PXOR *pxtable;
  register int score = 0, can_win = CAN_WIN;
  register int side, majors, minors;
  register int phase, lscore;
  register int cutoff = queen_value;
  UINT64 *etable;
  UINT64 temp_hashkey;

/*
 **********************************************************************
 *                                                                    *
 *   initialize.                                                      *
 *                                                                    *
 **********************************************************************
 */
  maximum_ply = Max(maximum_ply, ply);
  tree->evaluations++;
  tree->score_mg = 0;
  tree->score_eg = 0;
/*
 *************************************************************
 *                                                           *
 *  The "dangerous" flag simply indicates whether that side  *
 *  has enough material to whip up a mating attack if the    *
 *  other side is careless (Q + minor or better, or RR + two *
 *  minors or better).                                       *
 *                                                           *
 *************************************************************
 */
    tree->dangerous[white] = (Queens(white) && TotalPieces(white, occupied) >= 1 * (p_vals[queen] + p_vals[knight])) ||
      (TotalPieces(white, rook) > 1 && TotalPieces(white, occupied) >= 2 * (p_vals[rook] + p_vals[knight]));
    tree->dangerous[black] = (Queens(black) && TotalPieces(black, occupied) >= 1 * (p_vals[queen] + p_vals[knight])) ||
      (TotalPieces(black, rook) > 1 && TotalPieces(black, occupied) >= 2 * (p_vals[rook] + p_vals[knight]));
/*
 **********************************************************************
 *                                                                    *
 *  Material, a term determined by the sum of piece values of each    *
 *  side.  The material value is the most influential part of the     *
 *  evaluation.  Usually it is the sum of a constant value for each   *
 *  piece present, measured in units of a fraction of a pawn, for     *
 *  instance the common centipawn scale, which allows positional      *
 *  features of the position, worth less than a single pawn, to be    *
 *  evaluated without requiring fractions but a fixed point score.    *
 *                                                                    *
 *  In chess, it is known that certain material features provide an   *
 *  advantage, such as the bishop pair (which might be worth as much  *
 *  as half a pawn).  A program might increase the values of rooks    *
 *  when there are less pawns on the board or prefer knights when     *
 *  there are many pawns.  Larry Kaufman performed statistical tests  *
 *  on a variety of different material configurations to approximate  *
 *  their values.                                                     *
 *                                                                    *
 *  in Vafra we use Tracy Riegle's ideas on material imbalances.      *
 *                                                                    *
 **********************************************************************
 */
  EvaluateMaterial(tree);
/*
 **********************************************************************
 *                                                                    *
 *   side on move bonus.                                              *
 *                                                                    *
 **********************************************************************
 */
  tree->score_mg += ((wtm) ? wtm_bonus[mg] : -wtm_bonus[mg]);
  tree->score_eg += ((wtm) ? wtm_bonus[eg] : -wtm_bonus[eg]);
/*
 **********************************************************************
 *                                                                    *
 *   check for draws due to insufficient material and adjust the      *
 *   score as necessary.  this code also handles a special endgame    *
 *   case where one side has only a lone king, and the king has no    *
 *   legal moves.  this has been shown to break a few evaluation      *
 *   terms such as bishop + wrong color rook pawn.  if this case is   *
 *   detected, a drawscore is returned.                               *
 *                                                                    *
 **********************************************************************
 */
  if (TotalPieces(white, occupied) <= eg_mat && TotalPieces(black, occupied) <= eg_mat)
    do {
/*
 ************************************************************
 *                                                          *
 *   if neither side has any pieces, and both sides have    *
 *   non-rookpawns, then either side can win.               *
 *                                                          *
 ************************************************************
 */
      if (TotalPieces(white, occupied) == 0 && TotalPieces(black, occupied) == 0
          && (Pawns(white) & not_rook_pawns && Pawns(black) & not_rook_pawns))
        break;
/*
 ************************************************************
 *                                                          *
 *   if one side is an exchange up, but has no pawns, then  *
 *   that side can not possibly win.                        *
 *                                                          *
 ************************************************************
 */
      majors =
          TotalPieces(white, rook) + 2 * TotalPieces(white,
          queen) - TotalPieces(black, rook) - 2 * TotalPieces(black, queen);
      if (Abs(majors) == 1) {
        minors =
            TotalPieces(white, knight) + TotalPieces(white,
            bishop) - TotalPieces(black, knight) - TotalPieces(black, bishop);
        if (majors == -minors) {
          if (TotalPieces(black, pawn) == 0)
            SetCannotWin(black);
          if (TotalPieces(white, pawn) == 0)
            SetCannotWin(white);
        }
        if (can_win == CANNOT_WIN)
          break;
      }
/*
 ************************************************************
 *                                                          *
 *   now check several special cases, such as bishop + the  *
 *   wrong rook pawn and adjust can_win accordingly.        *
 *                                                          *
 ************************************************************
 */
      if (!EvaluateWinningChances(tree, white, wtm))
        SetCannotWin(white);
      if (!EvaluateWinningChances(tree, black, wtm))
        SetCannotWin(black);
    } while (0);
/*
 **********************************************************************
 *                                                                    *
 *   determine if this position should be evaluated to force mate     *
 *   (neither side has pawns) or if it should be evaluated normally.  *
 *                                                                    *
 *   note the special case of no pawns, one side is ahead in total    *
 *   material, but the game is a hopeless draw.  KRN vs KR is one     *
 *   example.  If EvaluateWinningChances() determines that the side   *
 *   with extra material can not win, the score is pulled closer to a *
 *   draw although it can not collapse completely to the drawscore as *
 *   it is possible to lose KRB vs KR if the KR side lets the king    *
 *   get trapped on the edge of the board.                            *
 *                                                                    *
 **********************************************************************
 */
  tree->all_pawns = Pawns(black) | Pawns(white);
  if (!tree->all_pawns) {
    if (TotalPieces(white, occupied) > p_vals[knight] || TotalPieces(black, occupied) > p_vals[knight]) {
      if (Material > 0)
        EvaluateMate(tree, white);
      else if (Material < 0)
        EvaluateMate(tree, black);
      if (tree->score_eg > DrawScore(white) && !CanWin(white))
        tree->score_eg = tree->score_eg / 4;
      if (tree->score_eg < DrawScore(black) && !CanWin(black))
        tree->score_eg = tree->score_eg / 4;
      score = tree->score_eg;
      return ((wtm) ? score : -score); /* from side to move perspective */
    }
  }
/*
 **********************************************************************
 *                                                                    *
 *   now evaluate pawns.  if the pawn hash signature has not changed  *
 *   from the last entry to Evaluate() then we already have every-    *
 *   thing we need in the pawn hash entry.  in this case, we do not   *
 *   need to call EvaluatePawns() at all.  EvaluatePawns() does all   *
 *   of the analysis for information specifically regarding only      *
 *   pawns.  in many cases, it merely records the presence/absence of *
 *   positional pawn feature because that feature also depends on     *
 *   pieces.  note that anything put into EvaluatePawns() can only    *
 *   consider the placement of pawns.  kings or other pieces can not  *
 *   influence the score because those pieces are not hashed into the *
 *   pawn hash signature.  violating this principle leads to lots of  *
 *   very difficult and challenging debugging problems.               *
 *                                                                    *
 **********************************************************************
 */
  else {
    if (PawnHashKey == tree->pawn_score.key) {
      tree->score_mg += tree->pawn_score.score_mg;
      tree->score_eg += tree->pawn_score.score_eg;
    }
/*
 ************************************************************
 *                                                          *
 *   first check to see if this position has been handled   *
 *   before.  if so, we can skip the work saved in the pawn *
 *   hash table.                                            *
 *                                                          *
 ************************************************************
 */
    else {
      ptable = pawn_hash_table + (PawnHashKey & pawn_hash_mask);
      pxtable = (PXOR *) &(tree->pawn_score);
      tree->pawn_score = *ptable;
      tree->pawn_score.key ^= pxtable->entry[1] ^ pxtable->entry[2] ^ pxtable->entry[3];
      if (tree->pawn_score.key != PawnHashKey) {
        tree->pawn_score.key = PawnHashKey;
        tree->pawn_score.open_files = 255;
        tree->pawn_score.score_mg = 0;
        tree->pawn_score.score_eg = 0;
        for (side = black; side <= white; side++)
          EvaluatePawns(tree, side);
        ptable->key =
            pxtable->entry[0] ^ pxtable->
            entry[1] ^ pxtable->entry[2] ^ pxtable->entry[3];
        memcpy((char *) ptable + 8, (char *) &(tree->pawn_score) + 8, 24);
      }
      tree->score_mg += tree->pawn_score.score_mg;
      tree->score_eg += tree->pawn_score.score_eg;
    }
/*
 **********************************************************************
 *                                                                    *
 *   if there are any passed pawns, first call EvaluatePassedPawns()  *
 *   to evaluate them.  then, if one side has a passed pawn and the   *
 *   other side has no pieces, call EvaluatePassedPawnRaces() to see  *
 *   if the passed pawn can be stopped from promoting.                *
 *                                                                    *
 **********************************************************************
 */
    if (tree->pawn_score.passed[black] || tree->pawn_score.passed[white]) {
      for (side = black; side <= white; side++)
        if (tree->pawn_score.passed[side])
          EvaluatePassedPawns(tree, side);
      if ((TotalPieces(white, occupied) == 0 && tree->pawn_score.passed[black])
          || (TotalPieces(black, occupied) == 0 &&
              tree->pawn_score.passed[white]))
        EvaluatePassedPawnRaces(tree, wtm);
    }
  }
/*
 **********************************************************************
 *                                                                    *
 *   call EvaluateDevelopment() to evaluate development.              *
 *                                                                    *
 **********************************************************************
 */
  for (side = black; side <= white; side++) {
    if (TotalPieces(side, occupied) > eg_mat)
      EvaluateDevelopment(tree, ply, side);
  }
/*
 **********************************************************************
 *                                                                    *
 *   now evaluate pieces.                                             *
 *                                                                    *
 **********************************************************************
 */
  phase = Max(2 * eg_mat, TotalPieces(white, occupied) + TotalPieces(black, occupied));
  lscore = (wtm) ? (tree->score_eg + (tree->score_mg - tree->score_eg) * (phase - 2 * eg_mat) / Kp) :
    -(tree->score_eg + (tree->score_mg - tree->score_eg) * (phase - 2 * eg_mat) / Kp);
/*
 *************************************************************
 *                                                           *
 *  Lazy evaluation test.  We have computed the large        *
 *  positional scores (except for king safety).  If the      *
 *  score is too far outside the alpha/beta window, we skip  *
 *  the piece scoring which is the most expensive of all the *
 *  evaluation terms, and simply use what we have at this    *
 *  point.                                                   *
 *                                                           *
 *************************************************************
 */
  if (lazy_eval_cutoff[ub] && lazy_eval_cutoff[lb])
    cutoff = (tree->dangerous[white] || tree->dangerous[black]) ? lazy_eval_cutoff[ub] : lazy_eval_cutoff[lb];
  tree->tropism[white] = 0;
  tree->tropism[black] = 0;
/*
 **********************************************************************
 *                                                                    *
 *   In the chess programming sense of the phrase, lazy evaluation    *
 *   means dividing all the tasks performed by evaluation function in *
 *   (usually two) stages.  If after performing all the tasks for a   *
 *   given stage score exceeds beta by a certain margin or if it      *
 *   falls below alpha by the same margin, the score is returned and  *
 *   no more evaluation is conducted.  The first stage usually        *
 *   consists of the evaluation terms that are relatively cheap,      *
 *   possibly incrementally updated, like material and piece-square   *
 *   tables.  Lazy evaluation may be viewed as taking the alpha-beta  *
 *   principle one step further.  If some parts of the evaluation     *
 *   function are omitted due to this heuristics, it means that one   *
 *   side is already winning "the evaluation game" by the margin      *
 *   unacceptable to the opponent, and therefore a cutoff occurs.     *
 *                                                                    *
 **********************************************************************
 */
  if (!DoLazy(lscore, cutoff)) {
    for (side = black; side <= white; side++) {
      EvaluateKnights(tree, side);
      EvaluateBishops(tree, side);
      EvaluateRooks(tree, side);
      EvaluateQueens(tree, side);
    }
  }
/*
 *************************************************************
 *                                                           *
 *  Note:  We MUST evaluate kings last, since their scoring  *
 *  depends on the tropism scores computed by the other      *
 *  piece evaluators.  Do NOT try to collapse the following  *
 *  loops into one loop.  That will break things since it    *
 *  would violate the kings last rule.                       *
 *                                                           *
 *************************************************************
 */
  if (!DoLazy(lscore, cutoff)) {
    for (side = black; side <= white; side++) {
      EvaluateKings(tree, ply, side);
    }
  }
/*
 **********************************************************************
 *                                                                    *
 *   now adjust the score if the game is drawish but one side appears *
 *   to be significantly better according to the computed score.      *
 *                                                                    *
 **********************************************************************
 */
  score = tree->score_eg + (tree->score_mg - tree->score_eg) * (phase - 2 * eg_mat) / Kp;
  score = EvaluateDraws(tree, wtm, ply, can_win, score);
/*#if defined(SKILL)
  score =
      skill * score / 100 + ((100 -
          skill) * PAWN_VALUE * (UINT64) Random32() / 0x100000000ULL) / 100;
#endif*/
  return ((wtm) ? score : -score); /* from side to move perspective */
}

/* last modified 07/06/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateBishops() is used to evaluate black/white bishops.                *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateBishops(TREE * RESTRICT tree, int side)
{
  register BITBOARD temp, moves, moves2;
  register int square, s, trop;
  register int i, tpawns;
  register int t, score_eg = 0, score_mg = 0, enemy = Flip(side);
  register int pair = 0;
  int ray;

  if (TotalPieces(side, bishop) > 1)
    pair = 1;
/*
 ************************************************************
 *                                                          *
 *   first, locate each bishop and add in its piece/square  *
 *   score.                                                 *
 *                                                          *
 ************************************************************
 */
  temp = Bishops(side);
  while (temp) {
    square = LSB(temp);
    score_mg += bval[mg][side][square];
    score_eg += bval[eg][side][square];
/*
 ************************************************************
 *                                                          *
 *   now, evaluate for "outposts" which is a bishop that    *
 *   can't be driven off by an enemy pawn, and which is     *
 *   supported by a friendly pawn.                          *
 *                                                          *
 *   if the enemy has NO minor to take this bishop, then    *
 *   increase the bonus.                                    *
 *                                                          *
 ************************************************************
 */
    t = bishop_outpost[side][square];
    if (t > 0) {
      if (!(mask_no_pattacks[enemy][square] & Pawns(enemy))) {
        score_eg += t;
        score_mg += t;
        if (t && pawn_attacks[enemy][square] & Pawns(side)) {
          score_eg += t / 2;
          score_mg += t / 2;
          if (!Knights(enemy) && !(Color(square) & Bishops(enemy))) {
            score_eg += t;
            score_mg += t;
          }
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *  Next we count the number of friendly pawns on the same  *
 *  color squares as the bishop.  This is a bad thing since *
 *  it restricts the bishop's ability to move.  We only do  *
 *  this if there is only one bishop for this side.         *
 *                                                          *
 ************************************************************
 */
    if (TotalPieces(side, bishop) == 1) {
      if (dark_squares & SetMask(square))
        tpawns = PopCnt(dark_squares & Pawns(side));
      else
        tpawns = PopCnt(~dark_squares & Pawns(side));
      score_mg -= tpawns * bishop_pawns_on_color[mg];
      score_eg -= tpawns * bishop_pawns_on_color[eg];
    }
/*
 ************************************************************
 *                                                          *
 *   check to see if the bishop is trapped at a7 or h7 with *
 *   a pawn at b6 or g6 that can advance one square and     *
 *   trap the bishop, or a pawn at b6 or g6 that has        *
 *   trapped the bishop already.  also test for the bishop  *
 *   at b8 or g8 as that might not be an escape.            *
 *                                                          *
 ************************************************************
 */
    if (square == sqflip[side][A7] && SetMask(sqflip[side][B6]) & Pawns(enemy)) {
      score_eg -= bishop_trapped;
      score_mg -= bishop_trapped;
    } else if (square == sqflip[side][B8] &&
        SetMask(sqflip[side][C7]) & Pawns(enemy)) {
      score_eg -= bishop_trapped;
      score_mg -= bishop_trapped;
    } else if (square == sqflip[side][H7] &&
        SetMask(sqflip[side][G6]) & Pawns(enemy)) {
      score_eg -= bishop_trapped;
      score_mg -= bishop_trapped;
    } else if (square == sqflip[side][G8] &&
        SetMask(sqflip[side][F7]) & Pawns(enemy)) {
      score_eg -= bishop_trapped;
      score_mg -= bishop_trapped;
    }
/*
 ************************************************************
 *                                                          *
 *   mobility counts the number of squares the piece        *
 *   attacks, excluding squares with friendly pieces,       *
 *   and weighs each square according to centralization.    *
 *                                                          *
 ************************************************************
 */
    moves = AttacksBishop(square, OccupiedSquares) & Occupied(side);
    moves2 = AttacksBishop(square, OccupiedSquares) & Occupied(enemy);
    if (tree->cache_b_friendly[square] != moves ||
        tree->cache_b_enemy[square] != moves2) {
      tree->cache_b_friendly[square] = moves;
      tree->cache_b_enemy[square] = moves2;
      moves = AttacksBishop(square, OccupiedSquares) & ~Occupied(side);
      t = -lower_b;
      for (i = 0; i < 4; i++)
        t += PopCnt(moves & mobility_mask_b[i]) * mobility_score_b[pair][i];
      tree->cache_b_mobility[square] = t;
    }
    t = tree->cache_b_mobility[square];
    if (t > 0 && AbsolutePin(tree, side, square, &ray) != 0) {
      if (ray == 1 || ray == 8)
        t = 0;
      else if (ray == 7 || ray == 9)
        t /= 2;
    }
    if (t > 0 && !IsActive(bishop, moves))
      t /= 2;
    score_mg += t;
    score_eg += t;
/*
 ************************************************************
 *                                                          *
 *   check for pawns on both wings, which makes a bishop    *
 *   even more valuable against an enemy knight             *
 *                                                          *
 ************************************************************
 */
    if (tree->all_pawns & mask_fgh && tree->all_pawns & mask_abc) {
      score_mg += bishop_with_wing_pawns[mg];
      score_eg += bishop_with_wing_pawns[eg];
    }
/*
 ************************************************************
 *                                                          *
 *   adjust the tropism count for this piece.               *
 *                                                          *
 ************************************************************
 */
    if (tree->dangerous[side]) {
      trop =
          ((AttacksBishop(square, OccupiedSquares & ~(Queens(side)))) &
          king_attacks[KingSQ(enemy)]) ? 1 : Distance(square, KingSQ(enemy));
      tree->tropism[side] += king_tropism_b[trop];
    }
    temp &= temp - 1;
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 05/23/13 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateDevelopment() is used to encourage the program to develop its     *
 *   pieces before moving its queen.  standard developmental principles are    *
 *   applied.  they include:  (1) don't move the queen until minor pieces are  *
 *   developed;  (2) advance the center pawns as soon as possible;  (3) don't  *
 *   move the king unless its a castling move.                                 *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateDevelopment(TREE * RESTRICT tree, int ply, int side)
{
  register int score_mg = 0;
  register int enemy = Flip(side);

/*
 ************************************************************
 *                                                          *
 *   first, some "thematic" things, which includes don't    *
 *   block the c-pawn in queen-pawn openings.               *
 *                                                          *
 ************************************************************
 */
  if (!(SetMask(sqflip[side][E4]) & Pawns(side)) &&
      SetMask(sqflip[side][D4]) & Pawns(side)) {
    if (SetMask(sqflip[side][C2]) & Pawns(side) &&
        SetMask(sqflip[side][C3]) & (Knights(side) | Bishops(side)))
      score_mg -= development_thematic;
  }
/*
 ************************************************************
 *                                                          *
 *   If the king hasn't moved at the beginning of the       *
 *   search, but it has moved somewhere in the current      *
 *   search path, make *sure* it's a castle move or else    *
 *   penalize the loss of castling privilege.               *
 *                                                          *
 ************************************************************
 */
  if (Castle(1, side) > 0) {
    int oq = (Queens(enemy)) ? 3 : 1;

    if (Castle(ply, side) != Castle(1, side)) {
      if (Castle(ply, side) == 0)
        score_mg -= oq * development_losing_castle;
      else if (Castle(ply, side) > 0)
        score_mg -= (oq * development_losing_castle) / 2;
    } else
      score_mg -= oq * development_not_castled;
  }
/*
 ************************************************************
 *                                                          *
 *   check for an undeveloped knight/rook combo             *
 *                                                          *
 ************************************************************
 */
  if (PcOnSq(sqflip[side][B1]) == pieces[side][knight] &&
      PcOnSq(sqflip[side][A1]) == pieces[side][rook])
    score_mg -= undeveloped_piece;
  if (PcOnSq(sqflip[side][G1]) == pieces[side][knight] &&
      PcOnSq(sqflip[side][H1]) == pieces[side][rook])
    score_mg -= undeveloped_piece;
/*
 ************************************************************
 *                                                          *
 *   check for early queen development                      *
 *                                                          *
 ************************************************************
 */
  if (PcOnSq(sqflip[side][B1]) == pieces[side][knight] ||
    PcOnSq(sqflip[side][C1]) == pieces[side][bishop] ||
    PcOnSq(sqflip[side][G1]) == pieces[side][knight] ||
    PcOnSq(sqflip[side][F1]) == pieces[side][bishop]) {
    int u = 0;
    if (PcOnSq(sqflip[side][B1]) == pieces[side][knight])
      u++;
    if (PcOnSq(sqflip[side][C1]) == pieces[side][bishop])
      u++;
    if (PcOnSq(sqflip[side][G1]) == pieces[side][knight])
      u++;
    if (PcOnSq(sqflip[side][F1]) == pieces[side][bishop])
      u++;
    if (PcOnSq(sqflip[side][D1]) != pieces[side][queen])
      score_mg -= development_early_queen * u / 4;
  }
  tree->score_mg += sign[side] * score_mg;
/*
 ************************************************************
 *                                                          *
 *   evaluate space for minor pieces                        *
 *                                                          *
 ************************************************************
 */
  if (development_space) {
    if (TotalPieces(white, occupied) > eg_mat || TotalPieces(black, occupied) > eg_mat) {
      if ((Bishops(side) | Knights(side))) {
          EvaluateSpace(tree, ply, side);
      }
    }
  }
}

/* last modified 05/23/13 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateSpace() is used to evaluate space for a given side.  the space    *
 *   evaluation is a simple bonus based on the number of safe squares          *
 *   available for minor pieces on the central four files between ranks 2 and  *
 *   4.                                                                        *
 *                                                                             *
 *   in the middlegame, each side is given a bonus based on how many squares   *
 *   inside the area of the board which is considered by the space evaluation  *
 *   are safe and available for friendly minor pieces.                         *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateSpace(TREE * RESTRICT tree, int ply, int side)
{
  register BITBOARD temp, pawns;
  register int square, sq;
  register int t = 0, score_mg = 0, enemy = Flip(side);

  temp = mask_space[side] & ~OccupiedSquares;
  while (temp) {
    square = LSB(temp);
    if (!((pawn_attacks[side][square] & Pawns(enemy)) ||
      (((AttacksBishop(square, OccupiedSquares) & Occupied(enemy)) | (AttacksKnight(square) & Occupied(enemy))) &&
      (!((AttacksBishop(square, OccupiedSquares) & Occupied(side)) | (AttacksKnight(square) & Occupied(side)) |
      (pawn_attacks[enemy][square] & Pawns(side))))))) {
      t++;
/*
 ************************************************************
 *                                                          *
 *   safe squares one, two or three squares behind          *
 *   friendly pawns are counted twice.                      *
 *                                                          *
 ************************************************************
 */
      pawns = file_mask[File(square)] & Pawns(side);
      while (pawns) {
        if ((sq = LSB(pawns)) && FileDistance(square, sq) <= 3 &&
          (Behind(side, Rank(square), Rank(sq)))) {
          t++;
        }
        pawns &= pawns - 1;
      }
    }
    temp &= temp - 1;
  }
  score_mg = t * development_space;
  tree->score_mg += sign[side] * score_mg;
}

/* last modified 04/25/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateDraws() is used to adjust the score based on whether the side     *
 *   that appears to be better according the computed score can actually win   *
 *   the game or not.  if the answer is "no" then the score is reduced         *
 *   significantly to reflect the lack of winning chances.                     *
 *                                                                             *
 *******************************************************************************
 */
int EvaluateDraws(TREE * RESTRICT tree, int wtm, int ply, int can_win, int score)
{
/*
 ************************************************************
 *                                                          *
 *   if the ending has only bishops of opposite colors, the *
 *   score is pulled closer to a draw.  if the score says   *
 *   one side is winning, but that side doesn't have enough *
 *   material to win, the score is set to DRAW.             *
 *                                                          *
 *   if this is a pure BOC ending, it is very drawish un-   *
 *   less one side has at least 4 pawns.  more pawns makes  *
 *   it harder for a bishop and king to stop them all from  *
 *   advancing.                                             *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(white, occupied) <= eg_mat && TotalPieces(black, occupied) <= eg_mat) {
    if (TotalPieces(white, bishop) == 1 && TotalPieces(black, bishop) == 1) {
      if (square_color[LSB(Bishops(black))] !=
          square_color[LSB(Bishops(white))]) {
        if (TotalPieces(white, occupied) == p_vals[bishop] &&
            TotalPieces(black, occupied) == p_vals[bishop] && ((TotalPieces(white, pawn) < 4
                    && TotalPieces(black, pawn) < 4) ||
                Abs(TotalPieces(white, pawn) - TotalPieces(black, pawn)) < 2))
          score = score / 2;
        else if (TotalPieces(white, occupied) == TotalPieces(black, occupied))
          score = 3 * score / 4;
      }
    }
  }
  if (can_win != CAN_WIN) {
    if (!CanWin(white) && score > DrawScore(white))
      score = DrawScore(white);
    if (!CanWin(black) && score < DrawScore(black))
      score = DrawScore(black);
  }
/*
 ************************************************************
 *                                                          *
 *   if we are running into the 50-move rule, then start    *
 *   dragging the score toward draw.  this is the idea of a *
 *   "weariness factor" as mentioned by Dave Slate many     *
 *   times.  this avoids slamming into a draw at move 50    *
 *   and having to move something quickly, rather than      *
 *   slowly discovering that the score is dropping and that *
 *   pushing a pawn or capturing something will cause it to *
 *   go back to its correct value a bit more smoothly.      *
 *                                                          *
 ************************************************************
 */
  if (Rule50Moves(ply) > 80) {
    int scale = Min(101 - Rule50Moves(ply), 0);
    score = DrawScore(wtm) + score * scale / 20;
  }
  return (score);
}

/* last modified 09/27/08 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateHasOpposition() is used to determine if one king stands in        *
 *   "opposition" to the other.  if the kings are opposed on the same file or  *
 *   else are opposed on the same diagonal, then the side not-to-move has the  *
 *   opposition and the side-to-move must give way.                            *
 *                                                                             *
 *******************************************************************************
 */
int EvaluateHasOpposition(int on_move, int king, int enemy_king) {
  int file_distance, rank_distance;

  file_distance = FileDistance(king, enemy_king);
  rank_distance = RankDistance(king, enemy_king);
  if (rank_distance < 2)
    return (1);
  if (on_move) {
    if (rank_distance & 1)
      rank_distance--;
    if (file_distance & 1)
      file_distance--;
  }
  if (!(file_distance & 1) && !(rank_distance & 1))
    return (1);
  return (0);
}

/* last modified 10/11/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateKings() is used to evaluate black/white kings.                    *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateKings(TREE * RESTRICT tree, int ply, int side)
{
  register int score_eg = 0, score_mg = 0, defects;
  int enemy = Flip(side);

/*
 ************************************************************
 *                                                          *
 *   first, check for where the king should be if this is   *
 *   an endgame.  ie with pawns on one wing, the king needs *
 *   to be on that wing.  with pawns on both wings, the     *
 *   king belongs in the center.                            *
 *                                                          *
 ************************************************************
 */
  if (tree->all_pawns) {
    if (tree->all_pawns & mask_efgh && tree->all_pawns & mask_abcd)
      score_eg += kval_n[side][KingSQ(side)];
    else if (tree->all_pawns & mask_efgh)
      score_eg += kval_k[side][KingSQ(side)];
    else
      score_eg += kval_q[side][KingSQ(side)];
  }
/*
 ************************************************************
 *                                                          *
 *   now, check for the "trojan horse" attack where the     *
 *   opponent offers a piece to open the h-file with a very *
 *   difficult to refute attack.                            *
 *                                                          *
 ************************************************************
 */
  if (tree->dangerous[enemy]) {
    if (trojan_check) {
      if (root_wtm == side && File(KingSQ(side)) >= FILEE) {
        if (!(tree->all_pawns & file_mask[FILEH])) {
          if (Rooks(enemy) && Queens(enemy))
            score_mg -= king_safety_mate_threat;
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   Now do castle scoring, if the king has castled, the    *
 *   pawns in front are important.  If not castled yet, the *
 *   pawns on the kingside should be preserved for this.    *
 *                                                          *
 ************************************************************
 */
    defects = 0;
    if (Castle(ply, side) <= 0) {
      if (File(KingSQ(side)) >= FILEE) {
        if (File(KingSQ(side)) > FILEE)
          defects = tree->pawn_score.defects_k[side];
        else
          defects = tree->pawn_score.defects_e[side];
      } else {
        if (File(KingSQ(side)) < FILED)
          defects = tree->pawn_score.defects_q[side];
        else
          defects = tree->pawn_score.defects_d[side];
      }
    } else {
      if (Castle(ply, side) == 3)
        defects =
            Min(Min(tree->pawn_score.defects_k[side],
                tree->pawn_score.defects_e[side]),
            tree->pawn_score.defects_q[side]);
      else if (Castle(ply, side) == 1)
        defects =
            Min(tree->pawn_score.defects_k[side],
            tree->pawn_score.defects_e[side]);
      else
        defects =
            Min(tree->pawn_score.defects_q[side],
            tree->pawn_score.defects_e[side]);
      if (defects < 3)
        defects = 3;
    }
/*
 ************************************************************
 *                                                          *
 *   now fold in the king tropism and king pawn shelter     *
 *   scores together.                                       *
 *                                                          *
 ************************************************************
 */
    if (tree->tropism[enemy] < 0)
      tree->tropism[enemy] = 0;
    else if (tree->tropism[enemy] > 15)
      tree->tropism[enemy] = 15;
    if (defects > 15)
      defects = 15;
    score_mg -= king_safety[defects][tree->tropism[enemy]];
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 09/27/08 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateKingsFile computes defects for a file, based on whether the file  *
 *   is open or half-open.  if there are friendly pawns still on the file,     *
 *   they are penalized for advancing in front of the king.                    *
 *                                                                             *
 *******************************************************************************
 */
int EvaluateKingsFile(TREE * RESTRICT tree, int whichfile, int side)
{
  register int defects = 0, file;
  register int enemy = Flip(side);

  for (file = whichfile - 1; file <= whichfile + 1; file++) {
    if (!(file_mask[file] & tree->all_pawns))
      defects += open_file[file];
    else {
      if (!(file_mask[file] & Pawns(enemy)))
        defects += half_open_file[file] / 2;
      else
        defects +=
            pawn_defects[side][Rank(Advanced(enemy,
                    file_mask[file] & Pawns(enemy)))];
      if (!(file_mask[file] & Pawns(side)))
        defects += half_open_file[file];
      else {
        if (!(Pawns(side) & SetMask(sqflip[side][A2] + file))) {
          defects++;
          if (!(Pawns(side) & SetMask(sqflip[side][A3] + file)))
            defects++;
        }
      }
    }
  }
  return (defects);
}

/* last modified 07/06/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateKnights() is used to evaluate black/white knights.                *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateKnights(TREE * RESTRICT tree, int side)
{
  register BITBOARD temp, moves;
  register int square, s, i, t, score_eg = 0, score_mg = 0;
  register int enemy = Flip(side);
  int ray;

/*
 ************************************************************
 *                                                          *
 *   first fold in centralization score from the piece/     *
 *   square table "nval".                                   *
 *                                                          *
 ************************************************************
 */
  temp = Knights(side);
  while (temp) {
    square = LSB(temp);
    score_mg += nval[mg][side][square];
    score_eg += nval[eg][side][square];
/*
 ************************************************************
 *                                                          *
 *   now, evaluate for "outposts" which is a knight that    *
 *   can't be driven off by an enemy pawn, and which is     *
 *   supported by a friendly pawn.                          *
 *                                                          *
 *   if the enemy has NO minor to take this knight, then    *
 *   increase the bonus.                                    *
 *                                                          *
 ************************************************************
 */
    t = knight_outpost[side][square];
    if (t > 0) {
      if (!(mask_no_pattacks[enemy][square] & Pawns(enemy))) {
        score_eg += t;
        score_mg += t;
        if (t && pawn_attacks[enemy][square] & Pawns(side)) {
          score_eg += t / 2;
          score_mg += t / 2;
          if (!Knights(enemy) && !(Color(square) & Bishops(enemy))) {
            score_eg += t;
            score_mg += t;
          }
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   mobility counts the number of squares the piece        *
 *   attacks, excluding squares with friendly pieces,       *
 *   and weighs each square according to centralization.    *
 *                                                          *
 ************************************************************
 */
    moves = AttacksKnight(square) & ~Occupied(side);
    if (tree->cache_n[square] != moves) {
      tree->cache_n[square] = moves;
      t = -lower_n;
      for (i = 0; i < 4; i++)
        t += PopCnt(moves & mobility_mask_n[i]) * mobility_score_n[i];
      tree->cache_n_mobility[square] = t;
    }
    t = tree->cache_n_mobility[square];
    if (t > 0 && AbsolutePin(tree, side, square, &ray) != 0)
      t = 0;
    if (t > 0 && !IsActive(knight, moves))
      t /= 2;
    score_mg += t;
    score_eg += t;
/*
 ************************************************************
 *                                                          *
 *   adjust the tropism count for this piece.               *
 *                                                          *
 ************************************************************
 */
    if (tree->dangerous[side]) {
      i = Distance(square, KingSQ(enemy));
      tree->tropism[side] += king_tropism_n[i];
    }
    temp &= temp - 1;
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 09/27/08 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateMate() is used to evaluate positions where neither side has pawns *
 *   and one side has enough material to force checkmate.  it simply trys to   *
 *   force the losing king to the edge of the board, and then to the corner    *
 *   where mates are easier to find.                                           *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateMate(TREE * RESTRICT tree, int side)
{
  register int mate_score = 0;
  register int enemy = Flip(side);

/*
 ************************************************************
 *                                                          *
 *   if one side has a bishop+knight and the other side has *
 *   no pieces or pawns, then use the special bishop_knight *
 *   scoring board for the losing king to force it to the   *
 *   right corner for mate.                                 *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(enemy, occupied) == 0 &&
      TotalPieces(side, occupied) == p_vals[bishop] + p_vals[knight] &&
      TotalPieces(side, bishop) == 1) {
    if (dark_squares & Bishops(side))
      mate_score = b_n_mate_dark_squares[KingSQ(enemy)];
    else
      mate_score = b_n_mate_light_squares[KingSQ(enemy)];
  }
/*
 ************************************************************
 *                                                          *
 *   if one side is winning, force the enemy king to the    *
 *   edge of the board.                                     *
 *                                                          *
 ************************************************************
 */
  else {
    mate_score = mate[KingSQ(enemy)];
    mate_score -=
        (Distance(KingSQ(side), KingSQ(enemy)) - 3) * king_king_tropism;
  }
  tree->score_mg += sign[side] * mate_score;
  tree->score_eg += sign[side] * mate_score;
}

/* last modified 01/25/10 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateMaterial() adds bonuses/penalties for each piece based on the     *
 *   number of pawns on the board and the number of pieces of each type.       *
 *   this comes from Tracy Riegle's ideas on material imbalances.              *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateMaterial(TREE * RESTRICT tree)
{
  register int side;
  register int score_mg = Material, score_eg = Material;
  register int majors, minors, i, j;
  register int pair;

/*
 **********************************************************************
 *                                                                    *
 *   test 1.  if Majors or Minors are not balanced, then if one side  *
 *   is only an exchange up or down, we do not give any sort of bad   *
 *   trade penalty/bonus.                                             *
 *                                                                    *
 *   test 2.  if Majors or Minors are not balanced, then if one side  *
 *   has more piece material points than the other (using normal      *
 *   piece values of 3, 3, 5, 9 for N, B, R and Q) then the side that *
 *   is behind in piece material gets a penalty.                      *
 *                                                                    *
 **********************************************************************
 */
  majors =
      TotalPieces(white, rook) + 2 * TotalPieces(white,
      queen) - TotalPieces(black, rook) - 2 * TotalPieces(black, queen);
  minors =
      TotalPieces(white, knight) + TotalPieces(white,
      bishop) - TotalPieces(black, knight) - TotalPieces(black, bishop);
  j = majors + 4;
  i = minors + 4;
  if (j < 0)
   j = 0;
  else if (j > 8)
    j = 8;
  if (i < 0)
    i = 0;
  else if (i > 8)
    i = 8;
  score_mg += material_imbalance[j][i];
  score_eg += material_imbalance[j][i];
/*
 ************************************************************
 *                                                          *
 *   Add a bonus if the side has a pair of bishops, which   *
 *   can become very strong in open positions.              *
 *                                                          *
 ************************************************************
 */
  for (side = black; side <= white; side++) {
      pair = 0;
    if (TotalPieces(side, bishop) > 1)
      pair = 1;
    if (pair) {
      score_mg += sign[side] * bishop_pair[mg];
      score_eg += sign[side] * bishop_pair[eg];
    }
  }
  tree->score_mg += score_mg;
  tree->score_eg += score_eg;
}

/* last modified 06/14/18 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluatePassedPawns() is used to evaluate passed pawns and the danger     *
 *   they produce.  This code considers pieces as well, so it has been         *
 *   separated from the normal EvaluatePawns() code that hashes information    *
 *   based only on pawn positions.                                             *
 *                                                                             *
 *******************************************************************************
 */
void EvaluatePassedPawns(TREE * RESTRICT tree, int side) {
  register BITBOARD behind;
  register int file, square, rsq, score_mg = 0, score_eg = 0;
  register int pawns, rank, rank_bonus, dist_bonus, blocking;
  register int enemy = Flip(side);

/*
 ************************************************************
 *                                                          *
 *   Initialize.                                            *
 *                                                          *
 ************************************************************
 */
  for (pawns = tree->pawn_score.passed[side]; pawns; pawns &= pawns - 1) {
    file = LSB8Bit(pawns);
    square = Advanced(side, Pawns(side) & file_mask[file]);
    rank = Rank(square);
    rank_bonus = pp_rank_bonus[rankflip[side][rank]];
    dist_bonus = pp_dist_bonus[rankflip[side][rank]];
/*
 *****************************************************
 *                                                   *
 *   We have located the most advanced pawn on this  *
 *   file, which is the only one that will get any   *
 *   sort of bonus.  Add in the MG/EG scores first.  *
 *                                                   *
 *****************************************************
 */
    score_mg += passed_pawn_value[mg][side][rank];
    score_eg += passed_pawn_value[eg][side][rank];
/*
 *****************************************************
 *                                                   *
 *   Add in a bonus if the passed pawn is connected  *
 *   with another pawn for support.                  *
 *                                                   *
 *****************************************************
 */
    if (mask_pawn_connected[side][square] & Pawns(side)) {
      score_mg += connected_passed_pawn_value[mg] * rank_bonus;
      score_eg += connected_passed_pawn_value[eg] * rank_bonus;
    }
/*
 ************************************************************
 *                                                          *
 *   See if this pawn is either supported by a friendly     *
 *   rook or queen behind it, or if it is attacked by an    *
 *   enemy rook or queen from behind.                       *
 *                                                          *
 ************************************************************
 */
    if (Rooks(white) | Rooks(black)) {
      behind =
          ((side) ? minus8dir[square] : plus8dir[square]) & OccupiedSquares;
      if (behind) {
        rsq = Advanced(side, behind);
        if (PcOnSq(rsq) == pieces[side][rook]) {
          score_mg += rook_behind_passed_pawn[mg][rank];
          score_eg += rook_behind_passed_pawn[eg][rank];
        } else if (PcOnSq(rsq) == pieces[enemy][rook]) {
          score_mg -= rook_behind_passed_pawn[mg][rank];
          score_eg -= rook_behind_passed_pawn[eg][rank];
        }
      }
    }
/*
 *****************************************************
 *                                                   *
 *   If the pawn is blockaded by an enemy piece, it  *
 *   cannot move and is therefore not nearly as      *
 *   valuable as if it were free to advance.  If it  *
 *   is blocked by a friendly piece, it is penalized *
 *   1/2 or 1/3 of the normal blockade amount to     *
 *   encourage the blocking piece to move so the     *
 *   pawn can advance.                               *
 *                                                   *
 *****************************************************
 */
    blocking = square + direction[side];
    if (PcOnSq(blocking) * sign[side] < 0) {
      score_mg -= blockading_passed_pawn_value[mg][side][Rank(square)];
      score_eg -= blockading_passed_pawn_value[eg][side][Rank(square)];
    } else if (PcOnSq(blocking)) {
      score_mg -= blockading_passed_pawn_value[mg][side][Rank(square)] / 2;
      score_eg -= blockading_passed_pawn_value[eg][side][Rank(square)] / 3;
    }
/*
 *****************************************************
 *                                                   *
 *   Add in a bonus based on how close the friendly  *
 *   king is, and a penalty based on how close the   *
 *   enemy king is.  The bonus/penalty is based on   *
 *   how advanced the pawn is to attract the kings   *
 *   toward the most advanced (and most dangerous)   *
 *   passed pawn.                                    *
 *                                                   *
 *****************************************************
 */
    score_eg -=
        Distance(blocking, KingSQ(side)) * dist_bonus - Distance(blocking,
        KingSQ(enemy)) * dist_bonus;
/*
 *****************************************************
 *                                                   *
 *   Add in a bonus in endings with bishops of       *
 *   opposite colours if the side has a passed pawn  *
 *   to queen on the square of the same colour as    *
 *   the bishop.                                     *
 *                                                   *
 *****************************************************
 */
    if (TotalPieces(side, occupied) <= eg_mat && TotalPieces(enemy, occupied) <= eg_mat) {
      if (TotalPieces(side, bishop) == 1 && TotalPieces(enemy, bishop) == 1) {
        if (square_color[LSB(Bishops(side))] != square_color[LSB(Bishops(enemy))]) {
          register int bonus = 0;
          switch (File(square)) {
          case FILEA:
            if (square_color[sqflip[side][A8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEB:
            if (square_color[sqflip[side][B8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEC:
            if (square_color[sqflip[side][C8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILED:
            if (square_color[sqflip[side][D8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEE:
            if (square_color[sqflip[side][E8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEF:
            if (square_color[sqflip[side][F8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEG:
            if (square_color[sqflip[side][G8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          case FILEH:
            if (square_color[sqflip[side][H8]] == square_color[LSB(Bishops(side))])
              bonus = 1;
            break;
          };
          if (bonus) {
            score_mg += passed_pawn_value[mg][side][rank] / 2;
            score_eg += passed_pawn_value[eg][side][rank] / 2;
            if (mask_pawn_connected[side][square] & Pawns(side)) {
              score_mg += (connected_passed_pawn_value[mg] * rank_bonus) / 2;
              score_eg += (connected_passed_pawn_value[eg] * rank_bonus) / 2;
            }
          }
        }
      }
    }
  }
/*
 ************************************************************
 *                                                          *
 *   Check to see if side has an outside passed pawn.       *
 *                                                          *
 ************************************************************
 */
  if (tree->pawn_score.passed[side]) {
    if (is_outside[tree->pawn_score.passed[side]]
        [tree->pawn_score.all[enemy]]) {
      score_mg += outside_passed[mg];
      score_eg += outside_passed[eg];
    }
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 09/27/08 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluatePassedPawnRaces() is used to evaluate passed pawns when one       *
 *   side has passed pawns and the other side (or neither) has pieces.  in     *
 *   such a case, the critical question is can the defending king stop the pawn*
 *   from queening or is it too far away?  if only one side has pawns that can *
 *   "run" then the situation is simple.  when both sides have pawns that can  *
 *   "run" it becomes more complex as it then becomes necessary to see if      *
 *   one side can use a forced king move to stop the other side, while the     *
 *   other side doesn't have the same ability to stop ours.                    *
 *                                                                             *
 *   in the case of king and pawn endings with exactly one pawn, the simple    *
 *   evaluation rules are used:  if the king is two squares in front of the    *
 *   pawn then it is a win, if the king is one one square in front with the    *
 *   opposition, then it is a win,  if the king is on the 6th rank with the    *
 *   pawn close by, it is a win.  rook pawns are handled separately and are    *
 *   more difficult to queen because the king can get trapped in front of the  *
 *   pawn blocking promotion.                                                  *
 *                                                                             *
 *******************************************************************************
 */
void EvaluatePassedPawnRaces(TREE * RESTRICT tree, int wtm)
{
  register int file, square;
  register int queen_distance;
  register int pawnsq;
  register BITBOARD pawns;
  register int passed;
  register int side, enemy;
  BITBOARD runners[2] = { 0, 0 };
  int queener[2] = { 8, 8 };
  int psquare[2] = { 0, 0 };
  int ppawn[2] = { 0, 0 };
  int forced_km[2] = { 0, 0 };
/*
 ************************************************************
 *                                                          *
 *   check to see if side has one pawn and neither side     *
 *   has any pieces.  if so, use the simple pawn evaluation *
 *   logic.                                                 *
 *                                                          *
 ************************************************************
 */
  for (side = black; side <= white; side++) {
    enemy = Flip(side);
    if (Pawns(side) && !Pawns(enemy) && TotalPieces(white, occupied) == 0 &&
        TotalPieces(black, occupied) == 0) {
      pawns = Pawns(side);
      while (pawns) {
        pawnsq = LSB(pawns);
        pawns &= pawns - 1;
/*
 **************************************************
 *                                                *
 *   king must be in front of the pawn or we      *
 *   go no further.                               *
 *                                                *
 **************************************************
 */
        if (sign[side] * Rank(KingSQ(side)) <= sign[side] * Rank(pawnsq))
          continue;
/*
 **************************************************
 *                                                *
 *   first a special case.  if this is a rook     *
 *   pawn, then the king must be on the adjacent  *
 *   file, and be closer to the queening square   *
 *   than the opposing king.                      *
 *                                                *
 **************************************************
 */
        if (File(pawnsq) == FILEA) {
          if ((File(KingSQ(side)) == FILEB) &&
              (Distance(KingSQ(side),
                      sqflip[side][A8]) < Distance(KingSQ(enemy),
                      sqflip[side][A8]))) {
            tree->score_eg += sign[side] * pawn_can_promote;
            return;
          }
          continue;
        } else if (File(pawnsq) == FILEH) {
          if ((File(KingSQ(side)) == FILEG) &&
              (Distance(KingSQ(side),
                      sqflip[side][H8]) < Distance(KingSQ(enemy),
                      sqflip[side][H8]))) {
            tree->score_eg += sign[side] * pawn_can_promote;
            return;
          }
          continue;
        }
/*
 **************************************************
 *                                                *
 *   if king is two squares in front of the pawn  *
 *   then it's a win immediately.  if the king is *
 *   on the 6th rank and closer to the pawn than  *
 *   the opposing king, it's also a win.          *
 *                                                *
 **************************************************
 */
        if (Distance(KingSQ(side), pawnsq) < Distance(KingSQ(enemy), pawnsq)) {
          if (sign[side] * Rank(KingSQ(side)) >
              sign[side] * (Rank(pawnsq) - 1 + 2 * side)) {
            tree->score_eg += sign[side] * pawn_can_promote;
            return;
          }
          if (Rank(KingSQ(side)) == rankflip[side][RANK6]) {
            tree->score_eg += sign[side] * pawn_can_promote;
            return;
          }
        }
/*
 **************************************************
 *                                                *
 *   last chance:  if the king is one square in   *
 *   front of the pawn and has the opposition,    *
 *   then it's still a win.                       *
 *                                                *
 **************************************************
 */
        if ((Rank(KingSQ(side)) == Rank(pawnsq) - 1 + 2 * side) &&
            EvaluateHasOpposition(wtm == side, KingSQ(side), KingSQ(enemy))) {
          tree->score_eg += sign[side] * pawn_can_promote;
          return;
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   check to see if black is out of pieces and white has   *
 *   passed pawns.  if so, see if any of these passed pawns *
 *   can outrun the defending king and promote.             *
 *                                                          *
 ************************************************************
 */
    if (TotalPieces(enemy, occupied) == 0 && tree->pawn_score.passed[side]) {
      passed = tree->pawn_score.passed[side];
      while (passed) {
        file = LSB8Bit(passed);
        passed &= passed - 1;
        square = Advanced(side, Pawns(side) & file_mask[file]);
        forced_km[enemy] =
            (pawn_race[side][wtm][square] & Kings(enemy)) !=
            (pawn_race[side][Flip(wtm)][square] & Kings(enemy));
        if (!(pawn_race[side][wtm][square] & Kings(enemy))) {
          queen_distance = Abs(rankflip[side][RANK8] - Rank(square));
          if (Kings(side) & ((side) ? plus8dir[square] : minus8dir[square])) {
            if (file == FILEA || file == FILEH)
              queen_distance = 99;
            queen_distance++;
          }
          if (Rank(square) == rankflip[side][RANK2])
            queen_distance--;
          if (queen_distance < queener[side]) {
            runners[side] = SetMask(square);
            queener[side] = queen_distance;
            psquare[side] = file + (rankflip[side][RANK8] << 3);
            ppawn[side] = square;
          } else if (queen_distance == queener[side])
            runners[side] |= SetMask(square);
        }
      }
    }
  }
  if ((queener[white] == 8) && (queener[black] == 8))
    return;
/*
 ************************************************************
 *                                                          *
 *   now that we know which pawns can outrun the kings for  *
 *   each side, we need to do the following:                *
 *                                                          *
 *     (1) if both sides are forced to move their king to   *
 *         prevent the opponent from promoting, we let the  *
 *         search resolve this as the depth increases.      *
 *                                                          *
 *     (2) if white can run while black can not, or then    *
 *         white wins, or vice-versa.                       *
 *                                                          *
 *     (3) if white queens and black's king can't stop it   *
 *         no matter who moves first, while black has a     *
 *         pawn that white can stop if a king move is made  *
 *         immediately, then white wins, and vice-versa.    *
 *                                                          *
 *     (4) other situations are left to the search to       *
 *         resolve.                                         *
 *                                                          *
 ************************************************************
 */
  if (forced_km[white] & forced_km[black])
    return;
  if ((queener[white] < 8) && (queener[black] == 8)) {
    tree->score_eg += pawn_can_promote + (5 - queener[white]) * 10;
    return;
  } else if ((queener[black] < 8) && (queener[white] == 8)) {
    tree->score_eg += -(pawn_can_promote + (5 - queener[black]) * 10);
    return;
  }
  if (queener[white] < queener[black] && forced_km[white] && !forced_km[black]) {
    tree->score_eg += pawn_can_promote + (5 - queener[white]) * 10;
    return;
  } else if (queener[black] < queener[white] && forced_km[black] &&
      forced_km[white]) {
    tree->score_eg += -(pawn_can_promote + (5 - queener[black]) * 10);
    return;
  }
}

/* last modified 04/01/20 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluatePawns() is used to evaluate pawns.  it evaluates pawns for only   *
 *   one side, and fills in the pawn hash entry information.  it requires two  *
 *   calls to evaluate all pawns on the board.  comments below indicate the    *
 *   particular pawn structure features that are evaluated.                    *
 *                                                                             *
 *   this procedure also fills in information (without scoring) that other     *
 *   evaluation procedures use, such as which pawns are passed or candidates,  *
 *   which pawns are weak, which files are open, and so forth.                 *
 *                                                                             *
 *******************************************************************************
 */
void EvaluatePawns(TREE * RESTRICT tree, int side)
{
  register BITBOARD pawns;
  register BITBOARD temp;
  BITBOARD p_moves[2], doubled;
  register int pns, square, file, rank, score_eg = 0, score_mg = 0;
  register int defenders, attackers, sq, blocking;
  register int enemy = Flip(side);
  unsigned int pawn_files = 0;

/*
 ************************************************************
 *                                                          *
 *   initialize.                                            *
 *                                                          *
 ************************************************************
 */
  tree->pawn_score.all[side] = 0;
  tree->pawn_score.candidates[side] = 0;
  tree->pawn_score.passed[side] = 0;
/*
 ************************************************************
 *                                                          *
 *   first, determine which squares pawns can reach.        *
 *                                                          *
 ************************************************************
 */
  pawns = Pawns(side);
  p_moves[side] = 0;
  while (pawns) {
    square = LSB(pawns);
    tree->pawn_score.all[side] |= 1 << File(square);
    for (sq = square; sq != File(square) + ((side) ? RANK7 << 3 : RANK2 << 3);
        sq += direction[side]) {
      p_moves[side] |= SetMask(sq);
      if (SetMask(sq + direction[side]) & tree->all_pawns)
        break;
      defenders =
          PopCnt(pawn_attacks[enemy][sq + direction[side]] & Pawns(side));
      attackers =
          PopCnt(pawn_attacks[side][sq + direction[side]] & Pawns(enemy));
      if (attackers - defenders > 0)
        break;
    }
    pawns &= pawns - 1;
  }
/*
 ************************************************************
 *                                                          *
 *   now loop through all pawns for this side.              *
 *                                                          *
 ************************************************************
 */
  pawns = Pawns(side);
  while (pawns) {
    register int isolated = 0;
    square = LSB(pawns);
    file = File(square);
    rank = Rank(square);
    pawn_files |= 1 << file;
    pawns &= pawns - 1;
/*
 ************************************************************
 *                                                          *
 *   the first thing we do is make a note that the current  *
 *   file can't be open since there is a pawn on it.        *
 *                                                          *
 ************************************************************
 */
    tree->pawn_score.open_files &= ~(1 << file);
/*
 ************************************************************
 *                                                          *
 *   evaluate pawn advances.  center pawns are encouraged   *
 *   to advance, while wing pawns are pretty much neutral.  *
 *   this is a simple piece/square value.                   *
 *                                                          *
 ************************************************************
 */
    score_mg += pval[mg][side][square];
    score_eg += pval[eg][side][square];
/*
 ************************************************************
 *                                                          *
 *   evaluate isolated pawns, which are penalized based on  *
 *   the file, with central isolani being worse than when   *
 *   on the wings.                                          *
 *                                                          *
 ************************************************************
 */
    if (!(mask_pawn_isolated[square] & Pawns(side))) {
      score_mg -= pawn_isolated[mg];
      score_eg -= pawn_isolated[eg];
      if (!(Pawns(enemy) & file_mask[file])) {
        score_mg -= pawn_isolated_open_file[mg];
        score_eg -= pawn_isolated_open_file[eg];
      }
      isolated = 1;
    }
/*
 ************************************************************
 *                                                          *
 *  evaluate weak pawns.  weak pawns are evaluated by the   *
 *  following rules:  (1) if a pawn is defended by a pawn,  *
 *  it isn't weak;  (2) if a pawn is undefended by a pawn   *
 *  and advances one (or two if it hasn't moved yet) ranks  *
 *  and is defended fewer times than it is attacked, it is  *
 *  weak.  note that the penalty is greater if the pawn is  *
 *  on an open file.  note that an isolated pawn is just    *
 *  another case of a weak pawn, since it can never be      *
 *  defended by a pawn.                                     *
 *                                                          *
 *  first, test the pawn where it sits to determine if it   *
 *  is defended more times than attacked.  if so, it is not *
 *  weak and we are done.  if it is weak where it sits, can *
 *  it advance one square and become not weak.  if so we    *
 *  are again finished with this pawn.  otherwise we fall   *
 *  into the next test.                                     *
 *                                                          *
 ************************************************************
 */
    else {
      register int weak = 0;
      do {
        attackers = 0;
        defenders = 0;
        temp = p_moves[side] & ((side) ? plus8dir[square] : minus8dir[square]);
        while (temp) {
          sq = LSB(temp);
          temp &= temp - 1;
          defenders = PopCnt(pawn_attacks[enemy][sq] & Pawns(side));
          attackers = PopCnt(pawn_attacks[side][sq] & Pawns(enemy));
          if (defenders && defenders >= attackers)
            break;
        }
/*
 ************************************************************
 *                                                          *
 *  if the pawn is defended by a pawn then this pawn is not *
 *  weak.                                                   *
 *                                                          *
 ************************************************************
 */
        if (defenders && defenders >= attackers)
          break;
/*
 ************************************************************
 *                                                          *
 *  if the pawn can be defended by a pawn, and that pawn    *
 *  can safely advance to actually defend this pawn, then   *
 *  this pawn is not weak, otherwise the pawn is weak.      *
 *                                                          *
 ************************************************************
 */
        if (!(pawn_attacks[enemy][square] & p_moves[side])) {
          weak = 1;
        }
      } while (0);
/*
 ************************************************************
 *                                                          *
 *  evaluate unsupported pawns, which provide a target      *
 *  since they are undefended by a pawn.  we exclude pawns  *
 *  that are isolated and weak since they have already been *
 *  given a penalty.                                        *
 *                                                          *
 ************************************************************
 */
      if (!weak && !isolated)
        if (!(Pawns(side) & pawn_attacks[enemy][square]))
        {
          score_mg -= pawn_unsupported[mg];
          score_eg -= pawn_unsupported[eg];
        }
/*
 ************************************************************
 *                                                          *
 *  test the pawn to see it if forms a "duo" which is two   *
 *  pawns side-by-side.                                     *
 *                                                          *
 ************************************************************
 */
      if (!weak && !isolated)
        if (mask_pawn_duo[square] & Pawns(side)) {
          score_mg += pawn_duo[mg];
          score_eg += pawn_duo[eg];
        }
    }
/*
 ************************************************************
 *                                                          *
 *  evaluate doubled pawns.  if there are other pawns on    *
 *  this file in front of this pawn, penalize this pawn.    *
 *  note that this will NOT penalize both pawns, just the   *
 *  most rearward one that is really almost worthless.      *
 *                                                          *
 *  the farther apart two doubled pawns (same file) are,    *
 *  the less weak they are, so the penalty is reduced as    *
 *  this distance increases.                                *
 *                                                          *
 ************************************************************
 */
    doubled = Pawns(side) & ((side) ? plus8dir[square] : minus8dir[square]);
    if (doubled) {
      score_mg -=
          doubled_pawn_value[mg][file] / RankDistance(square, MostAdvanced(side,
              doubled));
      score_eg -=
          doubled_pawn_value[eg][file] / RankDistance(square, MostAdvanced(side,
              doubled));
    }
/*
 ************************************************************
 *                                                          *
 *   discover and flag passed pawns for use later.          *
 *                                                          *
 ************************************************************
 */
    if (!(mask_pawn_passed[side][square] & Pawns(enemy))) {
      tree->pawn_score.passed[side] |= 1 << file;
    }
/*
 ************************************************************
 *                                                          *
 *   now determine if this pawn is a candidate passer,      *
 *   since we now know it isn't passed.  a candidate is a   *
 *   pawn on a file with no enemy pawns in front of it, and *
 *   if it advances until it contacts an enemy pawn, and it *
 *   is defended as many times as it is attacked when it    *
 *   reaches that pawn, then all that is left is to see if  *
 *   it is passed when the attacker(s) get removed.         *
 *                                                          *
 ************************************************************
 */
    else {
      if (!(file_mask[File(square)] & Pawns(enemy)) &&
          mask_pawn_isolated[square] & Pawns(side) &&
          !(pawn_attacks[side][square] & Pawns(enemy))) {
        attackers = 1;
        defenders = 0;
        for (sq = square;
            sq != File(square) + ((side) ? RANK7 << 3 : RANK2 << 3);
            sq += direction[side]) {
          if (SetMask(sq + direction[side]) & tree->all_pawns)
            break;
          defenders = PopCnt(pawn_attacks[enemy][sq] & p_moves[side]);
          attackers = PopCnt(pawn_attacks[side][sq] & Pawns(enemy));
          if (attackers)
            break;
        }
        if (attackers <= defenders) {
          if (!(mask_pawn_passed[side][sq + direction[side]] & Pawns(enemy))) {
            tree->pawn_score.candidates[side] |= 1 << file;
            score_mg += passed_pawn_candidate[mg][side][rank];
            score_eg += passed_pawn_candidate[eg][side][rank];
          }
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   evaluate "hidden" passed pawns.  simple case is a pawn *
 *   chain (white) at b5, a6, with a black pawn at a7.      *
 *   it appears the b-pawn is backward, with a ram at a6/a7 *
 *   but this is misleading, because the pawn at a6 is      *
 *   really passed when white plays b6.                     *
 *                                                          *
 ************************************************************
 */
    if (Rank(square) == ((side) ? RANK6 : RANK3) &&
        SetMask(square + direction[side]) & Pawns(enemy) &&
        ((File(square) < FILEH && SetMask(square + 9 - 16 * side) & Pawns(side)
                && !(mask_hidden_right[side][File(square)] & Pawns(enemy))) ||
            (File(square) > FILEA &&
                SetMask(square + 7 - 16 * side) & Pawns(side) &&
                !(mask_hidden_left[side][File(square)] & Pawns(enemy))))) {
      score_mg += passed_pawn_hidden[mg];
      score_eg += passed_pawn_hidden[eg];
    }
  }
/*
 ************************************************************
 *                                                          *
 *  Give a bonus for distance between left-most pawn and    *
 *  right-most pawn.  The idea is that the wider the gap    *
 *  between the pawns, the harder they are for a lone king  *
 *  to control in the endgame.  Botvinnik referred to this  *
 *  concept as "trousers" (pants with two legs, the farther *
 *  the legs are apart, the better for the side with those  *
 *  pawns).                                                 *
 *                                                          *
 ************************************************************
 */
  score_eg += trousers_bonus * (MSB8Bit(pawn_files) - LSB8Bit(pawn_files));
/*
 ************************************************************
 *                                                          *
 *   now evaluate king safety.                              *
 *                                                          *
 *   this uses the function EvaluateKingsFile() which       *
 *   looks at four possible positions for the king, either  *
 *   castled kingside, queenside or else standing on the    *
 *   d or e file stuck in the middle.  this essentially is  *
 *   about the pawns in front of the king and what kind of  *
 *   "shelter" they provide for the king during the         *
 *   middlegame.                                            *
 *                                                          *
 ************************************************************
 */
  tree->pawn_score.defects_q[side] = EvaluateKingsFile(tree, FILEB, side);
  tree->pawn_score.defects_d[side] = EvaluateKingsFile(tree, FILED, side);
  tree->pawn_score.defects_e[side] = EvaluateKingsFile(tree, FILEE, side);
  tree->pawn_score.defects_k[side] = EvaluateKingsFile(tree, FILEG, side);
  tree->pawn_score.score_mg += sign[side] * score_mg;
  tree->pawn_score.score_eg += sign[side] * score_eg;
}

/* last modified 07/06/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateQueens() is used to evaluate black/white queens.                  *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateQueens(TREE * RESTRICT tree, int side)
{
  register BITBOARD temp, moves;
  register int square, i, t, score_mg = 0, score_eg = 0;
  register int enemy = Flip(side);
  int trop;
  int ray;

/*
 ************************************************************
 *                                                          *
 *   first locate each queen and obtain it's centralization *
 *   score from the static piece/square table for queens.   *
 *                                                          *
 ************************************************************
 */
  temp = Queens(side);
  while (temp) {
    square = LSB(temp);
/*
 ************************************************************
 *                                                          *
 *   first, add in the piece/square table value for the     *
 *   queen                                                  *
 *                                                          *
 ************************************************************
*/
    score_mg += qval[mg][side][square];
    score_eg += qval[eg][side][square];
/*
 ************************************************************
 *                                                          *
 *   mobility counts the number of squares the piece        *
 *   attacks and weighs each square according to relevance. *
 *                                                          *
 ************************************************************
 */
    moves = AttacksQueen(square, OccupiedSquares) & ~Occupied(side);
    if (tree->cache_q[square] != moves) {
      tree->cache_q[square] = moves;
      moves = AttacksQueen(square, OccupiedSquares) & ~Occupied(side);
      t = -lower_q;
      for (i = 0; i < 4; i++)
        t += PopCnt(moves & (mobility_mask_b[i] | mobility_mask_r[i])) * mobility_score_q[i];
      tree->cache_q_mobility[square] = t;
    }
    t = tree->cache_q_mobility[square];
    if (t > 0 && AbsolutePin(tree, side, square, &ray) != 0) {
      if (ray == 1 || ray == 8)
        t /= 2;
      else if (ray == 7 || ray == 9)
        t /= 2;
    }
    if (t > 0 && !IsActive(queen, moves))
      t /= 2;
    score_mg += t;
    score_eg += t;
/*
 ************************************************************
 *                                                          *
 *   adjust the tropism count for this piece.               *
 *                                                          *
 *   now we notice whether the queen is on a file that is   *
 *   bearing on the enemy king and adjust tropism if so.    *
 *                                                          *
 ************************************************************
 */
    if (tree->dangerous[side]) {
      trop = Distance(square, KingSQ(enemy));
      tree->tropism[side] += king_tropism_q[trop];
    }
    tree->tropism[enemy] -= friendly_queen[Distance(square, KingSQ(side))];
    temp &= temp - 1;
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 07/06/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateRooks() is used to evaluate black/white rooks.                    *
 *                                                                             *
 *******************************************************************************
 */
void EvaluateRooks(TREE * RESTRICT tree, int side)
{
  register BITBOARD temp, moves, moves2, pb;
  register int square, s, file, trop, t, u, n, p, i, f;
  register int score_mg = 0, score_eg = 0;
  register int enemy = Flip(side);
  int ray;

/*
 ************************************************************
 *                                                          *
 *   Initialize.                                            *
 *                                                          *
 ************************************************************
 */
  temp = Rooks(side);
  while (temp) {
    square = LSB(temp);
    file = File(square);
/*
 ************************************************************
 *                                                          *
 *   determine if the rook is on an open file or on a half- *
 *   open file, either of which increases its ability to    *
 *   attack important squares.                              *
 *                                                          *
 ************************************************************
 */
    if (!(file_mask[file] & Pawns(side))) {
      if (!(file_mask[file] & Pawns(enemy))) {
        score_mg += rook_open_file[mg];
        score_eg += rook_open_file[eg];
      } else {
        score_mg += rook_half_open_file[mg];
        score_eg += rook_half_open_file[eg];
      }
    }
/*
 ************************************************************
 *                                                          *
 *   determine if the rook is on the 7th rank, with the     *
 *   enemy king trapped on the 8th rank.  if so the rook    *
 *   exerts a "cramping" effect that is valuable.           *
 *                                                          *
 ************************************************************
 */
    if (Rank(square) == rankflip[side][RANK7] &&
      (Rank(KingSQ(enemy)) == rankflip[side][RANK8]
        || Pawns(enemy) & rank_mask[rankflip[side][RANK7]])) {
      score_mg += rook_on_7th[mg];
      score_eg += rook_on_7th[eg];
    }
/*
 *****************************************************
 *                                                   *
 *   add in a bonus if the rook is connected with    *
 *   another rook for support.                       *
 *                                                   *
 *****************************************************
 */
    pb = Rooks(side);
    while (pb) {
      f = Advanced(side, pb);
      if (AttacksRook(f, OccupiedSquares) & Rooks(side)) {
        score_mg += connected_rook_value[mg];
        score_eg += connected_rook_value[eg];
      }
      Clear(f, pb);
    }
/*
 ************************************************************
 *                                                          *
 *   check to see if the king has been forced to move and   *
 *   has trapped a rook at a1/a2/b1/g1/h1/h2, if so, then   *
 *   penalize the trapped rook to help extricate it.        *
 *                                                          *
 ************************************************************
 */
    if (Rank(KingSQ(side)) == rankflip[side][RANK1]) {
      if (File(KingSQ(side)) < FILED) {
        if (sqflip[side][square] < sqflip[side][KingSQ(side)]) {
          score_mg -= rook_trapped;
          score_eg -= rook_trapped;
        }
      } else if (File(KingSQ(side)) > FILEE) {
        if (sqflip[side][square] > sqflip[side][KingSQ(side)] &&
            sqflip[side][square] < 8) {
          score_mg -= rook_trapped;
          score_eg -= rook_trapped;
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   mobility counts the number of squares the piece        *
 *   attacks, excluding squares with friendly pieces, and   *
 *   weighs each square according to centralization (file). *
 *                                                          *
 *   For efficiency, we use a small cache to eliminate the  *
 *   rook mobility calculation cost if the pieces on the    *
 *   same rank/file with this rook are identical to the     *
 *   placement the last time we evaluated a rook on this    *
 *   same square.                                           *
 *                                                          *
 ************************************************************
 */
    moves = AttacksRook(square, OccupiedSquares) & Occupied(side);
    moves2 = AttacksRook(square, OccupiedSquares) & Occupied(enemy);
    if (tree->cache_r_friendly[square] != moves ||
        tree->cache_r_enemy[square] != moves2) {
      tree->cache_r_friendly[square] = moves;
      tree->cache_r_enemy[square] = moves2;
      moves = AttacksRook(square, OccupiedSquares) & ~Occupied(side);
      t = 0;
      n = -5;
      for (i = 0; i < 4; i++) {
        p = PopCnt(moves & mobility_mask_r[i]);
        t += p * mobility_score_r[i];
        n += p;
      }
      t = Max(t, 0);
      t = Min(t, 47);
      tree->cache_r_mobility_mg[square] = mob_curve_r[t];
      tree->cache_r_mobility_eg[square] = 3 * n;
    }
    t = tree->cache_r_mobility_mg[square];
    u = tree->cache_r_mobility_eg[square];
    if (t > 0 && AbsolutePin(tree, side, square, &ray) != 0) {
      if (ray == 7 || ray == 9) {
        t /= 2;
        u /= 2;
      }
      else if (ray == 1 || ray == 8) {
        t /= 2;
        u /= 2;
      }
    }
    if (t > 0 && !IsActive(rook, moves)) {
      t /= 2;
      u /= 2;
    }
    score_mg += t;
    score_eg += u;
/*
 ************************************************************
 *                                                          *
 *   adjust the tropism count for this piece.               *
 *                                                          *
 ************************************************************
 */
    if (tree->dangerous[side]) {
      trop =
          (AttacksRook(square,
              OccupiedSquares & ~(Queens(side) | (Rooks(side)))) &
          king_attacks[KingSQ(enemy)]) ? 1 : Distance(square, KingSQ(enemy));
      tree->tropism[side] += king_tropism_r[trop];
    }
    temp &= temp - 1;
  }
  tree->score_mg += sign[side] * score_mg;
  tree->score_eg += sign[side] * score_eg;
}

/* last modified 01/25/10 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateWinningChances() is used to determine if one side has reached a   *
 *   position which can not be won, period, even though side may be ahead in   *
 *   material in some way.                                                     *
 *                                                                             *
 *   return values:                                                            *
 *        0    ->     side on move can not win.                                *
 *        1    ->     side on move can win.                                    *
 *                                                                             *
 *******************************************************************************
 */
int EvaluateWinningChances(TREE * RESTRICT tree, int side, int wtm)
{
  register int square, fkd, ekd, pd, promote;
  int enemy = Flip(side);

/*
 ************************************************************
 *                                                          *
 *   if side has a piece and no pawn, it can not possibly   *
 *   win.  if side is a piece ahead, the only way it can    *
 *   win is if the enemy is already trapped on the edge of  *
 *   the board (special case to handle KRB vs KR which can  *
 *   be won if the king gets trapped).                      *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn) == 0) {
    if (TotalPieces(side, occupied) <= p_vals[knight])
      return (0);
    if (TotalPieces(side, occupied) - TotalPieces(enemy, occupied) <= p_vals[knight] &&
        mask_not_edge & Kings(enemy))
      return (0);
  }
/*
 ************************************************************
 *                                                          *
 *   if "side" has a pawn, then either the pawn had better  *
 *   not be a rook pawn, or else white had better have the  *
 *   right color bishop or any other piece, otherwise it is *
 *   not winnable if the black king can get to the queening *
 *   square first.                                          *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn) && !(Pawns(side) & not_rook_pawns))
    do {
      if (TotalPieces(side, occupied) > p_vals[knight] || (TotalPieces(side, occupied) == p_vals[knight]
              && Knights(side)))
        continue;
      if (TotalPieces(side, occupied) == 0 && file_mask[FILEA] & Pawns(side) &&
          file_mask[FILEH] & Pawns(side))
        continue;
      if (Bishops(side)) {
        if (!TotalPieces(enemy, occupied)) {
          if (Bishops(side) & dark_squares) {
            if (file_mask[dark_corner[side]] & Pawns(side))
              continue;
          } else if (file_mask[light_corner[side]] & Pawns(side))
            continue;
        } else
          continue;
      }
      if (!(Pawns(side) & file_mask[FILEA]) ||
          !(Pawns(side) & file_mask[FILEH])) {
        if (Pawns(side) & file_mask[FILEA])
          promote = A8;
        else
          promote = H8;
        ekd = Distance(KingSQ(enemy), sqflip[side][promote]) - (wtm != side);
        if (ekd <= 1)
          return (0);
        else {
          fkd = Distance(KingSQ(side), sqflip[side][promote]) - (wtm == side);
          pd = Distance(Advanced(side, Pawns(side) & file_mask[File(promote)]),
              sqflip[side][promote]) - (wtm == side);
          if (ekd <= fkd && (ekd - 1) <= pd)
            return (0);
        }
      }
    } while (0);
/*
 ************************************************************
 *                                                          *
 *   if both sides have pawns, and we have made it through  *
 *   the previous tests, then this side has winning         *
 *   chances.                                               *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn))
    return (1);
/*
 ************************************************************
 *                                                          *
 *   if side has two bishops, and the enemy has only a      *
 *   single kinght, the two bishops win.                    *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn) == 0 && TotalPieces(side, occupied) == 2 * p_vals[bishop] &&
      TotalPieces(enemy, occupied) == p_vals[knight] && (Knights(side) || !Knights(enemy)))
    return (0);
/*
 ************************************************************
 *                                                          *
 *   if one side is two knights ahead and the opponent has  *
 *   no remaining material, it is a draw.                   *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn) == 0 && TotalPieces(side, occupied) == 2 * p_vals[knight] &&
      !Bishops(side) &&
      TotalPieces(enemy, occupied) == 0 && TotalPieces(enemy, pawn) == 0)
    return (0);
/*
 ************************************************************
 *                                                          *
 *   check to see if this is a KRP vs KR or KQP vs KQ type  *
 *   ending.  if so, and the losing king is in front of the *
 *   passer, then this is a drawish ending.                 *
 *                                                          *
 ************************************************************
 */
  if (TotalPieces(side, pawn) == 1 && TotalPieces(enemy, pawn) == 0 &&
      ((TotalPieces(side, occupied) == p_vals[rook] && TotalPieces(enemy, occupied) == p_vals[rook])
          || ((TotalPieces(side, occupied) == p_vals[queen] &&
                  TotalPieces(enemy, occupied) == p_vals[queen])))) {
    square = LSB(Pawns(side));
    if (FileDistance(KingSQ(enemy), square) <= 1 &&
        FrontOf(side, Rank(KingSQ(enemy)), Rank(square)))
      return (0);
    else if (FileDistance(KingSQ(side), square) > 1 ||
        Behind(side, Rank(KingSQ(side)), Rank(square)))
      return (0);
  }
  return (1);
}

/* last modified 04/25/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   EvaluateNNUE() is used to evaluate the chess board.  it uses NNUE C++     *
 *   code to statically evaluate a current chess position. ``An efficiently    *
 *   updatable neural network (NNUE), is a neural network-based evaluation     *
 *   function that runs efficiently on central processing units without a      *
 *   requirement for a graphics processing unit (GPU). NNUE was invented by    *
 *   Yu Nasu and introduced to computer shogi in 2018. On 6 August 2020, NNUE  *
 *   was integrated into the chess engine Stockfish.''                         *
 *                                                                             *
 *******************************************************************************
 */
int EvaluateNNUE(TREE * RESTRICT tree, int ply, int wtm, int alpha, int beta)
{
    register int score = 0, can_win = CAN_WIN;

/*
 **********************************************************************
 *                                                                    *
 *   initialize.                                                      *
 *                                                                    *
 **********************************************************************
 */
  maximum_ply = Max(maximum_ply, ply);
  tree->evaluations++;
/*
 **********************************************************************
 *                                                                    *
 *   check for draws due to insufficient material and adjust the      *
 *   score as necessary.  this code also handles a special endgame    *
 *   case where one side has only a lone king, and the king has no    *
 *   legal moves.  this has been shown to break a few evaluation      *
 *   terms such as bishop + wrong color rook pawn.  if this case is   *
 *   detected, a drawscore is returned.                               *
 *                                                                    *
 **********************************************************************
 */
  if (TotalPieces(white, occupied) <= eg_mat && TotalPieces(black, occupied) <= eg_mat)
    do {
      register int majors, minors;
/*
 ************************************************************
 *                                                          *
 *   if neither side has any pieces, and both sides have    *
 *   non-rookpawns, then either side can win.               *
 *                                                          *
 ************************************************************
 */
      if (TotalPieces(white, occupied) == 0 && TotalPieces(black, occupied) == 0
          && (Pawns(white) & not_rook_pawns && Pawns(black) & not_rook_pawns))
        break;
/*
 ************************************************************
 *                                                          *
 *   if one side is an exchange up, but has no pawns, then  *
 *   that side can not possibly win.                        *
 *                                                          *
 ************************************************************
 */
      majors =
          TotalPieces(white, rook) + 2 * TotalPieces(white,
          queen) - TotalPieces(black, rook) - 2 * TotalPieces(black, queen);
      if (Abs(majors) == 1) {
        minors =
            TotalPieces(white, knight) + TotalPieces(white,
            bishop) - TotalPieces(black, knight) - TotalPieces(black, bishop);
        if (majors == -minors) {
          if (TotalPieces(black, pawn) == 0)
            SetCannotWin(black);
          if (TotalPieces(white, pawn) == 0)
            SetCannotWin(white);
        }
        if (can_win == CANNOT_WIN)
          break;
      }
/*
 ************************************************************
 *                                                          *
 *   now check several special cases, such as bishop + the  *
 *   wrong rook pawn and adjust can_win accordingly.        *
 *                                                          *
 ************************************************************
 */
      if (!EvaluateWinningChances(tree, white, wtm))
        SetCannotWin(white);
      if (!EvaluateWinningChances(tree, black, wtm))
        SetCannotWin(black);
    } while (0);
/*
 **********************************************************************
 *                                                                    *
 *   now evaluate pieces.  we call here NNUEEvaluate() to do an       *
 *   efficiently updatable neural network (NNUE) evaluation, using    *
 *   NNUE C++ code, which is capable of using Stockfish trained       *
 *   NNUEs.                                                           *
 *                                                                    *
 **********************************************************************
 */
  score = NNUEEvaluate(tree, ply, wtm, alpha, beta);
/*
 **********************************************************************
 *                                                                    *
 *   now adjust the score if the game is drawish but one side appears *
 *   to be significantly better according to the computed score.      *
 *                                                                    *
 **********************************************************************
 */
  score = EvaluateDraws(tree, wtm, ply, can_win, score);
/*#if defined(SKILL)
  score =
      skill * score / 100 + ((100 -
          skill) * PAWN_VALUE * (UINT64) Random32() / 0x100000000ULL) / 100;
#endif*/
  return ((wtm) ? score : -score); /* from side to move perspective */
}

/* last modified 03/28/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   NNUEEvaluate() is a wrapper around NNUE C++ nnue_evaluate() function,     *
 *   where we prepare player, pieces and squares, as required by               *
 *   nnue_evaluate(), and then call nnue_evaluate() to obtain the score.       *
 *                                                                             *
 *******************************************************************************
 */
int NNUEEvaluate(TREE * RESTRICT tree, int ply, int wtm, int alpha, int beta)
{
  register int score = 0;
  register BITBOARD temp;
  register int square;
  int nnue_counter = 0;
  int nnue_player;
  int nnue_pieces[NNUE_ARRAY];
  int nnue_squares[NNUE_ARRAY];

/*
 ************************************************************
 *                                                          *
 *   NNUE input player.                                     *
 *                                                          *
 ************************************************************
 */
  nnue_player = (wtm) ? nnue_white : nnue_black;

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for kings.              *
 *                                                          *
 ************************************************************
 */
  square = KingSQ(white);
  nnue_pieces[nnue_counter] = nnue_wking;
  nnue_squares[nnue_counter] = square;
  nnue_counter++;
  square = KingSQ(black);
  nnue_pieces[nnue_counter] = nnue_bking;
  nnue_squares[nnue_counter] = square;
  nnue_counter++;

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for pawns.               *
 *                                                          *
 ************************************************************
 */
  temp = Pawns(white);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_wpawn;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }
  temp = Pawns(black);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_bpawn;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for knights.             *
 *                                                          *
 ************************************************************
 */
  temp = Knights(white);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_wknight;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }
  temp = Knights(black);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_bknight;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for bishops.             *
 *                                                          *
 ************************************************************
 */
  temp = Bishops(white);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_wbishop;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }
  temp = Bishops(black);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_bbishop;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for rooks.               *
 *                                                          *
 ************************************************************
 */
  temp = Rooks(white);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_wrook;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }
  temp = Rooks(black);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_brook;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for queens.              *
 *                                                          *
 ************************************************************
 */
  temp = Queens(white);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_wqueen;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }
  temp = Queens(black);
  while (temp) {
    square = LSB(temp);
    nnue_pieces[nnue_counter] = nnue_bqueen;
    nnue_squares[nnue_counter] = square;
    nnue_counter++;
    temp &= temp - 1;
  }

/*
 ************************************************************
 *                                                          *
 *   NNUE input pieces and squares for end of array.        *
 *                                                          *
 ************************************************************
 */
  nnue_pieces[nnue_counter] = nnue_empty;
  nnue_squares[nnue_counter] = BAD_SQUARE;

/*
 ************************************************************
 *                                                          *
 *   call NNUE_evaluate() to get the NNUE score.            *
 *                                                          *
 ************************************************************
 */
  score = NNUE_evaluate(nnue_player, nnue_pieces,
    nnue_squares, tree->thread_id);

  return ((wtm) ? score : -score); /* from white's perspective */
}

