#include "chess.h"
#include "data.h"

/* last modified 06/23/14 */
/*
 *******************************************************************************
 *                                                                             *
 *   History() maintains the history counters.  each time a move fails high,   *
 *   it's history count is incremented by depth * depth.  all the moves prior  *
 *   to that move have their history counters decremented by the 2x that       *
 *   amount, so that likely fail-high moves get searched first.                *
 *                                                                             *
 *   history heuristic is a dynamic move ordering method based on the number   *
 *   of cutoffs caused by a given move irrespectively from the position in     *
 *   which the move has been made.  the heuristic was invented by Jonathan     *
 *   Schaeffer in 1983 and works as follows: on a cutoff we increment a        *
 *   counter in a special table addressed by [piece][to].  the added value is  *
 *   depth * depth, based on the assumption that otherwise moves from the      *
 *   plies near the leaves would have to much impact on the result.  values    *
 *   retrieved from that table are used to order non-capturing moves.  this    *
 *   simple heuristics performs usually better than domain-dependent           *
 *   heuristics, though it may be combined with them.  for example, in Rebel   *
 *   only a few non-captures are ordered by history heuristics, then a piece-  *
 *   square approach is used.                                                  *
 *                                                                             *
 *******************************************************************************
 */
void History(TREE * RESTRICT tree, int ply, int depth, int side, int move,
    int searched[], int count) {
  int i, index, indexm, age = 0;
  if (move == NULL_MOVE)
    return;
  if (count < 1)
    return;
/*
 ************************************************************
 *                                                          *
 *  If the best move so far is a capture or a promotion,    *
 *  return, since we try good captures and promotions       *
 *  before searching history heuristic moves anyway.        *
 *                                                          *
 ************************************************************
 */
  if (CaptureOrPromote(move))
    return;
  Lock(lock_his);
/*
 ************************************************************
 *                                                          *
 *  Set the counter-move for the move at the previous ply   *
 *  to be this move that caused the fail-high.              *
 *                                                          *
 ************************************************************
 */
  if (ply > 1) {
    if (counter_move[tree->curmv[ply - 1] & 4095].move1 != move) {
      counter_move[tree->curmv[ply - 1] & 4095].move2 =
          counter_move[tree->curmv[ply - 1] & 4095].move1;
      counter_move[tree->curmv[ply - 1] & 4095].move1 = move;
    }
  }
/*
 ************************************************************
 *                                                          *
 *  Set the move-pair for the move two plies back so that   *
 *  if we play that move again, we will follow up with this *
 *  move to continue the "plan".                            *
 *                                                          *
 ************************************************************
 */
  if (ply > 2) {
    if (move_pair[tree->curmv[ply - 2] & 4095].move1 != move) {
      move_pair[tree->curmv[ply - 2] & 4095].move2 =
          move_pair[tree->curmv[ply - 2] & 4095].move1;
      move_pair[tree->curmv[ply - 2] & 4095].move1 = move;
    }
  }
/*
 ************************************************************
 *                                                          *
 *  adjust the history counter for the move that caused the *
 *  fail-high, limiting the max value to 2^20.              *
 *                                                          *
 ************************************************************
 */
  if (depth > 5) {
    indexm = HistoryIndex(move);
    history[side][indexm] += depth * depth;
    if (history[side][indexm] > (1 << 20))
      age = 1;
/*
 ************************************************************
 *                                                          *
 *  adjust the history counters for the moves that were     *
 *  searched but did not cause a fail-high, limiting the    *
 *  min value to -2^20.                                     *
 *                                                          *
 ************************************************************
 */
    for (i = 0; i < count; i++) {
      index = HistoryIndex(searched[i]);
      if (index != indexm) {
        history[side][index] -= depth * depth * 2;
        if (history[side][index] <= (-1 << 20))
          age = 1;
      }
    }
/*
 ************************************************************
 *                                                          *
 *  if any value reached +/-2^20, we divide all entries for *
 *  this side by 4 so that they don't threaten to wrap      *
 *  around which would wreck the counters.                  *
 *                                                          *
 ************************************************************
 */
    if (age)
      for (i = 0; i < 512; i++)
        history[side][i] = history[side][i] >> 4;
  }
  Unlock(lock_his);
}

/* last modified 06/23/14 */
/*
 *******************************************************************************
 *                                                                             *
 *   HistoryAge() is used to age the history data.  since we use saturating    *
 *   counters (counters reach 2^16 and stop increasing, or -2^16 and stop      *
 *   decreasing, we want to prevent counters from reaching some high value     *
 *   and being accessed rarely, but then giving a bogus impression about how   *
 *   good or bad the move is.  we do this at the start of a new search (not a  *
 *   new iteration) and we divide every value by 256.                          *
 *                                                                             *
 *******************************************************************************
 */
void HistoryAge() {
  int side, i;

  Lock(lock_his);
  for (side = black; side <= white; side++)
    for (i = 0; i < 512; i++)
      history[side][i] = history[side][i] / 256;
  Unlock(lock_his);
}
