#include "chess.h"
#include "data.h"
#include "epdglue.h"
#if defined(UNIX) || defined(AMIGA)
#  include <unistd.h>
#  include <sys/types.h>
#endif
#if defined(_WIN32) || defined(_WIN64)
#  include <stdint.h>
#endif
/* last modified 12/15/21 */
/*
 *******************************************************************************
 *                                                                             *
 *   Iterate() is the routine used to drive the iterated search.               *
 *   it repeatedly calls search, incrementing the search depth after each      *
 *   call, until either time is exhausted or the maximum set search depth is   *
 *   reached.                                                                  *
 *                                                                             *
 *******************************************************************************
 */
int Iterate(int wtm, int search_type, int root_list_done)
{
  ROOT_MOVE temp_rm;
  int i;
  register int value = 0, twtm;
  register int correct, correct_count, material = 0;
  char *fh_indicator, *fl_indicator;
  TREE *const tree = block[0];
  int savevalue = 0, bm_changes;
  unsigned int savetime = 0;
  PATH savepv;
  int print_ok = 0;
  int aspiration = aspiration_window;
  int fail_highs, fail_lows;

  if (wtm) {
    fh_indicator = "++";
    fl_indicator = "--";
  }
  else {
    fh_indicator = "--";
    fl_indicator = "++";
  }
  if (log_trace) {
    if (log_file) {
      Lock(lock_io);
      time_t t = time(NULL);
      struct tm tm = *localtime(&t);
      fprintf(log_file, "%02d/%02d/%02d %02d:%02d:%02d starting iterative search...\n",
        tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900,
        tm.tm_hour, tm.tm_min, tm.tm_sec);
      fflush(log_file);
      Unlock(lock_io);
    }
  }
/*
 ************************************************************
 *                                                          *
 *  initialize.                                             *
 *                                                          *
 *  produce the root move list, which is ordered and kept   *
 *  for the duration of this search (the order may change   *
 *  as new best moves are backed up to the root of course.) *
 *                                                          *
 ************************************************************
 */
  tree->curmv[0] = NULL_MOVE;
  if (wtm) {
    draw_score[0] = -abs_draw_score;
    draw_score[1] = abs_draw_score;
  } else {
    draw_score[0] = abs_draw_score;
    draw_score[1] = -abs_draw_score;
  }
  time_abort = 0;
  abort_search = 0;
  book_move = 0;
  program_start_time = ReadClock();
  start_time = ReadClock();
  program_start_cpu_time = ReadCPUClock();
  root_wtm = wtm;
  PreEvaluate(tree);
  kibitz_iteration_depth_last = kibitz_iteration_depth;
  kibitz_maximum_ply_last = kibitz_maximum_ply;
  kibitz_maximum_ply_last_last = kibitz_maximum_ply_last;
  kibitz_iteration_depth = 0;
  kibitz_maximum_ply = 0;
  tree->nodes_searched = 0;
  tree->quiescent_nodes_searched = 0;
  tree->fail_high = 0;
  tree->fail_high_first = 0;
  tree->evaluations = 0;
  parallel_splits = 0;
  parallel_aborts = 0;
  max_split_blocks = 0;
  if (booking || !Book(tree, wtm, root_list_done))
    do {
      if (abort_search)
        break;
      if (!root_list_done)
        RootMoveList(wtm);
#if !defined(NOEGTB)
      if (EGTB_draw && !puzzling && swindle_mode)
        EGTB_use = 0;
      else
        EGTB_use = EGTBlimit;
      if (EGTBlimit && !EGTB_use)
        Print(128, "Drawn at root, trying for swindle.\n");
#endif
      if (test_running) {
        correct_count = 0;
      }
      burp = 15 * 100;
      transposition_age = (transposition_age + 1) & 7;
      time_check = 0;
      tree->egtb_probes = 0;
      tree->egtb_probes_successful = 0;
      tree->extensions_done = 0;
      tree->qchecks_done = 0;
      tree->reductions_done = 0;
      HistoryAge();
      root_wtm = wtm;
/*
 ************************************************************
 *                                                          *
 *  if there are no legal moves, it is either mate or draw  *
 *  depending on whether the side to move is in check or    *
 *  not (mate or stalemate.)                                *
 *                                                          *
 ************************************************************
 */
      Lock(lock_root);
      if (n_root_moves == 0) {
        int value;
        Unlock(lock_root);
        program_end_cpu_time = ReadCPUClock();
        program_end_time = ReadClock();
        tree->pv[0].pathl = 0;
        tree->pv[0].pathd = 0;
        Lock(lock_root);
        if (Check(wtm))
          root_value = -(MATE - 1);
        else
          root_value = DrawScore(wtm);
        if (!puzzling) {
          last_root_value = root_value;
        }
        value = root_value;
        Unlock(lock_root);
        Print(6, "              depth   time  score   variation\n");
        Print(6, "                                    (no moves)\n");
        tree->nodes_searched = 1;
        return (value);
      } else
        Unlock(lock_root);
      MaterialRoot = Material;

/*
 ************************************************************
 *                                                          *
 *   now set the search time and iteratively call Search()  *
 *   to analyze the position deeper and deeper.  note that  *
 *   Search() is called with an alpha/beta window roughly   *
 *   1/3 of a pawn on either side of the score last         *
 *   returned by search.  also, after the first root move   *
 *   is searched, this window is collapsed to n and n+1     *
 *   (where n is the value for the first root move.)  often *
 *   a better move is found, which causes search to return  *
 *   <beta> as the score.  we then relax beta depending on  *
 *   its value:  if beta = alpha+1, set beta to alpha+1/3   *
 *   of a pawn;  if beta = alpha+1/3 pawn, then set beta to *
 *   + infinity.                                            *
 *                                                          *
 ************************************************************
 */
      TimeSet(tree, search_type);
      iteration_depth = 1;
      if (last_pv.pathd > 1)
        iteration_depth = last_pv.pathd + 1;
      else
        difficulty = 100;
      Print(6, "              depth   time  score   variation (%d)\n",
          iteration_depth);
      time_abort = 0;
      abort_search = 0;
/*
 ************************************************************
 *                                                          *
 *   set the initial search bounds based on the last search *
 *   or default values.                                     *
 *                                                          *
 *   "Aspiration windows are a way to reduce the search     *
 *   space in an alpha-beta search. The technique is to use *
 *   a guess of the expected value (usually from the last   *
 *   iteration in iterative deepening), and use a window    *
 *   around this as the alpha-beta bounds. Because the      *
 *   window is narrower, more beta cutoffs are achieved,    *
 *   and the search takes a shorter time. The drawback is   *
 *   that if the true score is outside this window, then a  *
 *   costly re-search must be made. Typical window sizes    *
 *   are 1/2 to 1/4 of a pawn on either side of the guess." *
 *                                                          *
 ************************************************************
 */
      tree->pv[0] = last_pv;
      Lock(lock_root);
      if (iteration_depth > 1) {
        root_alpha = Max(last_value - aspiration, -INFTY);
        root_beta = Min(last_value + aspiration, INFTY);
      } else {
        root_alpha = -INFTY;
        root_beta = INFTY;
      }
      root_value = root_alpha;
      value = root_value;
      Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *   if we are using multiple threads, and they have not    *
 *   been started yet, then start them now as the search    *
 *   is ready to begin.                                     *
 *                                                          *
 ************************************************************
 */
#if (CPUS > 1)
#  if defined(_WIN32) || defined(_WIN64)
      intptr_t proc;
#  else
      long proc;
      pthread_t pt;
#  endif
      Lock(lock_smp);
      if (smp_max_threads > smp_threads + 1) {
        initialized_threads = 1;
        if (smp_threads < 0)
          smp_threads = 0;
        Unlock(lock_smp);
        for (proc = smp_threads + 1; proc < smp_max_threads; proc++) {
          Print(128, "starting thread %d\n", proc);
          thread[proc] = 0;
#  if defined(_WIN32) || defined(_WIN64)
          NumaStartThread(ThreadInit, (void *) proc);
#  else
          pthread_create(&pt, &attributes, ThreadInit, (void *) proc);
#  endif
          Lock(lock_smp);
          smp_threads++;
          Unlock(lock_smp);
          Print(128, " <done>\n");
        }
      } else
        Unlock(lock_smp);
      WaitForAllThreadsInitialized();
#endif
      Lock(lock_root);
      root_print_ok = 0;
      Unlock(lock_root);
      for (; iteration_depth <= MAXPLY - MAXPLY_ULLAGE; iteration_depth++) { /* do iterate */
        maximum_ply = 0;
        fail_highs = fail_lows = 0;
/*
 ************************************************************
 *                                                          *
 *   now install the old PV into the hash table so that     *
 *   these moves will be followed first.                    *
 *                                                          *
 ************************************************************
 */
        twtm = wtm;
        for (i = 1; i < (int)tree->pv[0].pathl; i++) {
          if (!VerifyMove(tree, i, twtm, tree->pv[0].path[i])) {
            Print(4095, "ERROR, not installing bogus pv info at ply=%d\n", i);
            Print(4095, "not installing from=%d  to=%d  piece=%d\n",
                From(tree->pv[0].path[i]), To(tree->pv[0].path[i]),
                Piece(tree->pv[0].path[i]));
            Print(4095, "pathlen=%d\n", tree->pv[0].pathl);
            break;
          }
          HashStorePV(tree, twtm, i);
          MakeMove(tree, i, tree->pv[0].path[i], twtm);
          twtm = Flip(twtm);
        }
        for (i--; i > 0; i--) {
          twtm = Flip(twtm);
          UnmakeMove(tree, i, tree->pv[0].path[i], twtm);
        }
/*
 ************************************************************
 *                                                          *
 *  Now we call Search() and start the next search          *
 *  iteration.  We already have solid alpha/beta bounds set *
 *  up for the aspiration search.  When each iteration      *
 *  completes, these aspiration values are recomputed and   *
 *  used for the next iteration.                            *
 *                                                          *
 *  We need to set "nodes_between_time_checks" to a value   *
 *  that will force us to check the time reasonably often   *
 *  without wasting excessive time doing this check.  As    *
 *  the target time limit gets shorter, we shorten the      *
 *  interval between time checks to avoid burning time off  *
 *  of the clock unnecessarily.                             *
 *                                                          *
 ************************************************************
 */
        if (trace_level) {
          printf("==================================\n");
          printf("=      search iteration %2d       =\n", iteration_depth);
          printf("==================================\n");
        }
        time_check = 0;
/*
 ************************************************************
 *                                                          *
 *  This loop will execute until we either run out of time  *
 *  or complete this iteration.  Since we can return to     *
 *  Iterate() multiple times during this iteration, due to  *
 *  multiple fail highs (and perhaps even an initial fail   *
 *  low) we stick in this loop until we have completed all  *
 *  root moves or TimeCheck() tells us it is time to stop.  *
 *                                                          *
 ************************************************************
 */
        while (1) { /* do search root */
          root_moves[0].status &= ~1;
          root_moves[0].status &= ~2;
          thread[0] = block[0];
          if (log_trace) {
            if (log_file) {
              Lock(lock_io);
              time_t t = time(NULL);
              struct tm tm = *localtime(&t);
              fprintf(log_file, "%02d/%02d/%02d %02d:%02d:%02d starting root search...\n",
                tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900,
                tm.tm_hour, tm.tm_min, tm.tm_sec);
              fflush(log_file);
              Unlock(lock_io);
            }
          }
          value = SearchRoot(tree, root_alpha, root_beta, wtm, iteration_depth);
          if (log_trace) {
            if (log_file) {
              Lock(lock_io);
              time_t t = time(NULL);
              struct tm tm = *localtime(&t);
              fprintf(log_file, "%02d/%02d/%02d %02d:%02d:%02d root search ended\n",
                tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900,
                tm.tm_hour, tm.tm_min, tm.tm_sec);
              fflush(log_file);
              Unlock(lock_io);
            }
          }
          if (abort_search || time_abort)
            break;
          end_time = ReadClock();
          Lock(lock_root);
          root_print_ok = (end_time - start_time > noise_level) ? 1 : 0;
          Unlock(lock_root);
          if (TimeCheck(tree, 0))
            break;
/*
 ************************************************************
 *                                                          *
 *   Check for the case where we get a score back that is   *
 *   greater than or equal to beta.  This is called a fail  *
 *   high condition and requires a re-search with a better  *
 *   (more optimistic) beta value so that we can discover   *
 *   just how good this move really is.                     *
 *                                                          *
 ************************************************************
 */
          Lock(lock_root);
          if (value >= root_beta) {
            if (!(root_moves[0].status & 1)) {
              fail_highs++;
              fail_lows = 0;
              root_moves[0].status |= 2;
              root_value = root_alpha;
              root_beta = SetRootBeta(root_beta, fail_highs);
              root_moves[0].nodes = 0;
              Unlock(lock_root);
              if (!time_abort && !abort_search) {
                if (root_print_ok) {
                  Print(2, "               %2i   %s     %2s   ", iteration_depth,
                      DisplayTime(end_time - start_time), fh_indicator);
                  if (display_options & 64)
                    Print(2, "%d. ", move_number);
                  if ((display_options & 64) && !wtm)
                    Print(2, "... ");
                  Print(2, "%s!\n", OutputMove(tree, tree->pv[1].path[1], 1, wtm));
                }
              }
            } else
            Unlock(lock_root);
          } else
            Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *   Check for the case where we get a score back that is   *
 *   less than or equal to alpha.  This is called a fail    *
 *   low condition and requires a re-search with a better   *
 *   (more optimistic) alpha value so that we can discover  *
 *   just how good this move really is.                     *
 *                                                          *
 ************************************************************
 */
          Lock(lock_root);
          if (value <= root_alpha) {
            if (!(root_moves[0].status & 2)) {
              fail_lows++;
              fail_highs = 0;
              root_moves[0].status |= 1;
              root_alpha = SetRootAlpha(root_alpha, fail_lows);
              root_value = root_alpha;
              root_moves[0].nodes = 0;
              Unlock(lock_root);
              if (!time_abort && !abort_search) {
                if (root_print_ok) {
                  Print(2, "               %2i   %s     %2s   ", iteration_depth,
                      DisplayTime(end_time - start_time), fl_indicator);
                  if (display_options & 64)
                    Print(2, "%d. ", move_number);
                  if ((display_options & 64) && !wtm)
                    Print(2, "... ");
                  Print(2, "%s?\n", OutputMove(tree, root_moves[0].move, 1, wtm));
                }
              }
            } else
              Unlock(lock_root);
          } else
            Unlock(lock_root);
          if (!(root_moves[0].status & 1) && !(root_moves[0].status & 2))
            break; /* not failed low and not failed high */
          root_moves[0].status &= 255 - 16;
        } /* end do search root */
        Lock(lock_root);
        if (root_value > root_alpha && root_value < root_beta)
          last_root_value = root_value;
        Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *   if we are running a test suite, check to see if we can *
 *   exit the search.  this happens when N successive       *
 *   iterations produce the correct solution.  N is set by  *
 *   the test command in Option().                          *
 *                                                          *
 ************************************************************
 */
        if (test_running) {
          correct = solution_type;
          for (i = 0; i < number_of_solutions; i++) {
            if (!solution_type) {
              if (solutions[i] == tree->pv[0].path[1])
                correct = 1;
            } else if (solutions[i] == tree->pv[0].path[1])
              correct = 0;
          }
          if (correct)
            correct_count++;
          else
            correct_count = 0;
        }
        twtm = wtm;
/*
 ************************************************************
 *                                                          *
 *   A quick kludge to make certain the PV move comes first *
 *   in the move list.  On rare occasions, multiple moves   *
 *   can fail high on the null-window search at the root,   *
 *   and some of them will then fail low on the re-search.  *
 *   The last move to fail high ends up on the top of the   *
 *   move list, while a different move shows up as best by  *
 *   not failing low on the re-search.  This makes certain  *
 *   that the PV move from the previous iteration is always *
 *   searched first.                                        *
 *                                                          *
 ************************************************************
 */
        Lock(lock_root);
        if (tree->pv[0].path[1] != root_moves[0].move) {
          for (i = 0; i < n_root_moves; i++)
            if (tree->pv[0].path[1] == root_moves[i].move)
              break;
          if (i < n_root_moves) {
            temp_rm = root_moves[i];
            for (; i > 0; i--)
              root_moves[i] = root_moves[i - 1];
            root_moves[0] = temp_rm;
          }
        }
        Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *  Notice that we don't search moves that were best over   *
 *  the last 3 iterations in parallel, nor do we reduce     *
 *  those since they are potential best moves again.        *
 *                                                          *
 ************************************************************
 */
        Lock(lock_root);
        for (i = 0; i < n_root_moves; i++) {
          root_moves[i].status &= 3;
          if (root_moves[i].bm_age)
            root_moves[i].bm_age--;
          if (root_moves[i].bm_age)
            root_moves[i].status |= 4;
        }
        Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *   if requested, print the ply=1 move list along with the *
 *   node counts for the tree each move produced.           *
 *                                                          *
 ************************************************************
 */
        if (display_options & 256) {
          UINT64 total_nodes = 0;
          Print(128, "       move       nodes     R/P\n");
          Lock(lock_root);
          for (i = 0; i < n_root_moves; i++) {
            total_nodes += root_moves[i].nodes;
            Unlock(lock_root);
            Print(128, " %10s  " BMF10 "     %d %d\n", OutputMove(tree,
                    root_moves[i].move, 1, wtm), root_moves[i].nodes,
                (root_moves[i].status & 8) == 0,
                (root_moves[i].status & 4) == 0);
            Lock(lock_root);
          }
          Unlock(lock_root);
          Print(256, "      total  " BMF10 "\n", total_nodes);
        }
/*
 ************************************************************
 *                                                          *
 *   now adjust the "difficulty" value.  The idea here is   *
 *   that a position is easier if the best move does not    *
 *   change during an iteration, and harder if it does.     *
 *   We use this to adjust the target time during the       *
 *   search to spend more or less time when appropriate.    *
 *                                                          *
 *   if the PV did not change, we lower the difficulty      *
 *   setting, but we weigh the past difficulty more, so     *
 *   it takes several iterations with no change to reach    *
 *   the max easy setting.  Ditto for cases where there     *
 *   were changes, except that the more changes in the last *
 *   iteration, the more difficult we consider this search. *
 *                                                          *
 ************************************************************
 */
        bm_changes = 0;
        Lock(lock_root);
        for (i = 0; i < n_root_moves; i++)
          if (root_moves[i].bm_age == 3)
            bm_changes++;
        if (bm_changes <= 1)
          difficulty = 90 * difficulty / 100;
        else {
          if (difficulty < 100)
            difficulty = 100 + 20 * bm_changes;
          else
            difficulty = 80 * difficulty / 100 + 20 * bm_changes;
        }
        difficulty = Max(60, Min(difficulty, 200));
        for (i = 0; i < n_root_moves; i++)
          root_moves[i].nodes = 0;
        if (end_time - start_time > 10)
          nodes_per_second = (unsigned int)
              (((UINT64) (tree->nodes_searched) * 100 /
              (UINT64) (end_time - start_time)));
        else
          nodes_per_second = 1000000;
        if (test_running) { /* test */
          if (value <= test_score_threshold) {
            correct_count = 0;
          }
          if (!time_abort && !abort_search && (root_print_ok ||
                  correct_count >= early_exit || Abs(value) > NOT_MATE ||
                  tree->pv[0].pathh == 2)) {
            if (Abs(value) != (MATE - 1)) {
              Unlock(lock_root);
              DisplayPV(tree, 5, wtm, end_time - start_time, value, &tree->pv[0]);
              Lock(lock_root);
            }
          }
        } else { /* not test */
/*
 ************************************************************
 *                                                          *
 *   Vafra may display here time, score and PV for the      *
 *   current finished iteration in SearchRoot().            *
 *                                                          *
 ************************************************************
 */
          if (!time_abort && !abort_search) {
            if (root_print_ok) {
              Unlock(lock_root);
              DisplayPV(tree, 5, wtm, end_time - start_time, value, &tree->pv[0]);
              Lock(lock_root);
              print_ok = 0;
            } else {
              savevalue = value;
              savepv = tree->pv[0];
              savetime = end_time - start_time;
              print_ok = 1;
            }
          }
        }
        if (!time_abort && !abort_search) {
          root_alpha = Max(value - aspiration, -INFTY);
          root_value = root_alpha;
          root_beta = Min(value + aspiration, INFTY);
        }
        Unlock(lock_root);
/*
 ************************************************************
 *                                                          *
 *  There are multiple termination criteria that are used.  *
 *  The first and most obvious is that we have exceeded the *
 *  target time limit.  Others include reaching a user-set  *
 *  maximum search depth, finding a mate and we searched so *
 *  deep there is little chance of another iteration find-  *
 *  ing a shorter mate; the search was aborted due to some  *
 *  sort of user input (usually during pondering);  and     *
 *  finally, when running a test suite, we had the correct  *
 *  best move for N successive iterations and the user      *
 *  asked us to stop after that number.                     *
 *                                                          *
 ************************************************************
 */
        end_time = ReadClock();
        if (TimeCheck(tree, 0))
          break;
        if ((iteration_depth >= search_depth) && search_depth)
          break;
        if (time_abort || abort_search)
          break;
        if (test_running) {
          if (correct_count >= early_exit)
            break;
        }
        if (search_nodes && tree->nodes_searched >= search_nodes)
          break;
      } /* end do iterate */
/*
 ************************************************************
 *                                                          *
 *   search done, now display statistics, depending on the  *
 *   verbosity level set.                                   *
 *                                                          *
 ************************************************************
 */
      end_time = ReadClock();
      program_end_cpu_time = ReadCPUClock();
      program_end_time = ReadClock();
/*
 ************************************************************
 *                                                          *
 *   Vafra may display here time, score and PV for the      *
 *   last finished iteration in SearchRoot().               *
 *                                                          *
 ************************************************************
 */
      if (print_ok) {
        root_print_ok = print_ok;
        DisplayPV(tree, 5, wtm, savetime, savevalue, &savepv);
      }
      tree->evaluations = (tree->evaluations) ? tree->evaluations : 1;
      if ((!abort_search || time_abort) && !puzzling) {
        tree->fail_high++;
        tree->fail_high_first++;
        material = Material / PieceValues(white, pawn);
        if (kibitz_iteration_depth == 0)
          kibitz_iteration_depth = kibitz_iteration_depth_last;
        if (kibitz_maximum_ply == 0)
          kibitz_maximum_ply = kibitz_maximum_ply_last;
        if (kibitz_maximum_ply <= kibitz_iteration_depth)
          kibitz_maximum_ply = kibitz_maximum_ply_last_last;
        if (kibitz_maximum_ply < kibitz_iteration_depth)
          kibitz_maximum_ply = kibitz_iteration_depth;
        Print(8, "              depth=%d/%d", kibitz_iteration_depth, kibitz_maximum_ply);
        Print(8, "  time=%s  mat=%d",
            DisplayTimeKibitz(end_time - start_time), material);
        Print(8, "  n=" BMF, tree->nodes_searched);
        Print(8, "  fh=%u%%",
            (int) ((UINT64) tree->fail_high_first * 100 /
                (UINT64) tree->fail_high));
        Print(8, "  nps=%s\n", DisplayKM(nodes_per_second));
        Print(16, "              extended=%s ",
            DisplayKM(tree->extensions_done));
        Print(16, "qchecks=%s ", DisplayKM(tree->qchecks_done));
        Print(16, "reduced=%s\n", DisplayKM(tree->reductions_done));
        Print(16, "              predicted=%d  evals=%s  50move=%d",
            predicted, DisplayKM(tree->evaluations), Rule50Moves(0));
        Print(16, "  EGTBprobes=%s  hits=%s\n", DisplayKM(tree->egtb_probes),
            DisplayKM(tree->egtb_probes_successful));
#if defined(WINDOWS)
        Print(16, "              SMP->  splits=%d  aborts=%d  data=%d/%d  cpu=%.1f%%\n",
            parallel_splits, parallel_aborts, max_split_blocks, MAX_BLOCKS, CPUPercent);
#else
        Print(16, "              SMP->  splits=%d  aborts=%d  data=%d/%d  \n",
            parallel_splits, parallel_aborts, max_split_blocks, MAX_BLOCKS);
#endif
      }
    } while (0);
  else {
    Lock(lock_root);
    last_root_value = 0;
    root_value = 0;
    Unlock(lock_root);
    book_move = 1;
    tree->pv[0] = tree->pv[1];
    if (analyze_mode)
      Kibitz(4, wtm, 0, 0, 0, 0, 0, kibitz_text);
  }
  if (smp_nice && ponder == 0 && smp_threads) {
    Print(128, "terminating SMP processes\n");
    quit = 1;
    while (smp_threads);
    quit = 0;
    Lock(lock_smp);
    smp_idle = 0;
    Unlock(lock_smp);
  }
  //program_end_time = ReadClock();
  search_move = 0;
  if (log_trace) {
    if (log_file) {
      Lock(lock_io);
      time_t t = time(NULL);
      struct tm tm = *localtime(&t);
#if defined(WINDOWS)
      fprintf(log_file, "%02d/%02d/%02d %02d:%02d:%02d iterative search ended (cpu=%.1f%%)\n",
        tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900,
        tm.tm_hour, tm.tm_min, tm.tm_sec, CPUPercent);
#else
      fprintf(log_file, "%02d/%02d/%02d %02d:%02d:%02d iterative search ended\n",
        tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900,
        tm.tm_hour, tm.tm_min, tm.tm_sec);
#endif
      fflush(log_file);
      Unlock(lock_io);
    }
  }
  return (last_root_value);
}

/*
 *******************************************************************************
 *                                                                             *
 *   SetRootAlpha() is used to set the root alpha value by looking at how many *
 *   times this move has failed low.                                           *
 *                                                                             *
 *******************************************************************************
 */
int SetRootAlpha(int old_root_alpha, int fail_lows)
{
  int alpha = -INFTY;
  int i;

  if (fail_lows <= 0)
    return (old_root_alpha);
  if (fail_lows > ITERATE_FAIL_DELTA)
    i = ITERATE_FAIL_DELTA - 1;
  else
    i = fail_lows - 1;

  if (iterate_fail_delta[i] <= queen_value)
    alpha = Max(old_root_alpha - iterate_fail_delta[i], -INFTY);
  else
    alpha = -INFTY;

  return (alpha);
}

/*
 *******************************************************************************
 *                                                                             *
 *   SetRootBeta() is used to set the root beta value by looking at how many   *
 *   times this move has failed high.                                          *
 *                                                                             *
 *******************************************************************************
 */
int SetRootBeta(int old_root_beta, int fail_highs)
{
  int beta = INFTY;
  int i;

  if (fail_highs <= 0)
    return (old_root_beta);
  if (fail_highs > ITERATE_FAIL_DELTA)
    i = ITERATE_FAIL_DELTA - 1;
  else
    i = fail_highs - 1;

  if (iterate_fail_delta[i] <= queen_value)
    beta = Min(old_root_beta + iterate_fail_delta[i], INFTY);
  else
    beta = INFTY;

  return (beta);
}



