/* last modified 12/07/21 */
#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include "nnue.h"

/*
 *******************************************************************************
 *                                                                             *
 *   The neural network consists of four layers. The input layer is heavily    *
 *   overparametrized, feeding in the board representation for all king        *
 *   placements per side.                                                      *
 *                                                                             *
 *   The so called HalfKP structure consists of two halves covering input      *
 *   layer and first hidden layer, each half of the input layer associated to  *
 *   one of the two kings, cross coupled with the side to move or not to move  *
 *   halves of the first hidden layer. For each either black or white king     *
 *   placement, the 10 none king pieces on their particular squares are the    *
 *   boolean {0,1} inputs, along with a relict from Shogi piece drop           *
 *   (BONA_PIECE_ZERO), 64 x (64 x 10 + 1) = 41,024 inputs for each half,      *
 *   which are multiplied by a 16-bit integer weight vector for 256 outputs    *
 *   per half, in total, 256 x 41,024 = 10,502,144 weights. As emphasized by   *
 *   Ronald de Man in a CCC forum discussion, the input weights are arranged   *
 *   in such a way, that colour flipped king-piece configurations in both      *
 *   halves share the same index. However, and that seems also a relict from   *
 *   Shogi with its 180 degrees rotational 9x9 board symmetry, instead of      *
 *   vertical flipping (xor 56), rotation is applied (xor 63).                 *
 *                                                                             *
 *   The efficiency of NNUE is due to incremental update of the input layer    *
 *   outputs in make and unmake move, where only a tiny fraction of its        *
 *   neurons need to be considered in case of none king moves. The remaining   *
 *   three layers with 2x256x32, 32x32 and 32x1 weights are computational less *
 *   expensive, hidden layers apply a ReLu activation, best calculated using   *
 *   appropriate SIMD instructions performing fast 8-bit/16-bit integer vector *
 *   arithmetic, like MMX, SSE2 or AVX2 on x86/x86-64, or, if available,       *
 *   AVX-512.                                                                  *
 *                                                                             *
 *   https://www.chessprogramming.org/Stockfish_NNUE                           *
 *                                                                             *
 *******************************************************************************
 */

//--------------------
#ifdef _MSC_VER
#if defined(_WIN32) || defined(_WIN64)
#if defined(_WIN64)
// Vafra.w64.AVX2.3.0.0.exe, etc. (works on Intel i7-4770)
#ifdef _DOUSE_AVX2 // _WIN64
#  define USE_AVX2   1
#  define USE_SSE41  1
#  define USE_SSE3   1
#  define USE_SSE2   1
#  define USE_SSE    1
#  define IS_64BIT   1
#endif
// Vafra.w64.SSE41.3.0.0.exe, etc.
#ifdef _DOUSE_SSE41 // _WIN64
#  define USE_SSE41  1
#  define USE_SSE3   1
#  define USE_SSE2   1
#  define USE_SSE    1
#  define IS_64BIT   1
#endif
// Vafra.w64.SSE3.3.0.0.exe, etc. (may or may not work on Intel i7-2670QM)
#ifdef _DOUSE_SSE3 // _WIN64
#  define USE_SSE3   1
#  define USE_SSE2   1
#  define USE_SSE    1
#  define IS_64BIT   1
#endif
// Vafra.w64.SSE2.3.0.0.exe, etc. (should work on Intel i7-2670QM)
#ifdef _DOUSE_SSE2 // _WIN64
#  define USE_SSE2   1
#  define USE_SSE    1
#  define IS_64BIT   1
#endif
#else
// Vafra.w32.SSE2.3.0.0.exe, etc.
#ifdef _DOUSE_SSE2 // _WIN32
#  define USE_SSE2   1
#endif
#endif
#endif
#endif
//-------------------

#if defined(USE_AVX2)
#include <immintrin.h>

#elif defined(USE_SSE41)
#include <smmintrin.h>

#elif defined(USE_SSSE3)
#include <tmmintrin.h>

#elif defined(USE_SSE2)
#include <emmintrin.h>

#elif defined(USE_SSE)
#include <xmmintrin.h>

#endif

/**
* bsf and bsr
*/
#if defined(_WIN32) || defined(_WIN64)
#if !defined(_WIN64)
extern "C" unsigned char msb[65536];
extern "C" unsigned char lsb[65536];
#endif
#include "vcinline.h"
#define bsf(b) LSB(b)
#define bsr(b) MSB(b)
#endif

/**
* Position
*/
Position *POS;

/**
* NNUEdata
*/
#ifdef _NNUE_EVALUATE_INCREMENTAL
NNUEdata *NNUE;
#else
NNUEdata *NNUE;
#endif

//-------------------
#if defined(_WIN32) || defined(_WIN64)
#define INLINE  __forceinline
#else
#define INLINE  __inline
#endif

INLINE uint32_t nnue_readu_le_u32(const void *p)
{
  const uint8_t *q = (const uint8_t*) p;
  return q[0] | (q[1] << 8) | (q[2] << 16) | (q[3] << 24);
}

INLINE uint16_t nnue_readu_le_u16(const void *p)
{
  const uint8_t *q = (const uint8_t*) p;
  return q[0] | (q[1] << 8);
}

#if defined(_WIN32) || defined(_WIN64)
extern "C" HANDLE OpenNNUEFile(const char *);
extern "C" void CloseNNUEFile(HANDLE);
extern "C" __int64 GetNNUEFileSize(HANDLE);
extern "C" const void *MapNNUEFile(HANDLE, HANDLE *);
extern "C" void UnmapNNUEFile(const void *, HANDLE);
extern "C" void DecodeNNUEFen(const char *, int *, int* , int *, int *, int *, int *);
#endif

#define nnue_KING(c)    ( (c) ? bking : wking )
#define nnue_IS_KING(p) ( ((p) == wking) || ((p) == bking) )
//-------------------

#define nnue_B(x) (buf.x)

enum {
  PS_W_PAWN   =  1,
  PS_B_PAWN   =  1 * 64 + 1,
  PS_W_KNIGHT =  2 * 64 + 1,
  PS_B_KNIGHT =  3 * 64 + 1,
  PS_W_BISHOP =  4 * 64 + 1,
  PS_B_BISHOP =  5 * 64 + 1,
  PS_W_ROOK   =  6 * 64 + 1,
  PS_B_ROOK   =  7 * 64 + 1,
  PS_W_QUEEN  =  8 * 64 + 1,
  PS_B_QUEEN  =  9 * 64 + 1,
  PS_END      = 10 * 64 + 1  // 641
};

uint32_t PieceToIndex[2][14] = {
  { 0, 0, PS_W_QUEEN, PS_W_ROOK, PS_W_BISHOP, PS_W_KNIGHT, PS_W_PAWN,
       0, PS_B_QUEEN, PS_B_ROOK, PS_B_BISHOP, PS_B_KNIGHT, PS_B_PAWN, 0},
  { 0, 0, PS_B_QUEEN, PS_B_ROOK, PS_B_BISHOP, PS_B_KNIGHT, PS_B_PAWN,
       0, PS_W_QUEEN, PS_W_ROOK, PS_W_BISHOP, PS_W_KNIGHT, PS_W_PAWN, 0}
};

// Version of the evaluation file
// net architecture HalfKP 256x2-32-32-1
static const uint32_t NnueVersions[2] = {0x7af32f16u, 0x7af32f17u};
static int net_rotated = 0;
static int net_architecture = 0;

// Constants used in evaluation value calculation
enum {
  FV_SCALE = 16,
  SHIFT = 6
};

enum {
  K_HALF_DIMENSIONS = 256,
  FT_IN_DIMS = 64 * PS_END, // 64 * 641
  FT_OUT_DIMS = K_HALF_DIMENSIONS * 2
};

static_assert(K_HALF_DIMENSIONS % 256 == 0, "K_HALF_DIMENSIONS should be a multiple of 256");

#define VECTOR

#ifdef USE_AVX512
#define SIMD_WIDTH 512
typedef __m512i vec16_t;
typedef __m512i vec8_t;
typedef __mmask64 mask_t;
#define vec_add_16(a,b) _mm512_add_epi16(a,b)
#define vec_sub_16(a,b) _mm512_sub_epi16(a,b)
#define vec_packs(a,b) _mm512_packs_epi16(a,b)
#define vec_mask_pos(a) _mm512_cmpgt_epi8_mask(a,_mm512_setzero_si512())
#define NUM_REGS 8 // only 8 are needed

#elif USE_AVX2
#define SIMD_WIDTH 256
typedef __m256i vec16_t;
typedef __m256i vec8_t;
typedef uint32_t mask_t;
#define vec_add_16(a,b) _mm256_add_epi16(a,b)
#define vec_sub_16(a,b) _mm256_sub_epi16(a,b)
#define vec_packs(a,b) _mm256_packs_epi16(a,b)
#define vec_mask_pos(a) _mm256_movemask_epi8(_mm256_cmpgt_epi8(a,_mm256_setzero_si256()))
#define NUM_REGS 16

#elif USE_SSE2
#define SIMD_WIDTH 128
typedef __m128i vec16_t;
typedef __m128i vec8_t;
typedef uint16_t mask_t;
#define vec_add_16(a,b) _mm_add_epi16(a,b)
#define vec_sub_16(a,b) _mm_sub_epi16(a,b)
#define vec_packs(a,b) _mm_packs_epi16(a,b)
#define vec_mask_pos(a) _mm_movemask_epi8(_mm_cmpgt_epi8(a,_mm_setzero_si128()))
#ifdef IS_64BIT
#define NUM_REGS 16
#else
#define NUM_REGS 8
#endif

#else
#undef VECTOR
#define SIMD_WIDTH 16 // dummy
typedef uint8_t mask_t; // dummy

#endif

#ifdef IS_64BIT
typedef uint64_t mask2_t;
#else
typedef uint32_t mask2_t;
#endif

typedef int8_t clipped_t;
#if defined(USE_SSE2) && !defined(USE_AVX2)
typedef int16_t weight_t;
#else
typedef int8_t weight_t;
#endif

typedef struct {
  size_t size;
  unsigned values[30];
} IndexList;

INLINE int nnue_orient(int c, int s)
{
  if (net_rotated)
    return s ^ (c == nwhite ? 0x00 : 0x3f);
  else
    return s ^ (c == nwhite ? 0x00 : 0x38);
}

INLINE unsigned nnue_make_index(int c, int s, int pc, int ksq)
{
  return nnue_orient(c, s) + PieceToIndex[c][pc] + PS_END * ksq;
}

static void nnue_half_kp_append_active_indices(const Position *pos, const int c,
    IndexList *active)
{
  int ksq = pos->square[c];
  ksq = nnue_orient(c, ksq);
  for (int i = 2; pos->piece[i]; i++) {
    int sq = pos->square[i];
    int pc = pos->piece[i];
    active->values[active->size++] = nnue_make_index(c, sq, pc, ksq);
  }
}

static void nnue_half_kp_append_changed_indices(const Position *pos, const int c,
    const DirtyPiece *dp, IndexList *removed, IndexList *added)
{
  int ksq = pos->square[c];
  ksq = nnue_orient(c, ksq);
  for (int i = 0; i < dp->dirtyNum; i++) {
    int pc = dp->pc[i];
    if (nnue_IS_KING(pc)) continue;
    if (dp->from[i] != 64)
      removed->values[removed->size++] = nnue_make_index(c, dp->from[i], pc, ksq);
    if (dp->to[i] != 64)
      added->values[added->size++] = nnue_make_index(c, dp->to[i], pc, ksq);
  }
}

static void nnue_append_active_indices(const Position *pos, IndexList active[2])
{
  for (unsigned c = 0; c < 2; c++)
    nnue_half_kp_append_active_indices(pos, c, &active[c]);
}

#ifdef _NNUE_EVALUATE_INCREMENTAL
static void nnue_append_changed_indices(const Position *pos, IndexList removed[2],
    IndexList added[2], bool reset[2])
{
  const DirtyPiece *dp = &(pos->nnue[0]->dirtyPiece);
  // assert(dp->dirtyNum != 0);

  if (pos->nnue[1]->accumulator.computedAccumulation) {
    for (unsigned c = 0; c < 2; c++) {
      reset[c] = dp->pc[0] == (int)nnue_KING(c);
      if (reset[c])
        nnue_half_kp_append_active_indices(pos, c, &added[c]);
      else
        nnue_half_kp_append_changed_indices(pos, c, dp, &removed[c], &added[c]);
    }
  } else {
    const DirtyPiece *dp2 = &(pos->nnue[1]->dirtyPiece);
    for (unsigned c = 0; c < 2; c++) {
      reset[c] =   dp->pc[0] == (int)nnue_KING(c)
                || dp2->pc[0] == (int)nnue_KING(c);
      if (reset[c])
        nnue_half_kp_append_active_indices(pos, c, &added[c]);
      else {
        nnue_half_kp_append_changed_indices(pos, c, dp, &removed[c], &added[c]);
        nnue_half_kp_append_changed_indices(pos, c, dp2, &removed[c], &added[c]);
      }
    }
  }
}
#endif

// InputLayer = InputSlice<256 * 2>
// out: 512 x clipped_t

// Hidden1Layer = ClippedReLu<AffineTransform<InputLayer, 32>>
// 512 x clipped_t -> 32 x int32_t -> 32 x clipped_t

// Hidden2Layer = ClippedReLu<AffineTransform<hidden1, 32>>
// 32 x clipped_t -> 32 x int32_t -> 32 x clipped_t

// OutputLayer = AffineTransform<HiddenLayer2, 1>
// 32 x clipped_t -> 1 x int32_t

#if !defined(USE_AVX512)
static weight_t hidden1_weights alignas(64) [32 * 512];
static weight_t hidden2_weights alignas(64) [32 * 32];
#else
static weight_t hidden1_weights alignas(64) [64 * 512];
static weight_t hidden2_weights alignas(64) [64 * 32];
#endif
static weight_t output_weights alignas(64) [1 * 32];

static int32_t hidden1_biases alignas(64) [32];
static int32_t hidden2_biases alignas(64) [32];
static int32_t output_biases[1];

static_assert(FT_OUT_DIMS % 64 == 0, "FT_OUT_DIMS not a multiple of 64");

#ifdef VECTOR
INLINE bool nnue_next_idx(unsigned *idx, unsigned *offset, mask2_t *v,
    mask_t *mask, unsigned inDims)
{
  while (*v == 0) {
    *offset += 8 * sizeof(mask2_t);
    if (*offset >= inDims) return false;
    memcpy(v, (char *)mask + (*offset / 8), sizeof(mask2_t));
  }
#ifdef IS_64BIT
  *idx = *offset + bsf(*v);
#else
  *idx = *offset + bsf(*v);
#endif
  *v &= *v - 1;
  return true;
}

#endif

// Input feature converter
static int16_t ft_biases alignas(64) [K_HALF_DIMENSIONS];
static int16_t ft_weights alignas(64) [K_HALF_DIMENSIONS * FT_IN_DIMS];

#ifdef VECTOR
#define TILE_HEIGHT (NUM_REGS * SIMD_WIDTH / 16)
#endif

// Calculate cumulative value without using difference calculation
INLINE void nnue_refresh_accumulator(Position *pos)
{
  Accumulator *accumulator = &(pos->nnue[0]->accumulator);

  IndexList activeIndices[2];
  activeIndices[0].size = activeIndices[1].size = 0;
  nnue_append_active_indices(pos, activeIndices);

  for (unsigned c = 0; c < 2; c++) {
#ifdef VECTOR
    for (unsigned i = 0; i < K_HALF_DIMENSIONS / TILE_HEIGHT; i++) {
      vec16_t *ft_biases_tile = (vec16_t *)&ft_biases[i * TILE_HEIGHT];
      vec16_t *accTile = (vec16_t *)&accumulator->accumulation[c][i * TILE_HEIGHT];
      vec16_t acc[NUM_REGS];

      for (unsigned j = 0; j < NUM_REGS; j++)
        acc[j] = ft_biases_tile[j];

      for (size_t k = 0; k < activeIndices[c].size; k++) {
        unsigned index = activeIndices[c].values[k];
        unsigned offset = K_HALF_DIMENSIONS * index + i * TILE_HEIGHT;
        vec16_t *column = (vec16_t *)&ft_weights[offset];

        for (unsigned j = 0; j < NUM_REGS; j++)
          acc[j] = vec_add_16(acc[j], column[j]);
      }

      for (unsigned j = 0; j < NUM_REGS; j++)
        accTile[j] = acc[j];
    }
#else
    memcpy(accumulator->accumulation[c], ft_biases,
        K_HALF_DIMENSIONS * sizeof(int16_t));

    for (size_t k = 0; k < activeIndices[c].size; k++) {
      unsigned index = activeIndices[c].values[k];
      unsigned offset = K_HALF_DIMENSIONS * index;

      for (unsigned j = 0; j < K_HALF_DIMENSIONS; j++)
        accumulator->accumulation[c][j] += ft_weights[offset + j];
    }
#endif
  }

  accumulator->computedAccumulation = 1;
}

#ifdef _NNUE_EVALUATE_INCREMENTAL
// Calculate cumulative value using difference calculation if possible
INLINE bool nnue_update_accumulator(Position *pos)
{
  Accumulator *accumulator = &(pos->nnue[0]->accumulator);
  if (accumulator->computedAccumulation)
    return true;

  Accumulator *prevAcc;
  if (   (!pos->nnue[1] || !(prevAcc = &pos->nnue[1]->accumulator)->computedAccumulation)
      && (!pos->nnue[2] || !(prevAcc = &pos->nnue[2]->accumulator)->computedAccumulation) )
    return false;

  IndexList removed_indices[2], added_indices[2];
  removed_indices[0].size = removed_indices[1].size = 0;
  added_indices[0].size = added_indices[1].size = 0;
  bool reset[2];
  nnue_append_changed_indices(pos, removed_indices, added_indices, reset);

#ifdef VECTOR
  for (unsigned i = 0; i< K_HALF_DIMENSIONS / TILE_HEIGHT; i++) {
    for (unsigned c = 0; c < 2; c++) {
      vec16_t *accTile = (vec16_t *)&accumulator->accumulation[c][i * TILE_HEIGHT];
      vec16_t acc[NUM_REGS];

      if (reset[c]) {
        vec16_t *ft_b_tile = (vec16_t *)&ft_biases[i * TILE_HEIGHT];
        for (unsigned j = 0; j < NUM_REGS; j++)
          acc[j] = ft_b_tile[j];
      } else {
        vec16_t *prevAccTile = (vec16_t *)&prevAcc->accumulation[c][i * TILE_HEIGHT];
        for (unsigned j = 0; j < NUM_REGS; j++)
          acc[j] = prevAccTile[j];

        // Difference calculation for the deactivated features
        for (unsigned k = 0; k < removed_indices[c].size; k++) {
          unsigned index = removed_indices[c].values[k];
          const unsigned offset = K_HALF_DIMENSIONS * index + i * TILE_HEIGHT;

          vec16_t *column = (vec16_t *)&ft_weights[offset];
          for (unsigned j = 0; j < NUM_REGS; j++)
            acc[j] = vec_sub_16(acc[j], column[j]);
        }
      }

      // Difference calculation for the activated features
      for (unsigned k = 0; k < added_indices[c].size; k++) {
        unsigned index = added_indices[c].values[k];
        const unsigned offset = K_HALF_DIMENSIONS * index + i * TILE_HEIGHT;

        vec16_t *column = (vec16_t *)&ft_weights[offset];
        for (unsigned j = 0; j < NUM_REGS; j++)
          acc[j] = vec_add_16(acc[j], column[j]);
      }

      for (unsigned j = 0; j < NUM_REGS; j++)
        accTile[j] = acc[j];
    }
  }
#else
  for (unsigned c = 0; c < 2; c++) {
    if (reset[c]) {
      memcpy(accumulator->accumulation[c], ft_biases,
          K_HALF_DIMENSIONS * sizeof(int16_t));
    } else {
      memcpy(accumulator->accumulation[c], prevAcc->accumulation[c],
          K_HALF_DIMENSIONS * sizeof(int16_t));
      // Difference calculation for the deactivated features
      for (unsigned k = 0; k < removed_indices[c].size; k++) {
        unsigned index = removed_indices[c].values[k];
        const unsigned offset = K_HALF_DIMENSIONS * index;

        for (unsigned j = 0; j < K_HALF_DIMENSIONS; j++)
          accumulator->accumulation[c][j] -= ft_weights[offset + j];
      }
    }

    // Difference calculation for the activated features
    for (unsigned k = 0; k < added_indices[c].size; k++) {
      unsigned index = added_indices[c].values[k];
      const unsigned offset = K_HALF_DIMENSIONS * index;

      for (unsigned j = 0; j < K_HALF_DIMENSIONS; j++)
        accumulator->accumulation[c][j] += ft_weights[offset + j];
    }
  }
#endif

  accumulator->computedAccumulation = 1;
  return true;
}
#endif

struct NetData {
  alignas(64) clipped_t input[FT_OUT_DIMS];
  clipped_t hidden1_out[32];
#if defined(USE_SSE2) && !defined(USE_AVX2)
  int16_t hidden2_out[32];
#else
  int8_t hidden2_out[32];
#endif
};

// Calculate the output of the output layer
INLINE int32_t nnue_affine_propagate(clipped_t *input, int32_t *biases,
    weight_t *weights)
{
#if defined(USE_AVX2)
  __m256i *iv = (__m256i *)input;
  __m256i *row = (__m256i *)weights;
#if defined(USE_VNNI)
  __m256i prod = _mm256_dpbusd_epi32(_mm256_setzero_si256(), iv[0], row[0]);
#else
  __m256i prod = _mm256_maddubs_epi16(iv[0], row[0]);
  prod = _mm256_madd_epi16(prod, _mm256_set1_epi16(1));
#endif
  __m128i sum = _mm_add_epi32(
      _mm256_castsi256_si128(prod), _mm256_extracti128_si256(prod, 1));
  sum = _mm_add_epi32(sum, _mm_shuffle_epi32(sum, 0x1b));
  return _mm_cvtsi128_si32(sum) + _mm_extract_epi32(sum, 1) + biases[0];

#elif defined(USE_SSE2)
  __m128i *iv = (__m128i *)input;
  __m128i *row = (__m128i *)weights;
  __m128i p0 = _mm_madd_epi16(iv[0], row[0]);
  __m128i p1 = _mm_madd_epi16(iv[1], row[1]);
  __m128i p2 = _mm_madd_epi16(iv[2], row[2]);
  __m128i p3 = _mm_madd_epi16(iv[3], row[3]);
  __m128i sum = _mm_add_epi32(_mm_add_epi32(p0, p1), _mm_add_epi32(p2, p3));
  sum = _mm_add_epi32(sum, _mm_shuffle_epi32(sum, 0xb));
#if defined(USE_SSE41)
  return _mm_cvtsi128_si32(sum) + _mm_extract_epi32(sum, 1) + biases[0];
#else
  sum = _mm_add_epi32(sum, _mm_shuffle_epi32(sum, 0x1));
  return _mm_cvtsi128_si32(sum) + biases[0];
#endif

#else
  int32_t sum = biases[0];
  for (unsigned j = 0; j < 32; j++)
    sum += weights[j] * input[j];
  return sum;

#endif
}

// Calculate the output of the first or the second hidden layer
#if defined(USE_AVX512)
INLINE void nnue_affine_txfm(int8_t *input, void *output, unsigned inDims,
    unsigned outDims, const int32_t *biases, const weight_t *weights,
    mask_t *inMask, mask_t *outMask, const bool pack8_and_calc_mask)
{
  assert(outDims == 32);

  (void)outDims;
  const __m512i kZero = _mm512_setzero_si512();
  __m512i out_0 = ((__m512i *)biases)[0];
  __m512i out_1 = ((__m512i *)biases)[1];
  __m512i first, second;
  mask2_t v;
  unsigned idx;

  memcpy(&v, inMask, sizeof(mask2_t));
  for (unsigned offset = 0; offset < inDims;) {
    if (!nnue_next_idx(&idx, &offset, &v, inMask, inDims))
      break;
    first = ((__m512i *)weights)[idx];
    uint16_t factor = input[idx];
    if (nnue_next_idx(&idx, &offset, &v, inMask, inDims)) {
      second = ((__m512i *)weights)[idx];
      factor |= input[idx] << 8;
    } else {
      second = kZero;
    }
    __m512i mul = _mm512_set1_epi16(factor), prod, signs;
    prod = _mm512_maddubs_epi16(mul, _mm512_unpacklo_epi8(first, second));
    signs = _mm512_srai_epi16(prod, 15);
    out_0 = _mm512_add_epi32(out_0, _mm512_unpacklo_epi16(prod, signs));
    out_1 = _mm512_add_epi32(out_1, _mm512_unpackhi_epi16(prod, signs));
  }

  __m512i out16 = _mm512_srai_epi16(_mm512_packs_epi32(out_0, out_1), SHIFT);

  __m256i *outVec = (__m256i *)output;
  const __m256i kZero256 = _mm256_setzero_si256();
  outVec[0] = _mm256_packs_epi16(
      _mm512_castsi512_si256(out16),_mm512_extracti64x4_epi64(out16, 1));
  if (pack8_and_calc_mask)
    outMask[0] = (uint32_t)_mm256_movemask_epi8(_mm256_cmpgt_epi8(outVec[0], kZero256));
  else
    outVec[0] = _mm256_max_epi8(outVec[0], kZero256);
}
#elif defined(USE_AVX2)
INLINE void nnue_affine_txfm(int8_t *input, void *output, unsigned inDims,
    unsigned outDims, const int32_t *biases, const weight_t *weights,
    mask_t *inMask, mask_t *outMask, const bool pack8_and_calc_mask)
{
  assert(outDims == 32);

  (void)outDims;
  const __m256i kZero = _mm256_setzero_si256();
  __m256i out_0 = ((__m256i *)biases)[0];
  __m256i out_1 = ((__m256i *)biases)[1];
  __m256i out_2 = ((__m256i *)biases)[2];
  __m256i out_3 = ((__m256i *)biases)[3];
  __m256i first, second;
  mask2_t v;
  unsigned idx;

  memcpy(&v, inMask, sizeof(mask2_t));
  for (unsigned offset = 0; offset < inDims;) {
    if (!nnue_next_idx(&idx, &offset, &v, inMask, inDims))
      break;
    first = ((__m256i *)weights)[idx];
    uint16_t factor = input[idx];
    if (nnue_next_idx(&idx, &offset, &v, inMask, inDims)) {
      second = ((__m256i *)weights)[idx];
      factor |= input[idx] << 8;
    } else {
      second = kZero;
    }
    __m256i mul = _mm256_set1_epi16(factor), prod, signs;
    prod = _mm256_maddubs_epi16(mul, _mm256_unpacklo_epi8(first, second));
    signs = _mm256_cmpgt_epi16(kZero, prod);
    out_0 = _mm256_add_epi32(out_0, _mm256_unpacklo_epi16(prod, signs));
    out_1 = _mm256_add_epi32(out_1, _mm256_unpackhi_epi16(prod, signs));
    prod = _mm256_maddubs_epi16(mul, _mm256_unpackhi_epi8(first, second));
    signs = _mm256_cmpgt_epi16(kZero, prod);
    out_2 = _mm256_add_epi32(out_2, _mm256_unpacklo_epi16(prod, signs));
    out_3 = _mm256_add_epi32(out_3, _mm256_unpackhi_epi16(prod, signs));
  }

  __m256i out16_0 = _mm256_srai_epi16(_mm256_packs_epi32(out_0, out_1), SHIFT);
  __m256i out16_1 = _mm256_srai_epi16(_mm256_packs_epi32(out_2, out_3), SHIFT);

  __m256i *outVec = (__m256i *)output;
  outVec[0] = _mm256_packs_epi16(out16_0, out16_1);
  if (pack8_and_calc_mask)
    outMask[0] = _mm256_movemask_epi8(_mm256_cmpgt_epi8(outVec[0], kZero));
  else
    outVec[0] = _mm256_max_epi8(outVec[0], kZero);
}
#elif defined(USE_SSE2)
INLINE void nnue_affine_txfm(clipped_t *input, void *output, unsigned inDims,
    unsigned outDims, const int32_t *biases, const weight_t *weights,
    mask_t *inMask, mask_t *outMask, const bool pack8_and_calc_mask)
{
  assert(outDims == 32);

  const __m128i kZeros[4] = { 0 };
  __m128i out_0 = ((__m128i *)biases)[0];
  __m128i out_1 = ((__m128i *)biases)[1];
  __m128i out_2 = ((__m128i *)biases)[2];
  __m128i out_3 = ((__m128i *)biases)[3];
  __m128i out_4 = ((__m128i *)biases)[4];
  __m128i out_5 = ((__m128i *)biases)[5];
  __m128i out_6 = ((__m128i *)biases)[6];
  __m128i out_7 = ((__m128i *)biases)[7];
  const __m128i *first, *second;
  mask2_t v;
  unsigned idx;

  memcpy(&v, inMask, sizeof(mask2_t));
  for (unsigned offset = 0; offset < inDims;) {
    if (!nnue_next_idx(&idx, &offset, &v, inMask, inDims))
      break;
    first = (__m128i *)&weights[outDims * idx];
    uint32_t factor = input[idx];
    if (nnue_next_idx(&idx, &offset, &v, inMask, inDims)) {
      second = (__m128i *)&weights[outDims * idx];
      factor |= input[idx] << 16;
    } else {
      second = kZeros;
    }
    __m128i mul = _mm_set1_epi32(factor);
    out_0 = _mm_add_epi32(out_0, _mm_madd_epi16(mul, _mm_unpacklo_epi16(first[0],second[0])));
    out_1 = _mm_add_epi32(out_1, _mm_madd_epi16(mul, _mm_unpackhi_epi16(first[0],second[0])));
    out_2 = _mm_add_epi32(out_2, _mm_madd_epi16(mul, _mm_unpacklo_epi16(first[1],second[1])));
    out_3 = _mm_add_epi32(out_3, _mm_madd_epi16(mul, _mm_unpackhi_epi16(first[1],second[1])));
    out_4 = _mm_add_epi32(out_4, _mm_madd_epi16(mul, _mm_unpacklo_epi16(first[2],second[2])));
    out_5 = _mm_add_epi32(out_5, _mm_madd_epi16(mul, _mm_unpackhi_epi16(first[2],second[2])));
    out_6 = _mm_add_epi32(out_6, _mm_madd_epi16(mul, _mm_unpacklo_epi16(first[3],second[3])));
    out_7 = _mm_add_epi32(out_7, _mm_madd_epi16(mul, _mm_unpackhi_epi16(first[3],second[3])));
  }

  __m128i out16_0 = _mm_srai_epi16(_mm_packs_epi32(out_0, out_1), SHIFT);
  __m128i out16_1 = _mm_srai_epi16(_mm_packs_epi32(out_2, out_3), SHIFT);
  __m128i out16_2 = _mm_srai_epi16(_mm_packs_epi32(out_4, out_5), SHIFT);
  __m128i out16_3 = _mm_srai_epi16(_mm_packs_epi32(out_6, out_7), SHIFT);

  __m128i *outVec = (__m128i *)output;
  if (pack8_and_calc_mask) {
    outVec[0] = _mm_packs_epi16(out16_0, out16_1);
    outMask[0] = _mm_movemask_epi8(_mm_cmpgt_epi8(outVec[0], kZeros[0]));
    outVec[1] = _mm_packs_epi16(out16_2, out16_3);
    outMask[1] = _mm_movemask_epi8(_mm_cmpgt_epi8(outVec[1], kZeros[0]));
  } else {
    const __m128i kx07f = _mm_set1_epi16(127);
    outVec[0] = _mm_min_epi16(_mm_max_epi16(out16_0, kZeros[0]), kx07f);
    outVec[1] = _mm_min_epi16(_mm_max_epi16(out16_1, kZeros[0]), kx07f);
    outVec[2] = _mm_min_epi16(_mm_max_epi16(out16_2, kZeros[0]), kx07f);
    outVec[3] = _mm_min_epi16(_mm_max_epi16(out16_3, kZeros[0]), kx07f);
  }
}
#else /* generic fallback */
INLINE void nnue_affine_txfm(clipped_t *input, void *output, unsigned inDims,
    unsigned outDims, int32_t *biases, const weight_t *weights,
    mask_t *inMask, mask_t *outMask, const bool pack8_and_calc_mask)
{
  (void)inMask; (void)outMask; (void)pack8_and_calc_mask;

  int32_t tmp[outDims];

  for (unsigned i = 0; i < outDims; i++)
    tmp[i] = biases[i];

  for (unsigned idx = 0; idx < inDims; idx++)
    if (input[idx])
      for (unsigned i = 0; i < outDims; i++)
        tmp[i] += (int8_t)input[idx] * weights[outDims * idx + i];

  clipped_t *outVec = (clipped_t *)output;
  for (unsigned i = 0; i < outDims; i++)
    outVec[i] = clamp(tmp[i] >> SHIFT, 0, 127);
}
#endif

// Calculate the output of the input layer
INLINE void nnue_transform(Position *pos, clipped_t *output, mask_t *outMask)
{
#ifdef _NNUE_EVALUATE_INCREMENTAL
  if (!nnue_update_accumulator(pos))
    nnue_refresh_accumulator(pos);
#else
  nnue_refresh_accumulator(pos);
#endif

  int16_t (*accumulation)[2][256] = &pos->nnue[0]->accumulator.accumulation;
  (void)outMask; // avoid compiler warning

  const int perspectives[2] = { pos->player, !pos->player };
  for (unsigned p = 0; p < 2; p++) {
    const unsigned offset = K_HALF_DIMENSIONS * p;

#ifdef VECTOR
    const unsigned numChunks = (16 * K_HALF_DIMENSIONS) / SIMD_WIDTH;
    vec8_t *out = (vec8_t *)&output[offset];
    for (unsigned i = 0; i < numChunks / 2; i++) {
      vec16_t s0 = ((vec16_t *)(*accumulation)[perspectives[p]])[i * 2];
      vec16_t s1 = ((vec16_t *)(*accumulation)[perspectives[p]])[i * 2 + 1];
      out[i] = vec_packs(s0, s1);
      *outMask++ = vec_mask_pos(out[i]);
    }

#else
    for (unsigned i = 0; i < K_HALF_DIMENSIONS; i++) {
      int16_t sum = (*accumulation)[perspectives[p]][i];
      output[offset + i] = clamp(sum, 0, 127);
    }
#endif

  }
}

// Evaluation function
int nnue_evaluate_pos(Position *pos)
{
  int32_t out_value;
  alignas(8) mask_t input_mask[FT_OUT_DIMS / (8 * sizeof(mask_t))];
  alignas(8) mask_t hidden1_mask[8 / sizeof(mask_t)] = { 0 };
  struct NetData buf;

  /* The "transform" step of the NNUE evaluation forms a 512-element
     vector of 8-bit ints where the first half is formed from the
     256-element vector of the side to move and the second half is
     formed from the 256-element vector of the other side. In this
     step the 16-bit elements are clipped/clamped to a value from 0
     to 127. This is the output of the input layer.
   */
  nnue_transform(pos, nnue_B(input), input_mask);

  /* This 512-element vector of 8-bit ints is then multiplied by a
     32x512 matrix of 8-bit weights to get a 32-element vector of
     32-bit ints, to which a vector of 32-bit biases is added. The
     sum vector is divided by 64 and clipped/clamped to a 32-element
     vector of 8-bit ints from 0 to 127. This is the output of the
     first hidden layer.
   */
  nnue_affine_txfm(nnue_B(input), nnue_B(hidden1_out), FT_OUT_DIMS, 32,
      hidden1_biases, hidden1_weights, input_mask, hidden1_mask, true);

  /* The resulting 32-element vector of 8-bit ints is multiplied by
     a 32x32 matrix of 8-bit weights to get a 32-element vector of
     32-bit ints, to which another vector of 32-bit biases is added.
     These ints are again divided by 64 and clipped/clamped to 32
     8-bit ints from 0 to 127. This is the output of the second
     hidden layer.
   */
  nnue_affine_txfm(nnue_B(hidden1_out), nnue_B(hidden2_out), 32, 32,
      hidden2_biases, hidden2_weights, hidden1_mask, NULL, false);

  /* This 32-element vector of 8-bits ints is then multiplied by a
     1x32 matrix of 8-bit weights (i.e. the inner product of two
     vectors is taken). This produces a 32-bit value to which a
     32-bit bias is added. This gives the output of the output layer.
   */
  out_value = nnue_affine_propagate((int8_t *)nnue_B(hidden2_out),
      output_biases, output_weights);

  /* The output of the output layer is divided by FV_SCALE = 16 to
     produce the NNUE evaluation. SF's evaluation then take some
     further steps such as adding a Tempo bonus (even though the
     NNUE evaluation inherently already takes into account the side
     to move in the "transform" step) and scaling the evaluation
     towards zero as rule50_count() approaches 50 moves.
   */
  return out_value / FV_SCALE;
}

static void nnue_read_output_weights(weight_t *w, const char *d)
{
  for (unsigned i = 0; i < 32; i++) {
    unsigned c = i;
#if defined(USE_AVX512)
    unsigned b = c & 0x18;
    b = (b << 1) | (b >> 1);
    c = (c & ~0x18) | (b & 0x18);
#endif
    w[c] = *d++;
  }
}

INLINE unsigned nnue_wt_idx(unsigned r, unsigned c, unsigned dims)
{
  (void)dims;

#if defined(USE_AVX512)
  if (dims > 32) {
    unsigned b = c & 0x38;
    b = (b << 1) | (b >> 2);
    c = (c & ~0x38) | (b & 0x38);
  }
  else if (dims == 32) {
    unsigned b = c & 0x18;
    b = (b << 1) | (b >> 1);
    c = (c & ~0x18) | (b & 0x18);
  }

#elif defined(USE_AVX2)
  if (dims > 32) {
    unsigned b = c & 0x18;
    b = (b << 1) | (b >> 1);
    c = (c & ~0x18) | (b & 0x18);
  }

#endif

#if defined(USE_AVX512)
  return c * 64 + r + (r & ~7);

#else
  return c * 32 + r;

#endif
}

static const char *nnue_read_hidden_weights(weight_t *w, unsigned dims,
    const char *d)
{
  for (unsigned r = 0; r < 32; r++)
    for (unsigned c = 0; c < dims; c++)
      w[nnue_wt_idx(r, c, dims)] = *d++;

  return d;
}

#ifdef USE_AVX2
static void nnue_permute_biases(int32_t *biases)
{
  __m128i *b = (__m128i *)biases;
  __m128i tmp[8];
#ifdef USE_AVX512
  tmp[0] = b[0];
  tmp[1] = b[2];
  tmp[2] = b[4];
  tmp[3] = b[6];
  tmp[4] = b[1];
  tmp[5] = b[3];
  tmp[6] = b[5];
  tmp[7] = b[7];
#elif USE_AVX2
  tmp[0] = b[0];
  tmp[1] = b[4];
  tmp[2] = b[1];
  tmp[3] = b[5];
  tmp[4] = b[2];
  tmp[5] = b[6];
  tmp[6] = b[3];
  tmp[7] = b[7];
#else
#error
#endif
  memcpy(b, tmp, 8 * sizeof(__m128i));
}
#endif

enum {
  TransformerStart = 3 * 4 + 177,
  NetworkStart = TransformerStart + 4 + 2 * 256 + 2 * 256 * 64 * 641
};

static bool nnue_verify_net(const void *evalData, size_t size)
{
  net_architecture = 0;

  if (size != 21022697) {
    net_architecture = NNUE_Architecture_unknown;
    return false;
  }
  const char *d = (const char*)evalData;
  if (nnue_readu_le_u32(d) != NnueVersions[0] && nnue_readu_le_u32(d) != NnueVersions[1]) {
    net_architecture = NNUE_Architecture_unknown;
    return false;
  }
  else {
    net_architecture = NNUE_Architecture_HalfKP_256_2_32_32_1;
  }
  if (nnue_readu_le_u32(d) == NnueVersions[0])
    net_rotated = 1;
  if (nnue_readu_le_u32(d + 4) != 0x3e5aa6eeU)
    return false;
  if (nnue_readu_le_u32(d + 8) != 177)
    return false;
  if (nnue_readu_le_u32(d + TransformerStart) != 0x5d69d7b8)
    return false;
  if (nnue_readu_le_u32(d + NetworkStart) != 0x63337156)
    return false;

  return true;
}

static void nnue_init_weights(const void *evalData)
{
  const char *d = (const char *)evalData + TransformerStart + 4;

  // Read transformer
  for (unsigned i = 0; i < K_HALF_DIMENSIONS; i++, d += 2)
    ft_biases[i] = nnue_readu_le_u16(d);
  for (unsigned i = 0; i < K_HALF_DIMENSIONS * FT_IN_DIMS; i++, d += 2)
    ft_weights[i] = nnue_readu_le_u16(d);

  // Read network
  d += 4;
  for (unsigned i = 0; i < 32; i++, d += 4)
    hidden1_biases[i] = nnue_readu_le_u32(d);
  d = nnue_read_hidden_weights(hidden1_weights, 512, d);
  for (unsigned i = 0; i < 32; i++, d += 4)
    hidden2_biases[i] = nnue_readu_le_u32(d);
  d = nnue_read_hidden_weights(hidden2_weights, 32, d);
  for (unsigned i = 0; i < 1; i++, d += 4)
    output_biases[i] = nnue_readu_le_u32(d);
  nnue_read_output_weights(output_weights, d);

#ifdef USE_AVX2
  nnue_permute_biases(hidden1_biases);
  nnue_permute_biases(hidden2_biases);
#endif
}

/*
 **********************************************************************
 *                                                                    *
 *   nnue_load_eval_file() is used to initialize the usage of NNUE.   *
 *   The NNUE eval file is read from disk, the NNUE with its weights  *
 *   is loaded into memory evalData and verified for validity.        *
 *                                                                    *
 **********************************************************************
 */
#if defined(_WIN32) || defined(_WIN64)
static bool nnue_load_eval_file(const char *evalFile)
{
  const void *evalData;
  HANDLE mapping;
  size_t size;

  {
    HANDLE fd = OpenNNUEFile(evalFile);
    if (fd == INVALID_HANDLE_VALUE)
      return false;
    evalData = MapNNUEFile(fd, &mapping);
    size = GetNNUEFileSize(fd);
    CloseNNUEFile(fd);
  }

  bool success = nnue_verify_net(evalData, size);
  if (success)
    nnue_init_weights(evalData);
  if (mapping) UnmapNNUEFile(evalData, mapping);
  return success;
}
#endif

/**
* Interfaces
*/

/*
 **********************************************************************
 *                                                                    *
 *   NNUE_init() is used to initialize the usage of NNUE.  The NNUE   *
 *   eval file is read from disk and the NNUE with its weights is     *
 *   loaded into memory.                                              *
 *                                                                    *
 **********************************************************************
 */
int NNUE_init(const char* evalFile, int *architecture, int *rotated)
{
  int success = 0;
  *rotated = 0;
  *architecture = 0;

  if (nnue_load_eval_file(evalFile)) {
    success = 1;
    *architecture = net_architecture;
    *rotated = net_rotated;
  } else {
    *architecture = net_architecture;
  }

  POS = (Position *)_aligned_malloc(CPUS * sizeof(Position), 64);
#ifdef _NNUE_EVALUATE_INCREMENTAL
  NNUE = (NNUEdata *)_aligned_malloc(CPUS * 3 * sizeof(NNUEdata), 64);
#else
  NNUE = (NNUEdata *)_aligned_malloc(CPUS * sizeof(NNUEdata), 64);
#endif
  if (POS == NULL || NNUE == NULL) {
    if (POS != NULL)
      _aligned_free(POS);
    if (NNUE != NULL)
      _aligned_free(NNUE);
    success = 0;
  }

  return success;
}

/*
 **********************************************************************
 *                                                                    *
 *   NNUE_evaluate() is used to evaluate chess position by passing    *
 *   NNUE description of the position in the form of player, and      *
 *   piece and square arrays, and thread_id.                          *
 *                                                                    *
 **********************************************************************
 */
int NNUE_evaluate(
  int player, int* piece, int* square, int thread_id)
{
  Position position = ((Position)*(POS + thread_id));
  NNUEdata data = ((NNUEdata)*(NNUE + thread_id + 0));
  Accumulator accumulator = data.accumulator;
  accumulator.computedAccumulation = 0;
  position.nnue[0] = &data;
#ifdef _NNUE_EVALUATE_INCREMENTAL
  position.nnue[1] = NULL;
  position.nnue[2] = NULL;
#endif
  position.player = player;
  position.piece = piece;
  position.square = square;
  return nnue_evaluate_pos(&position);
}

/*
 **********************************************************************
 *                                                                    *
 *   NNUE_evaluate_incremental() is used to evaluate chess position   *
 *   by passing NNUE description of the position in the form of       *
 *   player, and piece and square arrays, and thread_id.              *
 *                                                                    *
 **********************************************************************
 */
#ifdef _NNUE_EVALUATE_INCREMENTAL
int NNUE_evaluate_incremental(
  int player, int* piece, int* square, int thread_id)
{
  assert(&NNUE[thread_id][0] && (uint64_t)(&nnue[thread_id][0]->accumulator) % 64 == 0);

  POS[thread_id].nnue[0] = &NNUE[thread_id][0];
  /* new interface functions should be provided to set NNUE[][1]
     and NNUE[][2], etc. -- Robert Jurjevic
   */
  POS[thread_id].nnue[1] = &NNUE[thread_id][1];
  POS[thread_id].nnue[2] = &NNUE[thread_id][2];
  POS[thread_id].player = player;
  POS[thread_id].piece = piece;
  POS[thread_id].square = square;
  return nnue_evaluate_pos(&POS[thread_id]);
}
#endif

/*
 **********************************************************************
 *                                                                    *
 *   NNUE_evaluate_fen() is used to evaluate chess position by        *
 *   passing FEN description of the position.                         *
 *                                                                    *
 **********************************************************************
 */
int NNUE_evaluate_fen(const char* fen, int thread_id)
{
  int piece[33], square[33], player, castle, fifty, move_number;
  DecodeNNUEFen((char *)fen, &player, &castle, &fifty, &move_number, piece, square);
  return NNUE_evaluate(player, piece, square, thread_id);
}

