/* last modified 09/09/21 */
#ifndef NNUE_H
#define NNUE_H

#ifndef __cplusplus
#include <inttypes.h>
#endif

/**
* Internal piece representation
*     wking=1, wqueen=2, wrook=3, wbishop= 4, wknight= 5, wpawn= 6,
*     bking=7, bqueen=8, brook=9, bbishop=10, bknight=11, bpawn=12
*
* Make sure the pieces you pass to the library from your engine
* use this format.
*/
enum colors {
    nwhite, nblack
};
enum pieces {
    blank=0, wking, wqueen, wrook, wbishop, wknight, wpawn,
             bking, bqueen, brook, bbishop, bknight, bpawn
};

#define NNUE_Architecture_unknown -1
#define NNUE_Architecture_HalfKP_256_2_32_32_1 1

/**
* nnue data structure
*/

typedef struct DirtyPiece {
  int dirtyNum;
  int pc[3];
  int from[3];
  int to[3];
} DirtyPiece;

/* The accumulator has a "white king" half and a "black king" half,
   where each half is a 256-element vector of 16-bit ints, which is
   equal to the sum of the weights of the "active" (pt, sq, ksq)
   features  plus a 256-element vector of 16-bit biases.
 */
typedef struct Accumulator {
#ifndef __cplusplus
#else
  alignas(64)
#endif
  int16_t accumulation[2][256];
  int computedAccumulation;
} Accumulator;

typedef struct NNUEdata {
  Accumulator accumulator;
  DirtyPiece dirtyPiece;
} NNUEdata;

/**
* Position data structure passed to core subroutines
* See @nnue_evaluate for a description of parameters
*/
typedef struct Position {
  int player;
  int* piece;
  int* square;
#ifdef _NNUE_EVALUATE_INCREMENTAL
  NNUEdata* nnue[3];
#else
  NNUEdata* nnue[1];
#endif
} Position;

int nnue_evaluate_pos(Position* pos);

/************************************************************************
*         EXTERNAL INTERFACES
*
* Load a NNUE file using
*
*   NNUE_init(file_path)
*
* and then probe score using one of three functions, whichever
* is convenient. From easy to hard
*
*   a) NNUE_evaluate_fen         - accepts a fen string for evaluation
*   b) NNUE_evaluate             - suitable for use in chess engines
*   c) NNUE_evaluate_incremental - for ultimate performance but will need
*                                  some work on the engines side
*
**************************************************************************/

/**
* Load NNUE file
*/
#ifdef __cplusplus
extern "C"
#endif
int NNUE_init(
  const char * evalFile,            /** Path to NNUE file */
  int *architecture,                /** NNUE architecture ID */
  int *rotated                      /** Is the NNUE rotated or flipped */
);

/**
* Evaluation subroutine suitable for chess engines.
* -------------------------------------------------
* Piece codes are
*   wking=1, wqueen=2, wrook=3, wbishop= 4, wknight= 5, wpawn= 6,
*   bking=7, bqueen=8, brook=9, bbishop=10, bknight=11, bpawn=12,
* Squares are
*   A1=0, B1=1 ... H8=63
* Input format:
*   piece[0] is nwhite king, square[0] is its location
*   piece[1] is nblack king, square[1] is its location
*   ..
*   piece[i>1], square[i>1] can be in any order
*   ..
*   piece[n+1] is set to 0 to represent end of array
* Returns
*   Score relative to side to move in approximate centi-pawns
*/
#ifdef __cplusplus
extern "C"
#endif
int NNUE_evaluate(
  int player,                       /** Side to move: nwhite=0 nblack=1 */
  int* piece,                       /** Array of pieces */
  int* square,                      /** Corresponding array of squares each piece stands on */
  int thread_id                     /** Thread ID */
);

/**
* Incremental NNUE evaluation function.
* -------------------------------------------------
* Parameters and return type are as in NNUE_evaluate()
*
*/
#ifdef _NNUE_EVALUATE_INCREMENTAL
#ifdef __cplusplus
extern "C"
#endif
int NNUE_evaluate_incremental(
  int player,                       /** Side to move: nwhite=0 nblack=1 */
  int* piece,                       /** Array of pieces */
  int* square,                      /** Corresponding array of squares each piece stands on */
  int thread_id                     /** Thread ID */
);
#endif

/**
* Evaluate on FEN string
* Returns
*   Score relative to side to move in approximate centi-pawns
*/
#ifdef __cplusplus
extern "C"
#endif
int NNUE_evaluate_fen(
  const char* fen,                  /** FEN string to probe evaluation for */
  int thread_id                     /** Thread ID */
);

#endif

/**
  For incremental NNUE evaluation, you keep track of the NNUEdata structure
  (i.e. accumulator and dirtyPiece) for each thread. For ply < 2, you set
  the pointers to 0 for those plies you don't have data for. To see an example
  implementation of these look here
    https://github.com/dshawul/Scorpio/commit/04bfc4f951e8eed10919a1fa98910e7b7322bf84
  Basically, you update dirtyPiece in your doMove and doNullMove functions,
  and copy accumulator in your doNullMove. If you get it right, the search
  results should match exactly with the incremental, and non-incremental calls.
  It took me some time to weed out bugs.
  -- Daniel Shawul
*/

/**
  This is NNUE eval code is extracted from CFish and published by Daniel Shawul; and
  it is optimised for building Windows binaries using Microsoft Visual Studio. It
  seems that the original contribution to C++ Stockfish was done by a Japanese Shogi AI
  programmer, Hisayori Noda (nodchip on GitHub). My understanding is that Shogi AI
  did not work without NNUE, i.e., they had problems in eval function for Shogi, etc.;
  then, as I understand, Ronald de Man made quite a radical rewrite of the Hisayori's
  C++ code, which is now, as it would seem in CFish. What Ronald de Man did is to remove
  NNUE C++ classes and code all with C/C++ functions; the majority of the code is still
  in a C++ file, and that is, as I understand, to use a C++ memory alignment function,
  etc., which seems does not exist in C, etc. The structure of the NNUE is pretty much
  kept similar as it was in Shogi, but adapted to chess. "It would seems that the position
  is stored in the form of KingSquare*PieceType*Square for each side; so one has a boolean
  input for whether there is a black queen on d3 and white king on f5, for example; there
  are 5 non-king pieces times two for the two sides giving 10 piece types in total; this
  gives 40,960 inputs per side, of which at most 32 are non-zero; when a piece moves one
  can incrementally update the output for the input of the first hidden layer; this
  calculation is done for each side giving n outputs for white and n outputs for black;
  depending on whether it is white or black to more the outputs are concatenated
  differently; so either [white outputs, black outputs] or [black outputs, white outputs];
  the rest of the network is then just a standard fully connected feed-forward network; this
  encoding allows for the first layer to be calculated very efficiently while still having
  a large amount of weights and the most important relation of the pieces (to the two
  respective kings!) being stored explicitly." Here is the link to the original Japanese
  article on NNUE. http://yaneuraou.yaneu.com/2020/06/19/stockfish-nnue-the-complete-guide
  -- Robert Jurjevic (M&R Research Team, London, England)
*/

