#include "chess.h"
#include "data.h"
/* last modified 04/24/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   Quiesce() is the recursive routine used to implement the alpha/beta       *
 *   negamax search (similar to minimax but simpler to code.)  Quiesce() is    *
 *   called whenever there is no "depth" remaining so that only good capture   *
 *   moves and some checks are searched deeper.                                *
 *                                                                             *
 *******************************************************************************
 */
int Quiesce(TREE * RESTRICT tree, int alpha, int beta, int wtm,
  int ply, int check_depth, int do_delta)
{
  register int o_alpha = alpha, value = alpha;
  register int *next_move;
  register int moves_searched = 0;
  register int *movep, *sortv, temp, t;
  register int searched, skip;
  register int do_futil = 1;
  register int do_nnue = 1;

  if (ply >= MAXPLY - 1)
    return (beta);
/*
 ************************************************************
 *                                                          *
 *   initialize.                                            *
 *                                                          *
 ************************************************************
 */
  if (!(search_quiescent_flags & DO_QCHECKS) ||
    check_depth < 0)
    check_depth = 0;
  if (!(search_pruning_flags & DO_FUTILITY) ||
    pruning_margin[0] == 0 || !do_delta)
    do_futil = 0;
  if (!NNUE_setup)
    do_nnue = 0;
/*
 ************************************************************
 *                                                          *
 *   check for draw by repetition.                          *
 *                                                          *
 ************************************************************
 */
  if (check_depth || tree->inchk[ply]) {
    if (RepetitionCheck(tree, ply, wtm)) {
      value = DrawScore(wtm);
      if (value < beta) {
        SavePV(tree, ply, 0);
      }
#if defined(TRACE)
      if (ply <= trace_level)
        printf("draw by repetition detected, ply=%d.\n", ply);
#endif
      return (value);
    }
  }
/*
 ************************************************************
 *                                                          *
 *   now call Evaluate() or EvaluateNNUE() to produce the   *
 *   "stand-pat" score that will be returned if no capture  *
 *   is acceptable.  if this score is > alpha, then we also *
 *   have to save the "path" to this node as it is the PV   *
 *   that leads to this score.                              *
 *                                                          *
 ************************************************************
 */
  if (!tree->inchk[ply]) {
    if (do_nnue) {
      value = EvaluateNNUE(tree, ply, wtm, alpha, beta);
    }
    else {
      value = Evaluate(tree, ply, wtm, alpha, beta);
    }
#if defined(TRACE)
    if (trace_level >= 99)
      Trace(tree, ply, value, wtm, alpha, beta, "quiesce", EVALUATION);
#endif
    if (value > alpha) {
      if (value >= beta) {  /* "stand-pat" beta cut-off */
        return (value);
      }
      alpha = value;
      tree->pv[ply].pathl = ply;
      tree->pv[ply].pathh = 0;
      tree->pv[ply].pathd = iteration_depth;
    }
  }
/*
 ************************************************************
 *                                                          *
 *   if the side to move is in check generate check         *
 *   evasions.                                              *
 *                                                          *
 ************************************************************
 */
  if (tree->inchk[ply]) {
    tree->last[ply] = GenerateCheckEvasions(tree, ply, wtm, tree->last[ply - 1]);
  }
/*
 ************************************************************
 *                                                          *
 *   if the side to move is not in check generate captures, *
 *   pawn promotions, non-promotion pawn advances, and if   *
 *   there is a remaining check depth non-capture checks.   *
 *                                                          *
 ************************************************************
 */
  else {
    tree->last[ply] = GenerateCaptures(tree, ply, wtm, tree->last[ply - 1]); /* generates captures including non-capture queen promotions */
    if (check_depth) {
      tree->last[ply] = GenerateNoncaptureChecks(tree, ply, wtm, tree->last[ply]); /* generates non-promotion non-capture checks with some unwanted non-capture non-checks */
    }
  }
/*
 ************************************************************
 *                                                          *
 *   assign sort value for each generated move, in Quesce   *
 *   we will try only moves with non-negative sort value.   *
 *                                                          *
 ************************************************************
 */
  if (tree->last[ply] > tree->last[ply - 1]) {
    for (movep = tree->last[ply - 1], sortv = tree->sort_value;
      movep < tree->last[ply]; movep++, sortv++) {
      *sortv = 0;
      if (tree->inchk[ply]) {
        if (Captured(*movep)) { /* capture check evasions */
          t = 0;
          temp = MVV_LVA[Captured(*movep)][Piece(*movep)];
          if (temp < 0 && (t = Swap(tree, *movep, wtm)) >= 0)
            temp = t;
          if (Promote(*movep) && t == 0) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              temp += pc_values[Promote(*movep)];
          }
          if (temp < 0)
            temp = 0; /* try all capture check evasions */
        } else { /* non-capture promotions and other non-capture check evasions */
          temp = 0;
          if (Promote(*movep)) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              temp += pc_values[Promote(*movep)] - pc_values[pawn];
          }
        }
      }
      else {
        if (Captured(*movep)) { /* captures */
          t = 0;
          temp = MVV_LVA[Captured(*movep)][Piece(*movep)];
          if (temp < 0 && (t = Swap(tree, *movep, wtm)) >= 0)
            temp = t;
          if (Promote(*movep) && t == 0) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              temp += pc_values[Promote(*movep)];
          }
/*
 ************************************************************
 *                                                          *
 *   if the capture removes the last opponent's piece or if *
 *   it is a pawn capture and the opponent has no pieces    *
 *   left, we always search this kind of capture since this *
 *   can be the move that allows a passed pawn to promote   *
 *   when the opponent has no piece to catch it.            *
 *                                                          *
 ************************************************************
 */
          if (temp < 0) {
            if (TotalPieces(Flip(wtm), occupied) - p_vals[Captured(*movep)] <= 0)
              temp = 0;
          }
        } else { /* non-capture promotions and non-capture non-promotion checks, some unwanted non-capture non-checks and pawn attacking advances */
          temp = 0;
          if (Promote(*movep)) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              temp += pc_values[Promote(*movep)] -  pc_values[pawn];
          }
        }
      }
      *sortv = temp;
    }
  }
/*
 ************************************************************
 *                                                          *
 *   now remove bad captures form the move list.            *
 *                                                          *
 ************************************************************
 */
  if (tree->last[ply] > tree->last[ply - 1]) {
    if (!tree->inchk[ply]) {
      for (movep = tree->last[ply - 1], sortv = tree->sort_value;
        movep < tree->last[ply]; movep++, sortv++) {
        if (Captured(*movep)) {
          if (*sortv < 0)
            *movep = NULL_MOVE;
        }
      }
    }
  }

/*
 ************************************************************
 *                                                          *
 *  This is a simple insertion sort algorithm.  It seems to *
 *  be no faster than a normal bubble sort, but using this  *
 *  eliminated a lot of explaining about "why?". :)         *
 *                                                          *
 ************************************************************
 */
  if (tree->last[ply] > tree->last[ply - 1] + 1) {
    register int move1, sortv1, *pmove0, *psortv0;
    sortv = tree->sort_value + 1;
    for (movep = tree->last[ply - 1] + 1; movep < tree->last[ply]; movep++, sortv++) {
      move1 = *movep;
      sortv1 = *sortv;
      pmove0 = movep - 1;
      psortv0 = sortv - 1;
      while (pmove0 >= tree->last[ply - 1] && ((*psortv0 < sortv1) || (*psortv0 == sortv1 &&
        *pmove0 && move1 && pc_values[Captured(*pmove0)] < pc_values[Captured(move1)]))) {
        *(psortv0 + 1) = *psortv0;
        *(pmove0 + 1) = *pmove0;
        pmove0--;
        psortv0--;
      }
      *(pmove0 + 1) = move1;
      *(psortv0 + 1) = sortv1;
    }
  }
  next_move = tree->last[ply - 1];
/*
 ************************************************************
 *                                                          *
 *   now iterate through the move list and search the       *
 *   resulting positions.                                   *
 *                                                          *
 ************************************************************
 */
  while (next_move < tree->last[ply]) {
    searched = 0;
    tree->curmv[ply] = *(next_move++);
    if (tree->curmv[ply] == NULL_MOVE)
      continue; /* the move is a bad capture */
#if defined(TRACE)
    if (ply <= trace_level)
      Trace(tree, ply, 0, wtm, alpha, beta, "quiesce", CAPTURE_MOVES);
#endif
    MakeMove(tree, ply, tree->curmv[ply], wtm);
    if (!Check(wtm)) { /* if legal */
      do { /* do search */
        if (Check(Flip(wtm))) /* if check */
          tree->inchk[ply + 1] = 1;
        else
          tree->inchk[ply + 1] = 0;
/*
 ************************************************************
 *                                                          *
 *   we decide here which moves to skip.                    *
 *                                                          *
 ************************************************************
 */
        skip = 0;
        if (!tree->inchk[ply]) {
          if (!Promote(tree->curmv[ply])) {
            if (!Captured(tree->curmv[ply])) {
              if (!tree->inchk[ply + 1]) {
                skip = 1; /* skip non-promotion non-capture non-checks which may be generated by GenerateNoncaptureChecks */
              }
              else {
                skip = 1; /* assume skipping non-promotion non-capture checks */
                if (MaterialBalanceRoot(pawn_value)) {
                  register int loss = NO_VALUE;
                  if (DoubleCheck(tree, wtm)) {
                    skip = 0; /* do not skip non-promotion non-capture double checks */
                  }
                  else {
                    if ((Piece(tree->curmv[ply]) == queen &&
                      Distance(KingSQ(Flip(wtm)), To(tree->curmv[ply])) < 2)) {
                      if (loss == NO_VALUE) {
                        register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                        loss = 0;
                        if (attacks & Occupied(Flip(wtm))) {
                          loss = SwapO(tree, tree->curmv[ply], wtm) -
                            pc_values[Captured(tree->curmv[ply])];
                        }
                      }
                      if (loss <= 0) /* no material loss */
                        skip = 0; /* do not skip non-promotion non-capture queen-king contact checks */
                    }
                    else {
                      int *mvp, *lastm, emoves[256], king_moves = 0;
                      lastm = GenerateCheckEvasions(tree, ply + 1, Flip(wtm), emoves);
                      for (mvp = emoves; mvp < lastm; mvp++) {
                        if (Piece(*mvp) == king)
                          king_moves++;
                        if (king_moves > 1)
                          break;
                      }
                      if (king_moves <= 1) {
                        if (loss == NO_VALUE) {
                          register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                          loss = 0;
                          if (attacks & Occupied(Flip(wtm))) {
                            loss = SwapO(tree, tree->curmv[ply], wtm) -
                              pc_values[Captured(tree->curmv[ply])];
                          }
                        }
                        if (loss <= 0) /* no material loss */
                          skip = 0; /* do not skip non-promotion non-capture checks which give opponent's king at most one escape square */
                      }
                    }
                  }
                }
              }
            }
          }
        }
        if (skip)
          continue;
/*
 ************************************************************
 *                                                          *
 *   delta pruning is a technique similar in concept to     *
 *   futility pruning, only used in the quiescence search.  *
 *   it works as follows: before we make a capture, we test *
 *   whether the captured piece value plus some safety      *
 *   margin (typically around 200 centipawns) are enough to *
 *   raise alpha for the current node.  for example, if the *
 *   side to move is a rook down, it does not bother to     *
 *   test captures of pawns, since they are unlikely to     *
 *   improve matters.  capturing a minor piece, however,    *
 *   might be sufficient, given enough positional           *
 *   compensation.  it follows that the safety margin       *
 *   (delta) should be around 200 centipawns, depending on  *
 *   the piece values used by the program.  for safety      *
 *   reasons, delta pruning should be switched off in the   *
 *   late endgame, since otherwise quiescence search would  *
 *   be blind to insufficient material issues and           *
 *   transitions into won endgames made at the expense of   *
 *   some material.  in particular if the move is a capture *
 *   removing the last opponent's piece or if it is a pawn  *
 *   capture and the opponent has no pieces left, then we   *
 *   may wish not to delta prune.                           *
 *                                                          *
 *   Exception:                                             *
 *                                                          *
 *     We do not prune if we are safely pushing a passed    *
 *     pawn and there is enough depth remaining that the    *
 *     pawn might make significant progress advancing.      *
 *                                                          *
 ************************************************************
 */
        if (do_futil) {
          if (ply > 1 && !tree->inchk[ply] && moves_searched &&
            !Promote(tree->curmv[ply]) && NotMate(alpha, beta)) {
            if (MaterialSTM(wtm) + pruning_margin[0] <= alpha) {
              if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
                rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
                if (!Captured(tree->curmv[ply]) ||
                  TotalPieces(Flip(wtm), occupied) > 0)
                  continue; /* skip move to be delta pruned */
              }
            }
          }
        }
        if (!tree->inchk[ply]) { /* if not in check */
          if (tree->inchk[ply + 1]) /* if check */
            tree->qchecks_done++;
        }
        tree->nodes_searched++;
        tree->quiescent_nodes_searched++;
        value = -Quiesce(tree, -beta, -alpha, Flip(wtm), ply + 1,
          ((check_depth > 0) ? ((tree->inchk[ply + 1]) ?
          ((!Captured(tree->curmv[ply])) ? check_depth - 1 : check_depth) :
          check_depth) : 0), do_delta);
        if (abort_search || tree->stop)
          return (0);
        searched = 1;
        moves_searched++;
      } while (0); /* end do search */
    }
    UnmakeMove(tree, ply, tree->curmv[ply], wtm);
    if (abort_search || tree->stop)
      return (0);
    if (searched) {
      if (value > alpha) {
        if (value >= beta) {
          return (value);
        }
        alpha = value;
      }
    }
  }
/*
 ************************************************************
 *                                                          *
 *   all moves have been searched.  return the search       *
 *   result that was found.  if the result is not the       *
 *   original alpha score, then we need to return the PV    *
 *   that is associated with this score.                    *
 *                                                          *
 ************************************************************
 */
  if (!moves_searched) { /* if leaf */
    if (tree->inchk[ply]) { /* mate */
      value = -(MATE - ply);
      if (value > alpha) {
        if (value >= beta) {
          return (value);
        }
        alpha = value;
        tree->pv[ply].pathl = ply;
        tree->pv[ply].pathh = 0;
        tree->pv[ply].pathd = iteration_depth;
      }
    }
  }
  if (alpha != o_alpha) {
    tree->pv[ply - 1] = tree->pv[ply];
    tree->pv[ply - 1].path[ply - 1] = tree->curmv[ply - 1];
  }
  return (alpha);
}

