#include "chess.h"
#include "data.h"
#include "epdglue.h"
#define BIG_SCORE_1 1000000
#define BIG_SCORE_2 2000000
#define BIG_SCORE_3 3000000
#define BIG_SCORE_4 4000000
/* last modified 05/26/20 */
/*
 *******************************************************************************
 *                                                                             *
 *   RootMoveList() is used to set up the ply one move list.  it is a  more    *
 *   accurate ordering of the move list than that done for plies deeper than   *
 *   one.  briefly, Quiesce() is used to obtain the positional score plus the  *
 *   expected gain/loss for pieces that can be captured.                       *
 *                                                                             *
 *******************************************************************************
 */
void RootMoveList(int wtm)
{
  int *mvp, *lastm, rmoves[256], sort_value[256];
  int i, done, temp, value;
  TREE *const tree = block[0];
  int tb_value;
  int mating_via_tb = 0;
  int ppv = 0;
  int ply = 1;

/*
 ************************************************************
 *                                                          *
 *   if the position at the root is a draw, based on EGTB   *
 *   results, we are going to behave differently.  we will  *
 *   extract the root moves that are draws, and toss the    *
 *   losers out.  then, we will do a normal search on the   *
 *   moves that draw to try and chose the most difficult    *
 *   drawing move.                                          *
 *                                                          *
 ************************************************************
 */
#if !defined(NOEGTB)
  EGTB_draw = 0;
  if (EGTBlimit && TotalAllPieces <= EGTBlimit &&
      Castle(ply, white) + Castle(ply, black) == 0 &&
      EGTBProbe(tree, ply, wtm, &tb_value)) {
    if (swindle_mode && (tb_value == DrawScore(wtm)))
      if ((wtm && Material > 0) || (!wtm && Material < 0))
        EGTB_draw = 1;
    if (tb_value > NOT_MATE)
      mating_via_tb = -tb_value - 1;
  }
#endif
/*
 ************************************************************
 *                                                          *
 *   first, use GenerateMoves() to generate the set of      *
 *   legal moves from the root position.                    *
 *                                                          *
 ************************************************************
 */
  lastm = GenerateCaptures(tree, ply, wtm, rmoves);
  best_root_capture = 0;
  if (lastm > rmoves) {
    int *movep, *sortv, t;
    for (movep = rmoves, sortv = sort_value;
        movep < lastm; movep++, sortv++) {
      *sortv = 0;
      t = 0;
      if (Captured(*movep)) {
        *sortv = MVV_LVA[Captured(*movep)][Piece(*movep)];
        if (*sortv < 0 && (t = Swap(tree, *movep, wtm)) >= 0)
          *sortv = t;
      }
      if (Promote(*movep) && t == 0) {
        register BITBOARD attacks = AttacksTo(tree, To(*movep));
        register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
        if (!loose)
          *sortv += pc_values[Promote(*movep)] -
            (Captured(*movep) ? 0 : pc_values[pawn]);
      }
      if (best_root_capture < *sortv + Piece(*movep) && t == 0) {
        register BITBOARD attacks = AttacksTo(tree, To(*movep));
        register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
          if (loose)
            best_root_capture = Max(best_root_capture, *sortv);
          else
            best_root_capture = Max(best_root_capture, *sortv +
              ((Captured(*movep)) ? Piece(*movep) : 0));
      }
    }
  }
  lastm = GenerateNoncaptures(tree, ply, wtm, lastm);
  n_root_moves = lastm - rmoves;
/*
 ************************************************************
 *                                                          *
 *   now make each move and use Evaluate() to compute the   *
 *   positional evaluation.                                 *
 *                                                          *
 ************************************************************
 */
  for (mvp = rmoves; mvp < lastm; mvp++) {
    value = -BIG_SCORE_4;
#if defined(TRACE)
    if (trace_level >= 1) {
      tree->curmv[ply] = *mvp;
      tree->phase[ply] = HASH_MOVE;
      Trace(tree, ply, 0, wtm, -INFTY, INFTY, "RootMoves()", tree->phase[ply]);
    }
#endif
    MakeMove(tree, ply, *mvp, wtm);
    if (!Check(wtm)) /* if legal */
      do { /* do search */
        tree->nodes_searched++;
        tree->curmv[ply] = *mvp;
#if !defined(NOEGTB)
        if (TotalAllPieces <= EGTBlimit && EGTB_draw &&
            Castle(ply, white) + Castle(ply, black) == 0) {
          i = EGTBProbe(tree, ply + 1, Flip(wtm), &tb_value);
          if (i && tb_value != DrawScore(Flip(wtm)))
            break;
        }
        if (mating_via_tb && TotalAllPieces <= EGTBlimit &&
            Castle(ply, white) + Castle(ply, black) == 0) {
          i = EGTBProbe(tree, ply + 1, Flip(wtm), &tb_value);
          if (i && ((mating_via_tb > DrawScore(Flip(wtm)) &&
                      tb_value < mating_via_tb) ||
                  (mating_via_tb < DrawScore(Flip(wtm)) &&
                      tb_value > mating_via_tb)))
            break;
        }
#endif
        value = -Quiesce(tree, -INFTY, INFTY, Flip(wtm), ply + 1,
          quiesce_check_depth, NO_DELTA);
/*
 ************************************************************
 *                                                          *
 *   add in a bonus if this move is part of the previous    *
 *   principal variation.  it was good in the search, we    *
 *   should try it first now.                               *
 *                                                          *
 ************************************************************
 */
        if ((Piece(*mvp) == Piece(last_pv.path[1])) &&
            (From(*mvp) == From(last_pv.path[1])) &&
            (To(*mvp) == To(last_pv.path[1])) &&
            (Captured(*mvp) == Captured(last_pv.path[1])) &&
            (Promote(*mvp) == Promote(last_pv.path[1])))
          value += BIG_SCORE_2;
/*
 ************************************************************
 *                                                          *
 *   fudge the score for promotions so that promotion to a  *
 *   queen is tried first.  since the positional score is   *
 *   computed before the piece is removed, under-promoting  *
 *   sometimes looks better.                                *
 *                                                          *
 ************************************************************
 */
        if (Promote(*mvp) && (Promote(*mvp) != queen))
          value -= pawn_value / 2;
      } while (0); /* end do search */
    sort_value[mvp - rmoves] = value;
    UnmakeMove(tree, ply, *mvp, wtm);
    if (abort_search || tree->stop)
      return;
  }
/*
 ************************************************************
 *                                                          *
 *   now sort the moves into order based on the sum of the  *
 *   positional score less the possible hung pieces, and    *
 *   factoring in a bonus for following the PV move.        *
 *                                                          *
 ************************************************************
 */
  do {
    done = 1;
    for (i = 0; i < lastm - rmoves - 1; i++) {
      if (sort_value[i] < sort_value[i + 1]) {
        temp = sort_value[i];
        sort_value[i] = sort_value[i + 1];
        sort_value[i + 1] = temp;
        temp = rmoves[i];
        rmoves[i] = rmoves[i + 1];
        rmoves[i + 1] = temp;
        done = 0;
      }
    }
  } while (!done);
/*
 ************************************************************
 *                                                          *
 *   now trim the move list to eliminate those moves that   *
 *   "hung" the king and are illegal.                       *
 *                                                          *
 ************************************************************
 */
  for (; n_root_moves; n_root_moves--)
    if (sort_value[n_root_moves - 1] > -BIG_SCORE_3)
      break;
  for (i = 0; i < n_root_moves; i++) {
    if (sort_value[i] > BIG_SCORE_1)
      sort_value[i] -= BIG_SCORE_2;
  }
/*
 ************************************************************
 *                                                          *
 *   debugging output to dump root move list and the stuff  *
 *   used to sort them, for testing and debugging.          *
 *                                                          *
 ************************************************************
 */
  if (display_options & 512) {
    Print(512, "%d moves at root\n", n_root_moves);
    Print(512, "        move   score\n");
    for (i = 0; i < n_root_moves; i++) {
      tree->curmv[ply] = rmoves[i];
      Print(512, "%12s", OutputMove(tree, rmoves[i], 1, wtm));
      Print(512, "%8d\n", sort_value[i]);
    }
  }
/*
 ************************************************************
 *                                                          *
 *   now check to see if we are in the special mode where   *
 *   moves need to be searched because of missing EGTBs.    *
 *                                                          *
 ************************************************************
 */
#if !defined(NOEGTB)
  if (mating_via_tb) {
    for (i = 0; i < n_root_moves; i++) {
      tree->curmv[ply] = rmoves[i];
      MakeMove(tree, ply, rmoves[i], wtm);
      if (mating_via_tb && TotalAllPieces <= EGTBlimit &&
          Castle(ply, white) + Castle(ply, black) == 0)
        temp =
            (EGTBProbe(tree, ply + 1, Flip(wtm), &tb_value) != DrawScore(Flip(wtm)));
      else
        temp = 0;
      UnmakeMove(tree, ply, rmoves[i], wtm);
      if (temp)
        break;
    }
    EGTB_search = (i == n_root_moves);
  } else
    EGTB_search = 0;
#endif
/*
 ************************************************************
 *                                                          *
 *   now copy the root moves into the root_move structure   *
 *   array for use by NextRootMove().                       *
 *                                                          *
 ************************************************************
 */
  for (i = 0; i < n_root_moves; i++) {
    root_moves[i].move = rmoves[i];
    root_moves[i].nodes = 0;
    root_moves[i].status = 0;
    root_moves[i].bm_age = 0;
  }
  for (i = 0; i < n_root_moves; i++)
    root_moves[i].status |= 8;
  if (n_root_moves)
    root_moves[0].status |= (4 + 8);
  return;
}
