#include "chess.h"
#include "data.h"
#include "epdglue.h"
/* last modified 05/15/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   Search() is the recursive routine used to implement the alpha/beta        *
 *   negamax search (similar to minimax but simpler to code.)  Search() is     *
 *   called whenever there is "depth" remaining so that all moves are subject  *
 *   to searching.  Search() recursively calls itself so long as there is at   *
 *   least one ply of depth left, otherwise it calls Quiesce() instead.        *
 *                                                                             *
 *******************************************************************************
 */
int Search(TREE * RESTRICT tree, int alpha, int beta, int wtm, int depth,
    int ply, int do_null)
{
  register int moves_searched = 0;
  register int value = alpha, o_alpha = alpha, t_beta = beta, mating_value;
  register int extension = 0, repeat, nws = 0;
  register int do_futil = 1;
  register int do_probcut = 1;
  register int do_internal_deepen = 1;
  register int extension_singular = 0;
  register int first_tried = NULL_MOVE;
  int best_capture = 0;
  NODE_TYPE node_type = (alpha != beta - 1) ? PV_NODE : CUT_NODE;
  UINT64 start_nodes = tree->nodes_searched;
  HASH hash;
  int searched[256];

  if (ply >= MAXPLY - 1)
    return (beta);
/*
 ************************************************************
 *                                                          *
 *   check to see if we have searched enough nodes that it  *
 *   is time to peek at how much time has been used, or if  *
 *   is time to check for operator keyboard input.  this is *
 *   usually enough nodes to force a time/input check about *
 *   once per second, except when the target time per move  *
 *   is very small, in which case we try to check the time  *
 *   at least 10 times during the search.                   *
 *                                                          *
 ************************************************************
 */
  if (time_check == 1) {
    time_check = 0;
    if (TimeCheck(tree, 1)) {
      time_abort++;
      abort_search = 1;
      return (0);
    }
  }
  if (tree->thread_id == 0) {
    if (CheckInput()) {
      Lock(lock_io);
      Interrupt(ply);
      Unlock(lock_io);
      if (abort_search)
        return (0);
    }
  }
  if (!(search_pruning_flags & DO_FUTILITY) ||
    pruning_margin[depth] == 0)
    do_futil = 0;
  if (!(do_probcut & DO_PROBCUT) ||
    probcut_margin == 0)
    do_probcut = 0;
  if (!(search_flags & DO_INTERNAL_DEEPEN))
    do_internal_deepen = 0;
  if (!(search_pruning_flags & DO_NULL_MOVE))
    do_null = 0;
/*
 ************************************************************
 *                                                          *
 *   Check for draw by repetition, which includes 50 move   *
 *   draws also.  This and the next two steps are skipped   *
 *   for root moves (ply = 1).                              *
 *                                                          *
 ************************************************************
 */

  tree->rep_list[wtm][Repetition(wtm) + (ply - 1) / 2] = HashKey;
  if (ply > 1) {
    if ((repeat = RepetitionCheck(tree, ply, wtm))) {
      if (repeat == 1 || !tree->inchk[ply]) {
        value = DrawScore(wtm);
        if (value < beta) {
          SavePV(tree, ply, 0);
        }
#if defined(TRACE)
        if (ply <= trace_level)
          printf("draw by repetition detected, ply=%d.\n", ply);
#endif
        return (value);
      }
    }
  }

/*
 ************************************************************
 *                                                          *
 *   If a forced mate has been found, mate distance pruning *
 *   will cut trees and adjust bounds of lines where no     *
 *   shorter mate is possible.  Mate distance pruning is a  *
 *   safe type of pruning, as it only cuts irrelevant       *
 *   nodes.  It will not add much to a programs playing     *
 *   strength as this pruning only comes into effect when a *
 *   position is already decided, still it helps with       *
 *   analysis and solving "Mate in n" problems.             *
 *                                                          *
 *   Assume we are in a winning position.  Through another  *
 *   line we have already found a forced mate in n plies.   *
 *   Thus our alpha is MATE - n.  From this information we  *
 *   can take that should we be searching a line that       *
 *   already is equal to or longer than n, we can not       *
 *   possibly increase alpha even if another mate was       *
 *   found.  This means we could safely prune that tree.    *
 *                                                          *
 *   Furthermore we can lower the upper bound. If we are n  *
 *   plies away from the root, the opponent can not         *
 *   possibly be evaluated greater than MATE - n.  Thus     *
 *   beta can be set to this value should it be higher.     *
 *                                                          *
 *   Contrary to the pruning if we are in a winning         *
 *   position, scores can also be adjusted and trees pruned *
 *   if we are in a losing position.                        *
 *                                                          *
 ************************************************************
 */
  mating_value = MATE - (ply - 1);
  if (mating_value < beta) {
     beta = mating_value;
     if (alpha >= mating_value)
       return (mating_value);
  }
  mating_value = -MATE + (ply - 1);
  if (mating_value > alpha) {
     alpha = mating_value;
     if (beta <= mating_value)
       return (mating_value);
  }
/*
 ************************************************************
 *                                                          *
 *   Now call HashProbe() to see if this position has been  *
 *   searched before.  If so, we may get a real score,      *
 *   produce a cutoff, or get nothing more than a good move *
 *   to try first.  There are four cases to handle:         *
 *                                                          *
 *   1. HashProbe() returns "HASH_HIT".  This terminates    *
 *   the search instantly and we simply return the value    *
 *   found in the hash table.  This value is simply the     *
 *   value we found when we did a real search in this       *
 *   position previously, and HashProbe() verifies that the *
 *   value is useful based on draft and current bounds.     *
 *                                                          *
 *   2. HashProbe() returns "AVOID_NULL_MOVE" which means   *
 *   the hashed score/bound was no good, but it indicated   *
 *   that trying a null-move in this position would be a    *
 *   waste of time since it will likely fail low, not high. *
 *                                                          *
 *   3. HashProbe() returns "HASH_MISS" when forces us to   *
 *   do a normal search to resolve this node.               *
 *                                                          *
 ************************************************************
 */
  switch (HashProbe(tree, ply, depth, wtm, alpha, beta, &hash)) {
    case HASH_HIT:
      return (hash.value);
    case AVOID_NULL_MOVE:
      do_null = 0;
      do_probcut = 0;
    case HASH_MISS:
      break;
  }
/*
 ************************************************************
 *                                                          *
 *   now it's time to try a probe into the endgame table-   *
 *   base files.  this is done if we notice that there are  *
 *   6 or fewer pieces left on the board.  EGTB_use tells   *
 *   us how many pieces to probe on.  note that this can be *
 *   zero when trying to swindle the opponent, so that no   *
 *   probes are done since we know it is a draw.            *
 *                                                          *
 ************************************************************
 */
#if !defined(NOEGTB)
    register int EGTB_depth = 3;
    if (depth > EGTB_depth && TotalAllPieces <= EGTB_use &&
        Castle(ply, white) + Castle(ply, black) == 0 &&
        Rule50Moves(ply) == 0) {
      int egtb_value;
      tree->egtb_probes++;
      if (EGTBProbe(tree, ply, wtm, &egtb_value)) {
        tree->egtb_probes_successful++;
        value = egtb_value;
        if (Abs(value) > NOT_MATE)
          value += (value > 0) ? -ply + 1 : ply;
        else if (value == 0) {
          value = DrawScore(wtm);
          if (swindle_mode) {
            if (Material > 0)
              value += (wtm) ? 1 : -1;
            else if (Material < 0)
              value -= (wtm) ? 1 : -1;
          }
        }
        if (value < beta)
            SavePV(tree, ply, 2);
        return (value);
      }
    }
#endif
  tree->inchk[ply + 1] = 0;
  tree->last[ply] = tree->last[ply - 1];
  tree->phase[ply] = NO_MOVE;
/*
 ************************************************************
 *                                                          *
 *  now, we try a null move to see if we can get a quick    *
 *  cutoff with only a little work.  this operates as       *
 *  follows.  instead of making a legal move, the side on   *
 *  move 'passes' and does nothing.  the resulting position *
 *  is searched to a shallower depth than normal (usually   *
 *  3 plies less but settable by the operator) this will    *
 *  result in a cutoff if our position is very good, but it *
 *  produces the cutoff much quicker since the search is    *
 *  far shallower than a normal search that would also be   *
 *  likely to fail high.                                    *
 *                                                          *
 *  this is skipped for any of the following reasons:       *
 *                                                          *
 *  1.  the side on move is in check.  the null move        *
 *      results in an illegal position.                     *
 *  2.  no more than one null move can appear in succession *
 *      or else the search will degenerate into nothing.    *
 *  3.  the side on move has little material left making    *
 *      zugzwang positions more likely.                     *
 *  4.  the transposition table probe found an entry that   *
 *      indicates that a null-move search will not fail     *
 *      high, so we avoid the wasted effort.                *
 *  5.  if the alpha/beta window is non-null, this is a PV  *
 *      node where the null move should be avoided.         *
 *                                                          *
 *  R is a common name for a constant or variable           *
 *  signifying depth reduction used in the Null Move        *
 *  Pruning. In most null move implementations it is either *
 *  2 or 3, the exact formula being depth - R - 1.          *
 *                                                          *
 *  When R becomes a variable, then we say that a program   *
 *  uses adaptive null move pruning, first described by     *
 *  Ernst A. Heinz. One common scheme is to use R=3 when    *
 *  normal search depth exceeds 6 plies and R=2 otherwise.  *
 *                                                          *
 *  The default for Fruit is to try a null search if the    *
 *  static evaluation is greater than beta, or the depth    *
 *  less than four. Newer versions of Fruit as well as Toga *
 *  default to always doing a null search, as long as there *
 *  is at least one piece of value greater than a pawn for  *
 *  that side.                                              *
 *                                                          *
 *  We review standard null-move pruning and introduce our  *
 *  extended version of it, which we call verified          *
 *  null-move pruning.  In verified null-move pruning,      *
 *  whenever the shallow null-move search indicates a       *
 *  fail-high, instead of cutting off the search from the   *
 *  current node, the search is continued with reduced      *
 *  depth.  Our experiments with verified null-move pruning *
 *  show that on average, it constructs a smaller search    *
 *  tree with greater tactical strength in comparison to    *
 *  standard null-move pruning.  Moreover, unlike standard  *
 *  null-move pruning, which fails badly in zugzwang        *
 *  positions, verified null-move pruning manages to detect *
 *  most zugzwangs and in such cases conducts a re-search   *
 *  to obtain the correct result.  In addition, verified    *
 *  null-move pruning is very easy to implement, and any    *
 *  standard null-move pruning program can use verified     *
 *  null-move pruning by modifying only a few lines of      *
 *  code. - Omid David-Tabibi, Nathan S Netanyahu           *
 *                                                          *
 ************************************************************
 */
  if (do_null && ply > 1 && node_type != PV_NODE && depth >= (Resourceful(wtm) ? 2 : 3) &&
    !tree->inchk[ply] && TotalPieces(wtm, occupied) && NotMate(alpha, beta)) {
    if (!Zugzwang(tree, wtm)) {
      register UINT64 save_hash_key;
      register int R = (depth >= 6) ? reduce_null_depth[ub] : reduce_null_depth[lb];
      int hash_move = tree->hash_move[ply];
      tree->curmv[ply] = NULL_MOVE;
      tree->phase[ply] = VOID_MOVE;
#if defined(TRACE)
      if (ply <= trace_level)
        Trace(tree, ply, depth, wtm, beta - 1, beta, "Search1", 0);
#endif
      tree->position[ply + 1] = tree->position[ply];
      Rule50Moves(ply + 1) = 0;
      save_hash_key = HashKey;
      if (EnPassant(ply)) {
        HashEP(EnPassant(ply + 1));
        EnPassant(ply + 1) = 0;
      }
      if (depth - R - 1 > 0)
        value =
            -Search(tree, -beta, -beta + 1, Flip(wtm), depth - R - 1,
            ply + 1, NO_NULL);
      else
        value = -Quiesce(tree, -beta, -beta + 1, Flip(wtm), ply + 1,
          quiesce_check_depth, NO_DELTA);
      if (abort_search || tree->stop)
        return (0);
      HashKey = save_hash_key;
      if (value > NOT_MATE) /* do not return unproven mate scores */
        value = beta;
      if (value >= beta) {
        if (depth >= 8) { /* do verification search at higher depths */
          if (depth - R > 0)
            value = Search(tree, beta - 1, beta, wtm, depth - R, ply,
              DO_NULL);
          else
            value = Quiesce(tree, beta - 1, beta, wtm, ply,
              quiesce_check_depth, NO_DELTA);
          if (abort_search || tree->stop)
            return (0);
        }
        if (value >= beta) {
          HashStore(tree, ply, depth, wtm, LOWER, value, hash_move);
          return (value);
        }
      }
      tree->next_status[ply].phase = HASH_MOVE;
      tree->hash_move[ply] = hash_move;
    }
  }
/*
 ************************************************************
 *                                                          *
 *   ProbCut is a selective search enhancement of the       *
 *   alpha-beta algorithm created in 1994 by Michael Buro   *
 *   as introduced in his PhD thesis.  It permits to        *
 *   exclude probably irrelevant subtrees from being        *
 *   searched deeply.  ProbCut and its improved variant     *
 *   Multi-ProbCut (MPC) have been shown to be effective in *
 *   Othello and Shogi, and a technique similar to ProbCut  *
 *   is used in the checkers program Chinook.  Schaeffer et *
 *   al. (1992) described their approach in a footnote:     *
 *   "Chinook performs forward cuts in positions with a     *
 *   material deficit, where a shallow search does not      *
 *   show an escape."  ProbCut is a generalization of this  *
 *   method in that it is game independent.  It was tested  *
 *   by incorporating it in Buro's already strong Othello   *
 *   program Logistello and increased the program's playing *
 *   strength.  Despite some promising results reported by  *
 *   Albert Xin Jiang and Michael Buro with Crafty, it      *
 *   seemed not that successful in chess programs which     *
 *   already perform Null Move Pruning and Late Move        *
 *   Reductions, until Stockfish proved otherwise as        *
 *   implemented by Gary Linscott.                          *
 *                                                          *
 ************************************************************
 */
  if (do_probcut && ply > 1 && depth >= 6 && node_type != PV_NODE &&
    !tree->inchk[ply] && NotMate(alpha, beta)) {
    int hash_move = tree->hash_move[ply];
    register int R = (depth >= 6) ? reduce_depth[ub] : reduce_depth[lb];
    register int r_beta = Min(beta + probcut_margin, INFTY);
    tree->next_status[ply].phase = HASH_MOVE;
    if (depth - R > 0)
      value = Search(tree, r_beta - 1, r_beta, wtm, depth - R, ply,
        DO_NULL);
    else
      value = Quiesce(tree, r_beta - 1, r_beta, wtm, ply,
        quiesce_check_depth, NO_DELTA);
    if (abort_search || tree->stop)
      return (0);
    if (value >= r_beta) {
      HashStore(tree, ply, depth, wtm, LOWER, value, hash_move);
      return (value);
    }
    tree->next_status[ply].phase = HASH_MOVE;
    tree->hash_move[ply] = hash_move;
  }
/*
 ************************************************************
 *                                                          *
 *   if there is no best move from the hash table, and this *
 *   is a PV node, then we need a good move to search       *
 *   first.  while killers and history moves are good, they *
 *   are not "good enough".  the simplest action is to try  *
 *   a shallow search depth - 3 to get a move.  note that   *
 *   when we call Search() with depth - 3, it, too, will    *
 *   not have a hash move, and will therefore recursively   *
 *   continue this process, hence the name "internal        *
 *   iterative deepening".                                  *
 *                                                          *
 ************************************************************
 */
  if (do_internal_deepen && !tree->hash_move[ply] && do_null &&
    !tree->inchk[ply] && node_type == PV_NODE && depth >= 6 && ply > 1) {
    register int R = (depth >= 6) ? reduce_depth[ub] : reduce_depth[lb];
    tree->next_status[ply].phase = HASH_MOVE;
    Lock(lock_root);
    if (alpha == ((ply & 1) ? root_alpha : -root_beta)
        && beta == ((ply & 1) ? root_beta : -root_alpha)) {
      Unlock(lock_root);
      do {
        tree->curmv[ply] = NULL_MOVE;
        if (depth - R > 0)
          value = Search(tree, alpha, beta, wtm, depth - R, ply,
            DO_NULL);
        else
          value = Quiesce(tree, alpha, beta, wtm, ply,
            quiesce_check_depth, NO_DELTA);
        if (abort_search || tree->stop)
          return (0);
        if (value > alpha) {
          if (value < beta) {
            if ((int) tree->pv[ply - 1].pathl > ply)
              tree->hash_move[ply] = tree->pv[ply - 1].path[ply];
          } else {
            tree->hash_move[ply] = tree->curmv[ply];
          }
          tree->last[ply] = tree->last[ply - 1];
        }
      } while (0);
    } else {
      Unlock(lock_root);
    }
    tree->next_status[ply].phase = HASH_MOVE;
  }
/*
 ************************************************************
 *                                                          *
 *   now iterate through the move list and search the       *
 *   resulting positions.  note that Search() culls any     *
 *   move that is not legal by using Check().  the special  *
 *   case is that we must find one legal move to search to  *
 *   confirm that it's not a mate or draw.                  *
 *                                                          *
 *   first step is to see if we need to extend this move    *
 *   for some tactical reason.  If not, we check to see if  *
 *   we can reduce the depth (LMR) to save time.  A final   *
 *   case is to determine if we can use AEL pruning         *
 *   (Heinz 2000) near the search frontier.  Note for those *
 *   that have read Heinz's paper.  Frontier nodes in       *
 *   crafty have a depth = 2.  pre-frontier nodes  have a   *
 *   depth = 3.  And finally, pre-pre-frontier nodes have a *
 *   depth = 4.                                             *
 *                                                          *
 ************************************************************
 */
  tree->next_status[ply].phase = HASH_MOVE;
  while (1) {
    tree->phase[ply] =
          (tree->inchk[ply]) ? NextEvasion(tree, ply, wtm) :
          NextMove(tree, alpha, beta, ply, wtm, depth,
          node_type, &best_capture);
    if (tree->phase[ply] == NONE)
      break;
    node_type = NodeType(ply);
#if defined(TRACE)
    if (ply <= trace_level)
      Trace(tree, ply, depth, wtm, alpha, t_beta, "Search3", tree->phase[ply]);
#endif
    MakeMove(tree, ply, tree->curmv[ply], wtm);
    if (!Check(wtm)) /* if legal */
      do { /* do search */
        if (Check(Flip(wtm))) /* if check */
          tree->inchk[ply + 1] = 1;
        else
          tree->inchk[ply + 1] = 0;
        searched[moves_searched] = tree->curmv[ply];
        if (!moves_searched)
          first_tried = tree->curmv[ply];
        tree->nodes_searched++;
 /*
  ************************************************************
  *                                                          *
  *   Many programs extend certain moves to try and find     *
  *   better moves faster, or to resolve tactical "noise"    *
  *   resulting from the horizon effect. To extend a move,   *
  *   its search depth (draft) is incremented by some        *
  *   amount, typically one ply.                             *
  *                                                          *
  ************************************************************
  */
        extension = SearchExtension(tree, wtm, ply, depth,
          moves_searched, node_type, &hash);
/*
 ************************************************************
 *                                                          *
 *   Now for the forward-pruning stuff.  The idea here is   *
 *   based on the old FUTILITY idea, where if the current   *
 *   material + a fudge factor is lower than alpha, then    *
 *   there is little promise in searching this move to make *
 *   up for the material deficit.                           *
 *                                                          *
 *   This is a useful idea in today's 20+ ply searches, as  *
 *   when we near the tips, if we are too far behind in     *
 *   material, there is little time left to recover it and  *
 *   moves that don't bring us closer to a reasonable       *
 *   material balance can safely be skipped.  It is much    *
 *   more dangerous in shallow searches.                    *
 *                                                          *
 *   We have an array of pruning margin values that are     *
 *   indexed by depth (remaining plies left until we drop   *
 *   into the quiescence search) and which increase with    *
 *   depth since more depth means a greater chance of       *
 *   bringing the score back up to alpha or beyond.  If the *
 *   current material + the bonus is less than alpha, we    *
 *   simply avoid searching this move at all, and skip to   *
 *   the next move without expending any more effort.  Note *
 *   that this is classic forward-pruning and can certainly *
 *   introduce errors into the search.  However, cluster    *
 *   testing has shown that this improves play in real      *
 *   games.  The current implementation only prunes in the  *
 *   last 4 plies before quiescence, although this can be   *
 *   tuned with the "eval" command changing the "pruning    *
 *   depth" value to something other than 5 (test is for    *
 *   depth < pruning depth, current value is 5 which prunes *
 *   in last 4 plies only).  Testing shows no benefit in    *
 *   larger values than 5, although this might change in    *
 *   future versions as other things are modified.          *
 *                                                          *
 *   Exception:                                             *
 *                                                          *
 *     We do not prune if we are safely pushing a passed    *
 *     pawn and there is enough depth remaining that the    *
 *     pawn might make significant progress advancing.      *
 *                                                          *
 *   forward pruning techniques are either applied at       *
 *   expected Cut-nodes or All-nodes and return either beta *
 *   as lower bound or alpha as upper bound.  the none      *
 *   recursive ProbCut is applied at both types of none     *
 *   PV-nodes.  rather than immediately returning alpha at  *
 *   strong expected All-nodes after a reduced or           *
 *   quiescence search, pruning techniques near the horizon *
 *   skip moves, which are very unlikely to exceed alpha,   *
 *   or in case of the quiescence search itself, all (most) *
 *   quiet moves as well as losing captures.                *
 *                                                          *
 ************************************************************
 */
        if (do_futil) {
          if (ply > 1 && DoFutility(ply) && !tree->inchk[ply] && !tree->inchk[ply + 1] &&
            moves_searched && !Promote(tree->curmv[ply]) && NotMate(alpha, beta)) {
            if (depth + extension < pruning_depth) {
              if (MaterialSTM(wtm) + pruning_margin[depth + extension] <= alpha) {
                if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
                  rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
                  t_beta = alpha + 1;
                  moves_searched++;
                  break; /* skip move to be forward pruned */
                }
              }
            }
/*
 ************************************************************
 *                                                          *
 *   LMP.  Further attempt at forward pruning based on the  *
 *   "late move pruning" idea (a take-off on LMR).          *
 *                                                          *
 *   The basic idea here is that once we have searched a    *
 *   significant number of moves at a ply, it becomes less  *
 *   and less likely that any of the moves left will        *
 *   produce a cutoff.  If the move appears to be simple    *
 *   (not a check, etc.) then we simply skip it, once the   *
 *   move count has been satisfied.  At present, we only    *
 *   do this in the last two plies although this might be   *
 *   changed in the future.                                 *
 *                                                          *
 ************************************************************
 */
            if (ply > 1 && DoLateFutility(ply) && !CaptureOrPromote(tree->curmv[ply])) {
              if (node_type == ALL_NODE) {
                if (depth < movecnt_depth) {
                  if (moves_searched > Moves(depth)) {
                    if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
                      rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
                      t_beta = alpha + 1;
                      moves_searched++;
                      break; /* skip move to be forward pruned */
                    }
                  }
                }
              }
            }
          }
        }
 /*
  ************************************************************
  *                                                          *
  *   Late Move Reductions (LMR) is a non-game specific      *
  *   enhancement to the alpha-beta algorithm and its        *
  *   variants which attempts to examine a game search tree  *
  *   more efficiently.  It uses the assumption that good    *
  *   game-specific move ordering causes a program to search *
  *   the most likely moves early.  If a cut-off is going to *
  *   happen in a search, the first few moves are the ones   *
  *   most likely to cause them.  In games like chess, most  *
  *   programs search winning captures and "killers" first.  *
  *   LMR will reduce the search depth for moves searched    *
  *   later at a given node.  This allows the program to     *
  *   search deeper along the critical lines, and play       *
  *   better.                                                *
  *                                                          *
  ************************************************************
  */
        if ((search_pruning_flags & DO_REDUCTIONS) && DoReduction(ply) && !extension)
          extension -= SearchReduction(tree, wtm, ply, depth, moves_searched,
            alpha, beta, node_type);
/*
 ************************************************************
 *                                                          *
 *   We have determined whether the depth is to be changed  *
 *   by an extension or a reduction.  If we get to this     *
 *   point, then the move is not being pruned.  So off we   *
 *   go to a recursive search/quiescence call to work our   *
 *   way toward a terminal node.                            *
 *                                                          *
 *   There are a couple of special-cases to handle.  If the *
 *   depth was reduced, and Search() returns a value >=     *
 *   beta, accepting that is risky (we reduced the move as  *
 *   we thought it was bad and expected it to fail low) so  *
 *   we repeat the search using the original (non-reduced)  *
 *   depth to see if the fail-high happens again.           *
 *                                                          *
 *   The other special-case is a result of the PVS idea and *
 *   is again a result of a fail-high.  Since we often      *
 *   narrow the window in PVS, if we narrow it at this ply  *
 *   and get a fail-high, we have to open it back up and    *
 *   repeat the search to see if it really fails high.  In  *
 *   most searches this is not done since we almost always  *
 *   reach this point with alpha = beta-1 so that there is  *
 *   no widening required.                                  *
 *                                                          *
 ************************************************************
 */
        if (depth + extension - 1 > 0) {
          value =
              -Search(tree, -t_beta, -alpha, Flip(wtm), depth + extension - 1,
              ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm), depth - 1,
                ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        } else {
          value = -Quiesce(tree, -t_beta, -alpha, Flip(wtm), ply + 1,
            quiesce_check_depth, DoDelta(ply));
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0 && depth - 1 > 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm), depth - 1,
                ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        }
/*
 ************************************************************
 *                                                          *
 *   This is the PVS re-search code.  If we reach this      *
 *   point and value > alpha and value < beta, then this    *
 *   can not be a null-window search.  We have to re-search *
 *   the position with the original beta value (not alpha+1 *
 *   as is the usual case in PVS) to see if it still fails  *
 *   high before we treat this as a real fail-high and back *
 *   up the value to the previous ply.                      *
 *                                                          *
 ************************************************************
 */
        if (value > alpha && value < beta && t_beta < beta) {
          node_type = PV_NODE;
          extension = Max(extension, 0);
          if (depth + extension - 1 > 0) {
            value =
                -Search(tree, -beta, -alpha, Flip(wtm), depth + extension - 1,
                ply + 1, DO_NULL);
          }
          else {
            value = -Quiesce(tree, -beta, -alpha, Flip(wtm), ply + 1,
              quiesce_check_depth, DoDelta(ply));
          }
          if (abort_search || tree->stop)
            break;
        }
/*
 ************************************************************
 *                                                          *
 *   Search (and/or re-search) has been completed.  Now we  *
 *   check for a fail high which terminates the search      *
 *   immediately as no further searching is required.       *
 *                                                          *
 *   Subtle code:  If ply == 1, we call Output() which will *
 *   dump the new PV.  But it also backs up the PV to ply=0 *
 *   which tells us which move to make.  We often time-out  *
 *   in the last iteration of a search and don't complete   *
 *   it.  We can't back up partial results, since they are  *
 *   not accurate, but when we dump a PV, it is from a full *
 *   search and it is safe to back it up to the root PV.    *
 *                                                          *
 ************************************************************
 */
        t_beta = alpha + 1;
        moves_searched++;
        if (value > alpha) {
          if (value >= beta) { /* beta cut-off */
            Killer(tree, ply, tree->curmv[ply]);
            History(tree, ply, depth, wtm, tree->curmv[ply], searched,
              moves_searched);
            HashStore(tree, ply, depth, wtm, LOWER, value, tree->curmv[ply]);
            UnmakeMove(tree, ply, tree->curmv[ply], wtm);
            tree->fail_high++;
            if (moves_searched == 1)
              tree->fail_high_first++;
            return (value);
          }
          alpha = value;
          t_beta = alpha + 1;
        }
      } while (0); /* end do search */
    UnmakeMove(tree, ply, tree->curmv[ply], wtm);
    if (abort_search || tree->stop)
      return (0);
/*
 ************************************************************
 *                                                          *
 *   if this is an SMP search, and we have idle processors, *
 *   now is the time to get them involved.  we have now     *
 *   satisfied the "young brothers wait" condition since we *
 *   have searched one move.  All that is left is to check  *
 *   the remaining depth so that we do not split too near   *
 *   the tips.                                              *
 *                                                          *
 *   In CB, I used this extensively, because you never      *
 *   want to split at a CUT node, and want to prefer an ALL *
 *   node.  YBW tries to recognize an ALL node by requiring *
 *   that at least one move be searched before splitting,   *
 *   since > 90% of CUT nodes are discovered on the first   *
 *   move searched.  But waiting on that first move         *
 *   introduces some wait time, and if you knew it was an   *
 *   ALL node from the get-go, you could split before the   *
 *   first move is completed.  That was the basic premise   *
 *   behind the iterated search and DTS algorithm in Cray   *
 *   Blitz...                                               *
 *                                                          *
 ************************************************************
 */
#if (CPUS > 1)
    Lock(lock_smp);
    if (smp_idle && moves_searched > 0 &&
      tree->nodes_searched - start_nodes > 2000 &&
      node_type != CUT_NODE && depth >= 5) {
      Unlock(lock_smp);
      do {
        tree->alpha = alpha;
        tree->beta = beta;
        tree->value = alpha;
        tree->wtm = wtm;
        tree->ply = ply;
        tree->depth = depth;
        tree->searched = searched;
        tree->moves_searched = moves_searched;
        tree->best_capture = best_capture;
        tree->node_type = node_type;
        if (Thread(tree)) {
          if (abort_search || tree->stop)
            return (0);
          value = tree->search_value;
          moves_searched = tree->moves_searched;
          best_capture = tree->best_capture;
          node_type = tree->node_type;
          if (value > alpha) {
            if (value >= beta) { /* beta cut-off */
              Killer(tree, ply, tree->cutmove);
              History(tree, ply, depth, wtm, tree->cutmove, searched,
                moves_searched);
              HashStore(tree, ply, depth, wtm, LOWER, value, tree->cutmove);
              return (value);
            }
            alpha = value;
            t_beta = alpha + 1;
            break;
          }
        }
      } while (0);
    } else
      Unlock(lock_smp);
#endif
  }
/*
 ************************************************************
 *                                                          *
 *   all moves have been searched.  if none were legal,     *
 *   return either MATE or DRAW depending on whether the    *
 *   side to move is in check or not.                       *
 *                                                          *
 ************************************************************
 */
  if (abort_search || tree->stop)
    return (0);
  if (moves_searched == 0) {
    value = (Check(wtm)) ? -(MATE - ply) : DrawScore(wtm);
    if (value >= alpha && value < beta) {
      SavePV(tree, ply, 0);
#if defined(TRACE)
      if (ply <= trace_level)
        printf("Search() no moves!  ply=%d\n", ply);
#endif
      }
      return (value);
  } else {
    if (repeat == 2 && alpha != -(MATE - ply - 1)) {
      value = DrawScore(wtm);
      if (value < beta)
        SavePV(tree, ply, 0);
#if defined(TRACE)
      if (ply <= trace_level)
        printf("draw by repetition detected, ply=%d.\n", ply);
#endif
      return (value);
    } else if (alpha != o_alpha) {
      tree->pv[ply - 1] = tree->pv[ply];
      tree->pv[ply - 1].path[ply - 1] = tree->curmv[ply - 1];
    }
    HashStore(tree, ply, depth, wtm, (alpha == o_alpha) ? UPPER : EXACT,
      alpha, (alpha == o_alpha) ? first_tried : tree->pv[ply].path[ply]);
    return (alpha);
  }
}

/* last modified 05/15/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   SearchExtension() is used to adjust the search depth for the tree below   *
 *   this move.  checking moves are problematic as they sometimes lead to a    *
 *   deep win of material or a checkmate, or sometimes they are used to delay  *
 *   something (push it beyond the search horizon) so that it is not even      *
 *   found by the search.                                                      *
 *                                                                             *
 *******************************************************************************
 */
#if defined(_WIN32) || defined(_WIN64)
__forceinline
#endif
int SearchExtension(TREE * RESTRICT tree, int wtm, int ply, int depth,
  int moves_searched, NODE_TYPE node_type, HASH *hash)
{
  register int adjustment = 0;
  if (ply + depth < MAXPLY - MAXPLY_ULLAGE / 2) {
    do { /* extend */
      register int loss = NO_VALUE;
/*
 ************************************************************
 *                                                          *
 *   if the move is a capture restoring material balance of *
 *   the root entering into a pawn endgame, then we may     *
 *   wish to extend the depth by one to three plies in      *
 *   order to search deeper.  (neutral)                     *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_CAPTURE_ENDGAME]) {
        if (Captured(tree->curmv[ply]) > pawn &&
          TotalPieces(wtm, occupied) == 0 &&
          TotalPieces(Flip(wtm), occupied) == 0) {
          if (MaterialBalanceRoot(pawn_value)) {
            adjustment = extension_depth[EXTENSION_CAPTURE_ENDGAME];
            adjustment = Max(Min(adjustment, 3), 0);
            if (adjustment >= 3)
              break;
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is castling in materially balanced         *
 *   position with the root, then we may wish to extend     *
 *   the depth by one to three plies in order to search     *
 *   deeper.  (neutral)                                     *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_CASTLING]) {
        register int btm = Flip(wtm);
        if (Castling(tree->curmv[ply]) && FileDistance(KingSQ(wtm),
          KingSQ(btm)) > 1) {
          if (node_type == PV_NODE && tree->phase[ply] < LATE_MOVES) {
            if (MaterialBalanceRoot(pawn_value)) {
              adjustment = extension_depth[EXTENSION_CASTLING];
              adjustment = Max(Min(adjustment, 3), 0);
              if (adjustment >= 2)
                break;
            }
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if this is a passed pawn push or a pawn capture which  *
 *   creates a passed pawn in a position which is           *
 *   materially balance with the root, then we may wish to  *
 *   extend the depth by one to two plies in order to       *
 *   search deeper.  the extension is done only if the      *
 *   pawn moved on the 6th/3rd or 7th/2nd rank.             *
 *   (win-seeking)                                          *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_PASSED_PAWN]) {
        if (Piece(tree->curmv[ply]) == pawn) {
          if (Passed(To(tree->curmv[ply]), wtm)) {
            if (Rank(To(tree->curmv[ply])) == rankflip[wtm][RANK7] ||
              Rank(To(tree->curmv[ply])) == rankflip[wtm][RANK6]) {
              if (MaterialBalanceRoot(pawn_value)) {
                if (loss == NO_VALUE) {
                  register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                  loss = 0;
                  if (attacks & Occupied(Flip(wtm))) {
                    loss = SwapO(tree, tree->curmv[ply], wtm) -
                      pc_values[Captured(tree->curmv[ply])];
                  }
                }
                if (loss <= 0) { /* no material loss */
                  adjustment = extension_depth[EXTENSION_PASSED_PAWN];
                  adjustment = Max(Min(adjustment, 2), 0);
                  if (adjustment >= 1)
                    break;
                }
              }
            }
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move checks the opponent, then we may wish to   *
 *   extend the depth by one ply in order to search deeper. *
 *   however, some programmers don't extend checks (and     *
 *   captures) with negative SEE or even reduce them.       *
 *   Robert Hyatt claimed a significant gain in Crafty by   *
 *   doing so.  (neutral)                                   *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_CHECK]) {
        if (tree->inchk[ply + 1]) { /* if check */
          if (loss == NO_VALUE) {
            register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
            loss = 0;
            if (attacks & Occupied(Flip(wtm))) {
              loss = SwapO(tree, tree->curmv[ply], wtm) -
                pc_values[Captured(tree->curmv[ply])];
            }
          }
          if (loss <= 0 || DoubleCheck(tree, wtm) || node_type == PV_NODE) { /* no material loss or double check or PV-node */
            adjustment = extension_depth[EXTENSION_CHECK];
            adjustment = Max(Min(adjustment, 1), 0);
            break;
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is evading a check and there is only one   *
 *   legal move, then we may wish to extend the depth by    *
 *   one ply in order to search deeper.  (loss-seeking)     *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_ONE_EVASION]) {
        if (ply > 2 && !tree->inchk[ply - 2] && tree->inchk[ply] &&
          tree->last[ply] - tree->last[ply - 1] == 1) {
            adjustment = extension_depth[EXTENSION_ONE_EVASION];
            adjustment = Max(Min(adjustment, 1), 0);
            break;
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is a pawn recapture restoring material     *
 *   balance of the root, then we may wish to extend the    *
 *   depth by one ply in order to search deeper.  (neutral) *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_PAWN_RECAPTURE]) {
        if (ply > 1 && Captured(tree->curmv[ply]) == pawn &&
          Captured(tree->curmv[ply - 1]) == pawn &&
          To(tree->curmv[ply]) == To(tree->curmv[ply - 1])) {
          if (node_type == PV_NODE && tree->phase[ply] < LATE_MOVES) {
            if (MaterialBalanceRoot(pawn_value)) {
              if (loss == NO_VALUE) {
                register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                loss = 0;
                if (attacks & Occupied(Flip(wtm))) {
                  loss = SwapO(tree, tree->curmv[ply], wtm) -
                    pc_values[Captured(tree->curmv[ply])];
                }
              }
              if (loss <= 0) { /* no material loss */
                adjustment = extension_depth[EXTENSION_PAWN_RECAPTURE];
                adjustment = Max(Min(adjustment, 1), 0);
                break;
              }
            }
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is a recapture restoring material balance  *
 *   of the root, then we may wish to extend the depth by   *
 *   one ply in order to search deeper.  (neutral)          *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_RECAPTURE]) {
        if (ply > 1  && Captured(tree->curmv[ply]) && Captured(tree->curmv[ply - 1]) &&
          Abs(pc_values[Captured(tree->curmv[ply])] -
          pc_values[Captured(tree->curmv[ply - 1])]) < pawn_value) {
          if (node_type == PV_NODE && tree->phase[ply] < LATE_MOVES) {
            if (MaterialBalanceRoot(pawn_value)) {
              if (loss == NO_VALUE) {
                register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                loss = 0;
                if (attacks & Occupied(Flip(wtm))) {
                  loss = SwapO(tree, tree->curmv[ply], wtm) -
                    pc_values[Captured(tree->curmv[ply])];
                }
              }
              if (loss <= 0) { /* no material loss */
                adjustment = extension_depth[EXTENSION_RECAPTURE];
                adjustment = Max(Min(adjustment, 1), 0);
                break;
              }
            }
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is a result of kicking the piece, in a     *
 *   position which is materially balanced with the root,   *
 *   then we may wish to extend the depth by one ply in     *
 *   order to search deeper.  (neutral)                     *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_KICK]) {
        if (ply > 1 && Piece(tree->curmv[ply]) != pawn &&
          pc_values[Piece(tree->curmv[ply])] -
          pc_values[Piece(tree->curmv[ply - 1])] > pawn_value) {
          if (node_type == PV_NODE && tree->phase[ply] < LATE_MOVES) {
            if (MaterialBalanceRoot(pawn_value)) {
              register BITBOARD attacks = AttacksTo(tree, From(tree->curmv[ply]));
              if (attacks & Pieces(Flip(wtm), Piece(tree->curmv[ply - 1]))) {
                if (loss == NO_VALUE) {
                  register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                  loss = 0;
                  if (attacks & Occupied(Flip(wtm))) {
                    loss = SwapO(tree, tree->curmv[ply], wtm) -
                      pc_values[Captured(tree->curmv[ply])];
                  }
                }
                if (loss <= 0) { /* no material loss */
                  adjustment = extension_depth[EXTENSION_KICK];
                  adjustment = Max(Min(adjustment, 1), 0);
                  break;
                }
              }
            }
          }
        }
      }
/*
 ************************************************************
 *                                                          *
 *   if the move is a delayed recapture restoring material  *
 *   balance of the root, then we may wish to extend the    *
 *   depth by one ply in order to search deeper.  (neutral) *
 *                                                          *
 ************************************************************
 */
      if (extension_depth[EXTENSION_RECAPTURE]) {
        if (ply > 3 && Captured(tree->curmv[ply]) &&
          Captured(tree->curmv[ply - 3]) && !tree->inchk[ply - 1] &&
          Abs(pc_values[Captured(tree->curmv[ply])] -
          pc_values[Captured(tree->curmv[ply - 3])]) < pawn_value) {
          if (node_type == PV_NODE && tree->phase[ply] < LATE_MOVES) {
            if (MaterialBalanceRoot(pawn_value)) {
              if (loss == NO_VALUE) {
                register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
                loss = 0;
                if (attacks & Occupied(Flip(wtm))) {
                  loss = SwapO(tree, tree->curmv[ply], wtm) -
                    pc_values[Captured(tree->curmv[ply])];
                }
              }
              if (loss <= 0) { /* no material loss */
                adjustment = extension_depth[EXTENSION_RECAPTURE];
                adjustment = Max(Min(adjustment, 1), 0);
                break;
              }
            }
          }
        }
      }
    } while (0);
    if (adjustment > 0)
      tree->extensions_done++;
  }
  return (adjustment);
}

/* last modified 04/24/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   SearchReduction() is used to adjust the search depth for the tree below   *
 *   this move.                                                                *
 *                                                                             *
 *   some moves do not deserve as much effort as others, and we try to         *
 *   recognize that here and rather than extending the search depth (and       *
 *   effort) we actually reduce the search depth to save time that can be      *
 *   better spent on other moves.  any move is tactical in nature (captures,   *
 *   checks, passed pawn pushes, and so forth) are not reduced.  normal moves  *
 *   that appear to be "good" are also not reduced, including the hash table   *
 *   "best move", killer moves, and other such moves.  that leaves moves that  *
 *   have not been identified as "good" to be searched with reduced depth.     *
 *   the final criterion here is that the "history value" must lie below some  *
 *   threshold indicating that this move has never been very good in the       *
 *   current search.                                                           *
 *                                                                             *
 *******************************************************************************
 */
#if defined(_WIN32) || defined(_WIN64)
__forceinline
#endif
int SearchReduction(TREE * RESTRICT tree, int wtm, int ply, int depth,
  int moves_searched, int alpha, int beta, NODE_TYPE node_type)
{
  register int adjustment = 0;
/*
 ************************************************************
 *                                                          *
 *   Late Move Reductions (LMR) is a non-game specific      *
 *   enhancement to the alpha-beta algorithm and its        *
 *   variants which attempts to examine a game search tree  *
 *   more efficiently.  It uses the assumption that good    *
 *   game-specific move ordering causes a program to search *
 *   the most likely moves early.  If a cut-off is going to *
 *   happen in a search, the first few moves are the ones   *
 *   most likely to cause them.  In games like chess, most  *
 *   programs search winning captures and "killers" first.  *
 *   LMR will reduce the search depth for moves searched    *
 *   later at a given node.  This allows the program to     *
 *   search deeper along the critical lines, and play       *
 *   better.                                                *
 *                                                          *
 ************************************************************
 */
/*
 ************************************************************
 *                                                          *
 *   Now it's time to try to reduce the search depth if the *
 *   move appears to be "poor".  To reduce the search, the  *
 *   following requirements must be met:                    *
 *                                                          *
 *   (1) We must be in the BAD_CAPTURE_MOVES part of the    *
 *       move ordering, so that we have nearly given up on  *
 *       failing high on any move.                          *
 *   (2) We must not be too close to the horizon (this is   *
 *       the LMR_min_depth value).                          *
 *   (3) The current move must not be a checking move and   *
 *       the side to move can not be in check.              *
 *   (4) The moving piece is not a passed pawn moving to a  *
 *       rank greater than 4.                               *
 *   (5) The current move can not affect the material       *
 *       balance, that is it can not be a capture or        *
 *       a pawn promotion.                                  *
 *                                                          *
 *   Most programs do not reduce these types of moves:      *
 *   . Tactical moves (captures and promotions)             *
 *   . Moves while in check                                 *
 *   . Moves which give check                               *
 *   . Moves that cause a search extension                  *
 *   . Anytime in a PV-Node in a PVS search                 *
 *   . Depth < 3 (sometimes depth < 2)                      *
 *                                                          *
 *   Less common conditions on moves not to reduce:         *
 *   . Passed Pawn Moves                                    *
 *   . Killer Moves                                         *
 *   . Moves threatening the King area                      *
 *   . Tactically threatening moves                         *
 *   . Moves with good past relative history                *
 *   . Any Pawn Moves                                       *
 *   . Allowing reductions of "bad" captures (SEE < 0)      *
 *   . Moves of a threatened piece to safety (often         *
 *     detected via a Null Move search)                     *
 *                                                          *
 ************************************************************
 */
  if (search_pruning_flags & DO_REDUCTIONS) {
    if (DoReduction(ply) && moves_searched > 0 && !tree->inchk[ply] && !tree->inchk[ply + 1] &&
      !CaptureOrPromote(tree->curmv[ply]) && !Castling((tree->curmv[ply])) && node_type != CUT_NODE) {
      if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
        rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
        register int min_depth = ((node_type != PV_NODE || tree->phase[ply] >= LATE_MOVES) ?
          LMR_min_depth[lb] : LMR_min_depth[ub]);
        if (depth - 1 - min_depth > 0) {
          adjustment = Max(Min(depth - 1 - min_depth, Reduction(node_type,
            tree->phase[ply] >= LATE_MOVES, depth, moves_searched)), 0);
        }
      }
      if (adjustment)
        tree->reductions_done++;
    }
  }
  return (adjustment);
}

