#include "chess.h"
#include "data.h"
#include "epdglue.h"
/* modified 05/15/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   SearchParallel() is the recursive routine used to implement alpha/beta    *
 *   negamax search using parallel threads.  when this code is called, the     *
 *   first move has already been searched, so all that is left is to search    *
 *   the remainder of the moves and then return.  note that the hash table and *
 *   such can't be modified here since this only represents a part of the      *
 *   search at this ply.                                                       *
 *                                                                             *
 *******************************************************************************
 */
int SearchParallel(TREE * RESTRICT tree, int alpha, int beta, int value,
    int wtm, int depth, int ply)
{
  register int t_beta = (tree->parent->moves_searched) ? alpha + 1 : beta;
  register int extension = 0;
  register int do_futil = 1;
  UINT64 begin_root_nodes;
  NODE_TYPE node_type = tree->parent->node_type;

  if (ply >= MAXPLY - 1)
    return (beta);
  if (!(search_pruning_flags & DO_FUTILITY) ||
    pruning_margin[depth] == 0)
    do_futil = 0;
/*
 ************************************************************
 *                                                          *
 *   now iterate through the move list and search the       *
 *   resulting positions.  note that Search() culls any     *
 *   move that is not legal by using Check().  the special  *
 *   case is that we must find one legal move to search to  *
 *   confirm that it's not a mate or draw.                  *
 *                                                          *
 *   first step is to see if we need to extend this move    *
 *   for some tactical reason.  If not, we check to see if  *
 *   we can reduce the depth (LMR) to save time.  A final   *
 *   case is to determine if we can use AEL pruning         *
 *   (Heinz 2000) near the search frontier.  Note for those *
 *   that have read Heinz's paper.  Frontier nodes in       *
 *   crafty have a depth = 2.  pre-frontier nodes  have a   *
 *   depth = 3.  And finally, pre-pre-frontier nodes have a *
 *   depth = 4.                                             *
 *                                                          *
 ************************************************************
 */
  while (1) {
    Lock(tree->parent->lock);
    if (ply == 1) {
      tree->phase[ply] = NextRootMove(tree->parent, tree, wtm,
        &tree->parent->best_capture);
      tree->root_move = tree->parent->root_move;
    } else {
      tree->phase[ply] =
          (tree->inchk[ply]) ? NextEvasion((TREE *) tree->parent, ply,
          wtm) : NextMove((TREE *) tree->parent, alpha, beta, ply, wtm,
          depth, node_type, &tree->parent->best_capture);
    }
    tree->curmv[ply] = tree->parent->curmv[ply];
    Unlock(tree->parent->lock);
    if (tree->phase[ply] == NONE)
      break;
    node_type = tree->parent->node_type = NodeType(ply);
#if defined(TRACE)
    if (ply <= trace_level)
      Trace(tree, ply, depth, wtm, alpha, t_beta, "SearchParallel",
          tree->phase[ply]);
#endif
    MakeMove(tree, ply, tree->curmv[ply], wtm);
    if (!Check(wtm)) /* if legal */
      do { /* do search */
        tree->parent->searched[tree->parent->moves_searched] = tree->curmv[ply];
        if (Check(Flip(wtm))) /* if check */
          tree->inchk[ply + 1] = 1;
        else
          tree->inchk[ply + 1] = 0;
        tree->nodes_searched++;
        begin_root_nodes = tree->nodes_searched;
 /*
  ************************************************************
  *                                                          *
  *   Many programs extend certain moves to try and find     *
  *   better moves faster, or to resolve tactical "noise"    *
  *   resulting from the horizon effect. To extend a move,   *
  *   its search depth (draft) is incremented by some        *
  *   amount, typically one ply.                             *
  *                                                          *
  ************************************************************
  */
        extension = SearchExtension(tree, wtm, ply, depth,
          tree->parent->moves_searched, node_type,
          NULL);
/*
 ************************************************************
 *                                                          *
 *   Now for the forward-pruning stuff.  The idea here is   *
 *   based on the old FUTILITY idea, where if the current   *
 *   material + a fudge factor is lower than alpha, then    *
 *   there is little promise in searching this move to make *
 *   up for the material deficit.                           *
 *                                                          *
 *   We have an array of pruning margin values that are     *
 *   indexed by depth (remaining plies left until we drop   *
 *   into the quiescence search) and which increase with    *
 *   depth since more depth means a greater chance of       *
 *   bringing the score back up to alpha or beyond.  If the *
 *   current material + the bonus is less than alpha, we    *
 *   simply avoid searching this move at all, and skip to   *
 *   the next move without expending any more effort.  Note *
 *   that this is classic forward-pruning and can certainly *
 *   introduce errors into the search.  However, cluster    *
 *   testing has shown that this improves play in real      *
 *   games.  The current implementation only prunes in the  *
 *   last 4 plies before quiescence, although this can be   *
 *   tuned with the "eval" command changing the "pruning    *
 *   depth" value to something other than 5 (test is for    *
 *   depth < pruning depth, current value is 5 which prunes *
 *   in last 4 plies only).  Testing shows no benefit in    *
 *   larger values than 5, although this might change in    *
 *   future versions as other things are modified.          *
 *   We do not prune if we are safely pushing a passed pawn *
 *   and there is enough depth remaining that the pawn      *
 *   might make significant progress advancing.             *
 *                                                          *
 *   forward pruning techniques are either applied at       *
 *   expected Cut-nodes or All-nodes and return either beta *
 *   as lower bound or alpha as upper bound.  the none      *
 *   recursive ProbCut is applied at both types of none     *
 *   PV-nodes.  rather than immediately returning alpha at  *
 *   strong expected All-nodes after a reduced or           *
 *   quiescence search, pruning techniques near the horizon *
 *   skip moves, which are very unlikely to exceed alpha,   *
 *   or in case of the quiescence search itself, all (most) *
 *   quiet moves as well as losing captures.                *
 *                                                          *
 ************************************************************
 */
        if (do_futil) {
          if (ply > 1 && DoFutility(ply) && !tree->inchk[ply] && !tree->inchk[ply + 1] &&
           tree->parent->moves_searched && !Promote(tree->curmv[ply]) && NotMate(alpha, beta)) {
            if (depth + extension < pruning_depth) {
              if (MaterialSTM(wtm) + pruning_margin[depth + extension] <= alpha) {
                if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
                  rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
                  t_beta = alpha + 1;
                  tree->parent->moves_searched++;
                  break; /* skip move to be forward pruned */
                }
              }
            }
/*
 ************************************************************
 *                                                          *
 *   LMP.  Further attempt at forward pruning based on the  *
 *   "late move pruning" idea (a take-off on LMR).          *
 *                                                          *
 *   The basic idea here is that once we have searched a    *
 *   significant number of moves at a ply, it becomes less  *
 *   and less likely that any of the moves left will        *
 *   produce a cutoff.  If the move appears to be simple    *
 *   (not a check, etc.) then we simply skip it, once the   *
 *   move count has been satisfied.  At present, we only    *
 *   do this in the last two plies although this might be   *
 *   changed in the future.                                 *
 *                                                          *
 ************************************************************
 */
            if (ply > 1 && DoLateFutility(ply) && !CaptureOrPromote(tree->curmv[ply])) {
              if (node_type == ALL_NODE) {
                if (depth < movecnt_depth) {
                  if (tree->parent->moves_searched > Moves(depth)) {
                    if (Piece(tree->curmv[ply]) != pawn || !Passed(To(tree->curmv[ply]), wtm) ||
                      rankflip[wtm][Rank(To(tree->curmv[ply]))] < RANK6) {
                      t_beta = alpha + 1;
                      tree->parent->moves_searched++;
                      break; /* skip move to be forward pruned */
                    }
                  }
                }
              }
            }
          }
        }
 /*
  ************************************************************
  *                                                          *
  *   Late Move Reductions (LMR) is a non-game specific      *
  *   enhancement to the alpha-beta algorithm and its        *
  *   variants which attempts to examine a game search tree  *
  *   more efficiently.  It uses the assumption that good    *
  *   game-specific move ordering causes a program to search *
  *   the most likely moves early.  If a cut-off is going to *
  *   happen in a search, the first few moves are the ones   *
  *   most likely to cause them.  In games like chess, most  *
  *   programs search winning captures and "killers" first.  *
  *   LMR will reduce the search depth for moves searched    *
  *   later at a given node.  This allows the program to     *
  *   search deeper along the critical lines, and play       *
  *   better.                                                *
  *                                                          *
  ************************************************************
  */
        if ((search_pruning_flags & DO_REDUCTIONS) && DoReduction(ply) && !extension)
          extension -= SearchReduction(tree, wtm, ply, depth, tree->parent->moves_searched,
            alpha, beta, node_type);
/*
 ************************************************************
 *                                                          *
 *   We have determined whether the depth is to be changed  *
 *   by an extension or a reduction.  If we get to this     *
 *   point, then the move is not being pruned.  So off we   *
 *   go to a recursive search/quiescence call to work our   *
 *   way toward a terminal node.                            *
 *                                                          *
 *   There are a couple of special-cases to handle.  If the *
 *   depth was reduced, and Search() returns a value >=     *
 *   beta, accepting that is risky (we reduced the move as  *
 *   we thought it was bad and expected it to fail low) so  *
 *   we repeat the search using the original (non-reduced)  *
 *   depth to see if the fail-high happens again.           *
 *                                                          *
 *   The other special-case is a result of the PVS idea and *
 *   is again a result of a fail-high.  Since we often      *
 *   narrow the window in PVS, if we narrow it at this ply  *
 *   and get a fail-high, we have to open it back up and    *
 *   repeat the search to see if it really fails high.  In  *
 *   most searches this is not done since we almost always  *
 *   reach this point with alpha = beta-1 so that there is  *
 *   no widening required.                                  *
 *                                                          *
 ************************************************************
 */
        if (depth + extension - 1 > 0) {
          value =
              -Search(tree, -t_beta, -alpha, Flip(wtm), depth + extension - 1, ply + 1,
              DO_NULL);
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm), depth - 1, ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        } else {
          value = -Quiesce(tree, -t_beta, -alpha, Flip(wtm), ply + 1,
            quiesce_check_depth, DoDelta(ply));
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0 && depth - 1 > 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm), depth - 1, ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        }
/*
 ************************************************************
 *                                                          *
 *   This is the PVS re-search code.  If we reach this      *
 *   point and value > alpha and value < beta, then this    *
 *   can not be a null-window search.  We have to re-search *
 *   the position with the original beta value (not alpha+1 *
 *   as is the usual case in PVS) to see if it still fails  *
 *   high before we treat this as a real fail-high and back *
 *   up the value to the previous ply.                      *
 *                                                          *
 ************************************************************
 */
        if (value > alpha && value < beta && t_beta < beta) {
          node_type = tree->parent->node_type = PV_NODE;
          extension = Max(extension, 0);
          if (depth + extension - 1 > 0) {
            value =
                -Search(tree, -beta, -alpha, Flip(wtm), depth + extension - 1,
                ply + 1, DO_NULL);
          }
          else {
            value = -Quiesce(tree, -beta, -alpha, Flip(wtm), ply + 1,
              quiesce_check_depth, DoDelta(ply));
          }
          if (abort_search || tree->stop)
            break;
        }
/*
 ************************************************************
 *                                                          *
 *   Search (and/or re-search) has been completed.  Now we  *
 *   check for a fail high which terminates the search      *
 *   immediately as no further searching is required.       *
 *                                                          *
 *   Subtle code:  If ply == 1, we call Output() which will *
 *   dump the new PV.  But it also backs up the PV to ply=0 *
 *   which tells us which move to make.  We often time-out  *
 *   in the last iteration of a search and don't complete   *
 *   it.  We can't back up partial results, since they are  *
 *   not accurate, but when we dump a PV, it is from a full *
 *   search and it is safe to back it up to the root PV.    *
 *                                                          *
 ************************************************************
 */
        if (ply == 1)
          root_moves[tree->root_move].nodes =
              tree->nodes_searched - begin_root_nodes;
        t_beta = alpha + 1;
        tree->parent->moves_searched++;
        if (value > alpha) {
          if (ply == 1) {
            Lock(lock_root);
            if (value > root_value) {
              Output(tree, value, beta);
              root_value = value;
            }
            Unlock(lock_root);
          }
          if (value >= beta) { /* beta cut-off */
            register int proc;
            parallel_aborts++;
            UnmakeMove(tree, ply, tree->curmv[ply], wtm);
            Lock(lock_smp);
            Lock(tree->parent->lock);
            if (!tree->stop) {
              for (proc = 0; proc < smp_max_threads; proc++)
                if (tree->parent->siblings[proc] && proc != tree->thread_id)
                  ThreadStop(tree->parent->siblings[proc]);
            }
            Unlock(tree->parent->lock);
            Unlock(lock_smp);
            return (value);
          }
          alpha = value;
          t_beta = alpha + 1;
        }
      } while (0); /* end do search */
    UnmakeMove(tree, ply, tree->curmv[ply], wtm);
    if (abort_search || tree->stop)
      break;
  }
/*
 ************************************************************
 *                                                          *
 *   there are no "end-of-search" things to do.  We have    *
 *   searched all the remaining moves at this ply in        *
 *   parallel, and now return and let the original search   *
 *   (that started this sub-tree) clean up, and do the      *
 *   tests for mate/stalemate, update the hash table, etc.  *
 *                                                          *
 *   we do need to flag the root move we tried to search,   *
 *   if we were stopped early due to another root move      *
 *   failing high.  otherwise this move appears to have     *
 *   been searched already and will not be searched again   *
 *   until the next iteration.                              *
 *                                                          *
 ************************************************************
 */
  if (tree->stop && ply == 1)
    root_moves[tree->root_move].status &= 255 - 16;
  return (alpha);
}
