#include "chess.h"
#include "data.h"
#include "epdglue.h"
/* modified 05/14/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   SearchRoot() is the recursive routine used to implement the alpha/beta    *
 *   negamax search (similar to minimax but simpler to code.)  SearchRoot() is *
 *   only called when ply=1.  it is somewhat different from Search() in that   *
 *   some things (null move search, hash lookup, etc.) are not useful at the   *
 *   root of the tree.  SearchRoot() calls Search() to search any positions    *
 *   that are below ply=1.                                                     *
 *                                                                             *
 *******************************************************************************
 */
int SearchRoot(TREE * RESTRICT tree, int alpha, int beta, int wtm, int depth)
{
  register int o_alpha = alpha, t_beta = beta;
  register int moves_searched = 0;
  register int value = alpha;
  register int extension = 0;
  register int first_tried = NULL_MOVE;
  register int ply = 1;
  NODE_TYPE node_type = PV_NODE;
  UINT64 begin_root_nodes;
  int searched[256];
  int best_capture = 0;

  if (Check(wtm)) /* if check */
    tree->inchk[ply] = 1;
  else
    tree->inchk[ply] = 0;
  tree->phase[ply] = NO_MOVE;
/*
 ************************************************************
 *                                                          *
 *   now iterate through the move list and search the       *
 *   resulting positions.  note that SearchRoot() does not  *
 *   search illegal moves since RootMoves() screened each   *
 *   move before adding it to the permanent ply-1 move list *
 *   earlier.                                               *
 *                                                          *
 ************************************************************
 */
  while ((tree->phase[ply] = NextRootMove(tree, tree, wtm, &best_capture))) {
#if defined(TRACE)
    if (trace_level > 0)
      Trace(tree, ply, depth, wtm, alpha, t_beta, "SearchRoot", tree->phase[ply]);
#endif
    MakeMove(tree, ply, tree->curmv[ply], wtm);
    if (!Check(wtm)) /* if legal */
      do { /* do search */
        searched[moves_searched] = tree->curmv[ply];
        if (!moves_searched)
          first_tried = tree->curmv[ply];
        if (Check(Flip(wtm))) /* if check */
          tree->inchk[ply + 1] = 1;
        else
          tree->inchk[ply + 1] = 0;
        tree->nodes_searched++;
 /*
  ************************************************************
  *                                                          *
  *   Many programs extend certain moves to try and find     *
  *   better moves faster, or to resolve tactical "noise"    *
  *   resulting from the horizon effect. To extend a move,   *
  *   its search depth (draft) is incremented by some        *
  *   amount, typically one ply.                             *
  *                                                          *
  ************************************************************
  */
        extension = SearchExtension(tree, wtm, ply, depth,
          moves_searched, node_type, NULL);
 /*
  ************************************************************
  *                                                          *
  *   Late Move Reductions (LMR) is a non-game specific      *
  *   enhancement to the alpha-beta algorithm and its        *
  *   variants which attempts to examine a game search tree  *
  *   more efficiently.  It uses the assumption that good    *
  *   game-specific move ordering causes a program to search *
  *   the most likely moves early.  If a cut-off is going to *
  *   happen in a search, the first few moves are the ones   *
  *   most likely to cause them.  In games like chess, most  *
  *   programs search winning captures and "killers" first.  *
  *   LMR will reduce the search depth for moves searched    *
  *   later at a given node.  This allows the program to     *
  *   search deeper along the critical lines, and play       *
  *   better.                                                *
  *                                                          *
  ************************************************************
  */
        if ((search_pruning_flags & DO_REDUCTIONS) && DoReduction(ply) && !extension)
          extension -= SearchReduction(tree, wtm, ply, depth, moves_searched,
            alpha, beta, node_type);
        begin_root_nodes = tree->nodes_searched;
  /*
   ************************************************************
   *                                                          *
   *   Bruce Moreland's PVS implementation waits until a move *
   *   is found that improves alpha, and then searches every  *
   *   move after that with a zero window around alpha. The   *
   *   alpha improvement usually occurs at the first move,    *
   *   and always at the leftmost nodes (assuming from left   *
   *   to right traversal) with a most open alpha-beta        *
   *   window of +-oo. In re-searches or with aspiration-     *
   *   windows the first moves may rarely not improve alpha.  *
   *   As pointed out by Edmund Moshammer, Gian-Carlo         *
   *   Pascutto, Robert Hyatt and Vincent Diepeveen, it is    *
   *   recommend to only search the first move with an open   *
   *   window, and then every other move after that with a    *
   *   zero window.                                           *
   *                                                          *
   ************************************************************
   */
  /*
   ************************************************************
   *                                                          *
   *   if this is the first move searched at ply=1, we search *
   *   using the normal alpha/beta bounds.  if not, we first  *
   *   search with alpha, alpha+1 (PVS search).  if this      *
   *   search fails high, we re-search using the normal       *
   *   bounds.  if allowed by inter-iteration analysis, some  *
   *   root moves might be searched with a reduced depth (as  *
   *   in the LMR idea).  if a reduced move fails high, we    *
   *   always re-search it with the normal depth before we    *
   *   accept the fail-high as a real one and re-search with  *
   *   the original alpha/beta bounds.  so a single move      *
   *   might fail high on the reduced search, get re-searched *
   *   with the original depth, and if it fails high here, we *
   *   then re-search it a third time with the original       *
   *   search bound to see if it is really a new best move.   *
   *                                                          *
   ************************************************************
   */
        if (depth - 1 + extension > 0) {
          value =
              -Search(tree, -t_beta, -alpha, Flip(wtm),
              depth - 1 + extension, ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm),
                depth - 1, ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        } else {
          value = -Quiesce(tree, -t_beta, -alpha, Flip(wtm), ply + 1,
            quiesce_check_depth, DoDelta(ply));
          if (abort_search || tree->stop)
            break;
          if (value > alpha && extension < 0 && depth - 1 > 0)
            value =
                -Search(tree, -t_beta, -alpha, Flip(wtm),
                depth - 1, ply + 1, DO_NULL);
          if (abort_search || tree->stop)
            break;
        }
        if (value > alpha && value < beta && t_beta < beta) { /* PVS re-search */
          node_type = PV_NODE;
          extension = Max(extension, 0);
          if (depth - 1 + extension > 0) {
            value =
                -Search(tree, -beta, -alpha, Flip(wtm), depth - 1 + extension,
                ply + 1, DO_NULL);
          }
          else {
            value = -Quiesce(tree, -beta, -alpha, Flip(wtm), ply + 1,
              quiesce_check_depth, DoDelta(ply));
          }
          if (abort_search || tree->stop)
            break;
        }
  /*
   ************************************************************
   *                                                          *
   *   now the move has been searched to a satisfactory       *
   *   conclusion.  we check to see if the search was aborted *
   *   due to time constraints, and if so, we just return and *
   *   do not modify the best move or anything.  if not, we   *
   *   then test for a beta cutoff and return to the control  *
   *   for the iterated search (Iterate()) to alter the       *
   *   aspiration window if needed.                           *
   *                                                          *
   ************************************************************
   */
        root_moves[tree->root_move].nodes =
            tree->nodes_searched - begin_root_nodes;
  /*
   ************************************************************
   *                                                          *
   *   Search (and/or re-search) has been completed.  Now we  *
   *   check for a fail high which terminates the search      *
   *   immediately as no further searching is required.       *
   *                                                          *
   ************************************************************
   */
        t_beta = alpha + 1;
        moves_searched++;
        if (value > alpha) {
          Lock(lock_root);
          Output(tree, value, beta);
          root_value = alpha;
          Unlock(lock_root);
          if (value >= beta) { /* beta cut-off */
            Killer(tree, ply, tree->curmv[ply]);
            History(tree, ply, depth, wtm, tree->curmv[ply], searched,
              moves_searched);
            HashStore(tree, ply, depth, wtm, LOWER, value, tree->curmv[ply]);
            UnmakeMove(tree, ply, tree->curmv[ply], wtm);
            tree->fail_high++;
            if (moves_searched == 1)
              tree->fail_high_first++;
            return (value);
          }
          alpha = value;
          t_beta = alpha + 1;
        }
        Lock(lock_root);
        root_value = alpha;
        Unlock(lock_root);
      } while (0); /* end do search */
    UnmakeMove(tree, ply, tree->curmv[ply], wtm);
    if (abort_search || tree->stop)
      return (0);
/*
 ************************************************************
 *                                                          *
 *   after searching the first move, we can now begin a     *
 *   parallel search at the root if root splitting is       *
 *   allowed, and the next move is not flagged as a "serial *
 *   search only" type move because we think it might be a  *
 *   new best move when searched further.  otherwise, back  *
 *   to the top of the NextRootMove() loop to search the    *
 *   remaining root moves one at a time.                    *
 *                                                          *
 ************************************************************
 */
#if (CPUS > 1)
    Lock(lock_smp);
    if (split_at_root && smp_idle && moves_searched > 0 &&
      NextRootMoveParallel() && alpha != o_alpha) {
      Unlock(lock_smp);
      do {
        tree->alpha = alpha;
        tree->beta = beta;
        tree->value = alpha;
        tree->wtm = wtm;
        tree->ply = ply;
        tree->depth = depth;
        tree->searched = searched;
        tree->moves_searched = moves_searched;
        tree->best_capture = best_capture;
        tree->node_type = node_type;
        if (Thread(tree)) {
          if (abort_search || tree->stop)
            return (0);
          value = tree->search_value;
          moves_searched = tree->moves_searched;
          best_capture = tree->best_capture;
          node_type = tree->node_type;
          if (value > alpha) {
            if (value >= beta) {
              Killer(tree, ply, tree->cutmove);
              History(tree, ply, depth, wtm, tree->cutmove, searched,
                moves_searched);
              HashStore(tree, ply, depth, wtm, LOWER, value, tree->cutmove);
              return (value);
            }
            alpha = value;
            t_beta = alpha + 1;
            break;
          }
        }
      } while (0);
    } else
      Unlock(lock_smp);
#endif
  }
/*
 ************************************************************
 *                                                          *
 *   all moves have been searched.  if none were legal,     *
 *   return either MATE or DRAW depending on whether the    *
 *   side to move is in check or not.                       *
 *                                                          *
 ************************************************************
 */
  if (abort_search || tree->stop || time_abort)
    return (0);
  if (moves_searched == 0) {
    value = (Check(wtm)) ? -(MATE - 1) : DrawScore(wtm);
    if (value >= alpha && value < beta) {
      tree->pv[0].pathl = 0;
      tree->pv[0].pathh = 0;
      tree->pv[0].pathd = iteration_depth;
      Lock(lock_root);
      Output(tree, value, beta);
      root_value = value;
      Unlock(lock_root);
#if defined(TRACE)
      if (trace_level > 0)
        printf("Search() no moves!  ply=%d\n", ply);
#endif
    }
    return (value);
  } else {
    HashStore(tree, ply, depth, wtm, (alpha == o_alpha) ? UPPER : EXACT,
      alpha, (alpha == o_alpha) ? first_tried : tree->pv[ply].path[ply]);
    return (alpha);
  }
}
