#include "chess.h"
#include "data.h"
#define BIG_SORT_VALUE 999999
/* last modified 12/12/17 */
/*
 *******************************************************************************
 *                                                                             *
 *   Swap() is used to analyze capture moves to see whether or not they appear *
 *   to be profitable.  the basic algorithm is extremely fast since it uses    *
 *   the bitmaps to determine which squares are attacking the [target] square. *
 *                                                                             *
 *   the algorithm is quite simple.  using the attack bitmaps, we enumerate    *
 *   all the pieces that are attacking [target] for either side.  then we      *
 *   simply use the lowest piece (value) for the correct side to capture on    *
 *   [target].  we continually "flip" sides taking the lowest piece each time. *
 *                                                                             *
 *   as a piece is used, if it is a sliding piece (pawn, bishop, rook or       *
 *   queen) we remove the piece, then generate moves of bishop/queen or        *
 *   rook/queen and then add those in to the attackers, removing any attacks   *
 *   that have already been used.                                              *
 *                                                                             *
 *******************************************************************************
 */
int Swap(TREE * RESTRICT tree, int move, int wtm)
{
  register BITBOARD attacks, temp = 0, toccupied;
  register int attacked_piece;
  register int piece;
  register int color, nc = 1;
  register int source = From(move);
  register int target = To(move);
  int swap_list[32];

/*
 ************************************************************
 *                                                          *
 *   determine which squares attack <target> for each side. *
 *   initialize by placing the piece on <target> first in   *
 *   the list as it is being captured to start things off.  *
 *                                                          *
 ************************************************************
 */
  toccupied = OccupiedSquares;
  attacks = AttacksTo(tree, target);
  attacked_piece = pc_values[Captured(move)];
/*
 ************************************************************
 *                                                          *
 *   the first piece to capture on <target> is the piece    *
 *   standing on <source>.                                  *
 *                                                          *
 ************************************************************
 */
  color = Flip(wtm);
  swap_list[0] = attacked_piece;
  piece = Piece(move);
  attacked_piece = pc_values[piece];
  Clear(source, toccupied);
  if (piece != knight && piece != king) {
    if (piece & 1)
      attacks |= AttacksBishop(target, toccupied) & BishopsQueens;
    if (piece == pawn || piece & 4)
      attacks |= AttacksRook(target, toccupied) & RooksQueens;
  }
  attacks &= toccupied;
/*
 ************************************************************
 *                                                          *
 *   now pick out the least valuable piece for the correct  *
 *   side that is bearing on <target>.  as we find one, we  *
 *   update the attacks (if this is a sliding piece) to get *
 *   the attacks for any sliding piece that is lined up     *
 *   behind the attacker we are removing.                   *
 *                                                          *
 *   once we know there is a piece attacking the last       *
 *   capturing piece.  add it to the swap list and repeat   *
 *   until one side has no more captures.                   *
 *                                                          *
 ************************************************************
 */
  if (piece != king || attacks) {
    while (attacks) {
      for (piece = pawn; ; piece++) {
        if ((temp = Pieces(color, piece) & attacks)) {
          if (piece != king || !(attacks & Occupied(Flip(color))))
            break;
        }
        if (piece > king)
          break;
      }
      if (piece > king)
        break;
      toccupied ^= (temp & -temp);
      if (piece != knight && piece != king) {
        if (piece & 1)
          attacks |= AttacksBishop(target, toccupied) & BishopsQueens;
        if (piece & 4)
          attacks |= AttacksRook(target, toccupied) & RooksQueens;
      }
      attacks &= toccupied;
      swap_list[nc] = -swap_list[nc - 1] + attacked_piece;
      attacked_piece = pc_values[piece];
      if (piece == king && attacks)
        break;
      nc++;
      color = Flip(color);
    }
  }
/*
 ************************************************************
 *                                                          *
 *   starting at the end of the sequence of values, use a   *
 *   "minimax" like procedure to decide where the captures  *
 *   will stop.                                             *
 *                                                          *
 ************************************************************
 */
  while (--nc)
    swap_list[nc - 1] = -Max(-swap_list[nc - 1], swap_list[nc]);
  return (swap_list[0]);
}

/* last modified 12/15/17 */
/*
 *******************************************************************************
 *                                                                             *
 *   SwapO() is used to analyze a move already made to see if it appears to be *
 *   safe.  It is similar to Swap() except that the move has already been made *
 *   and we are checking to see whether the opponent can gain material by      *
 *   capturing the piece just moved.                                           *
 *                                                                             *
 *******************************************************************************
 */
int SwapO(TREE * RESTRICT tree, int move, int wtm) {
  BITBOARD attacks, temp = 0, toccupied = OccupiedSquares;
  BITBOARD bsliders =
      Bishops(white) | Bishops(black) | Queens(white) | Queens(black);
  BITBOARD rsliders =
      Rooks(white) | Rooks(black) | Queens(white) | Queens(black);
  int attacked_piece, piece, color, nc = 1, swap_list[32];
  int target = To(move);

/*
 ************************************************************
 *                                                          *
 *   Determine which squares attack <target> for each side. *
 *   initialize by placing the piece on <target> first in   *
 *   the list as it is being captured to start things off.  *
 *                                                          *
 ************************************************************
 */
  attacks = AttacksTo(tree, target);
  attacked_piece = pc_values[Piece(move)];
/*
 ************************************************************
 *                                                          *
 *   The first piece to capture on <target> is the piece    *
 *   standing on that square.  We have to find out the      *
 *   least valuable attacker for that square first.         *
 *                                                          *
 ************************************************************
 */
  color = Flip(wtm);
  swap_list[0] = attacked_piece;
  for (piece = pawn; ; piece++) {
    if ((temp = Pieces(color, piece) & attacks)) {
      if (piece != king || !(attacks & Occupied(Flip(color))))
        break;
    }
    if (piece > king)
      break;
  }
  if (piece > king)
    return (0);
  toccupied ^= (temp & -temp);
  if (piece != knight && piece != king) {
    if (piece & 1)
      attacks |= AttacksBishop(target, toccupied) & bsliders;
    if (piece & 4)
      attacks |= AttacksRook(target, toccupied) & rsliders;
  }
  attacked_piece = pc_values[piece];
  color = Flip(color);
  attacks &= toccupied;
/*
 ************************************************************
 *                                                          *
 *   Now pick out the least valuable piece for the correct  *
 *   side that is bearing on <target>.  As we find one, we  *
 *   update the attacks (if this is a sliding piece) to get *
 *   the attacks for any sliding piece that is lined up     *
 *   behind the attacker we are removing.                   *
 *                                                          *
 *   Once we know there is a piece attacking the last       *
 *   capturing piece, add it to the swap list and repeat    *
 *   until one side has no more captures.                   *
 *                                                          *
 ************************************************************
 */
  if (piece != king || attacks) {
    while (attacks) {
      for (piece = pawn; ; piece++) {
        if ((temp = Pieces(color, piece) & attacks)) {
          if (piece != king || !(attacks & Occupied(Flip(color))))
            break;
        }
        if (piece > king)
          break;
      }
      if (piece > king)
        break;
      toccupied ^= (temp & -temp);
      if (piece != knight && piece != king) {
        if (piece & 1)
          attacks |= AttacksBishop(target, toccupied) & bsliders;
        if (piece & 4)
          attacks |= AttacksRook(target, toccupied) & rsliders;
      }
      attacks &= toccupied;
      swap_list[nc] = -swap_list[nc - 1] + attacked_piece;
      attacked_piece = pc_values[piece];
      if (piece == king && attacks)
        break;
      nc++;
      color = Flip(color);
    }
  }
/*
 ************************************************************
 *                                                          *
 *   Starting at the end of the sequence of values, use a   *
 *   "minimax" like procedure to decide where the captures  *
 *   will stop.                                             *
 *                                                          *
 ************************************************************
 */
  while (--nc)
    swap_list[nc - 1] = -Max(-swap_list[nc - 1], swap_list[nc]);
  return (swap_list[0]);
}

/* last modified 04/24/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   Move() is used to analyze non-capture moves to see whether or not they    *
 *   appear to be good.  the algorithm is fast but inaccurate.  we may also    *
 *   call MakeMove() and UnmakeMove() in order to calculate loss or gain.      *
 *                                                                             *
 *******************************************************************************
 */
int Move(TREE * RESTRICT tree, int alpha, int beta, int ply, int move, int wtm,
  NODE_TYPE node_type, int best_capture)
{
  if (!CaptureOrPromote(move))
  {
    register int score = 0;
    register int loss = 0, gain = 0;
    register int btm = Flip(wtm);
    register int source = From(move), target = To(move);
    register BITBOARD attacks = AttacksTo(tree, target);
/*
 ************************************************************
 *                                                          *
 *  This is a score calculation based on SEE loss.  For     *
 *  example,  normally it is bad to move a knight on a      *
 *  square where it would be en prise.                      *
 *                                                          *
 ************************************************************
 */
    if ((attacks & Occupied(btm))) {
      register int piece;
      for (piece = pawn; ; piece++) {
        if ((Pieces(btm, piece) & attacks)) {
          if (piece != king || !(attacks & Occupied(wtm)))
            break;
        }
        if (piece > king)
          break;
      }
      if (piece <= king) {
        loss = MVV_LVA[Piece(move)][piece];
        if (loss < 0) {
          register int l;
          MakeMove(tree, ply, move, wtm);
          if (!Check(wtm)) { /* if legal */
            if ((l = SwapO(tree, move, wtm)) >= 0)
              loss = l;
          }
          UnmakeMove(tree, ply, move, wtm);
          if (loss < 0)
            loss = 0;
        }
      }
    }
    score -= loss;
/*
 ************************************************************
 *                                                          *
 *  This is a score calculation based on the fact that if   *
 *  there was a good capture earlier, then we do not gain   *
 *  from it if we did not play it.  For example, as is bad  *
 *  to move a knight to a square where it is left en prise, *
 *  so should be not taking a queen with a knight and play  *
 *  some quiet move instead.                                *
 *                                                          *
 ************************************************************
 */
    score -= best_capture;
/*
 ************************************************************
 *                                                          *
 *  This is a score calculation based on SEE gain.  For     *
 *  example, if a pawn, which cannot be lost in the         *
 *  immediate exchanges, attacks a rook and a queen, the    *
 *  move wins the rook for the pawn.  Also, say if a pawn,  *
 *  which cannot be lost in the immediate exchanges, is     *
 *  attacking an opponent's rook only, some bonus is given, *
 *  as these moves may lead to forcing lines, so better     *
 *  that they are tried sooner, rather than later.  Also, a *
 *  bonus is given to checks, as they may lead to forcing   *
 *  lines, so better that they are tried sooner, rather     *
 *  than later.                                             *
 *                                                          *
 ************************************************************
 */
    if (!loss && Piece(move) < king) {
      register int btm = Flip(wtm);
      register int attacker = Piece(move);
      register int attacked;
      register int victim = empty;
      register int can_attack = 0;
      for (attacked = queen; attacked > attacker; attacked--) {
        if (attacker == knight)
          attacker = bishop;
        if (attacked == knight || attacked == bishop) {
          if (TotalPieces(btm, knight) || TotalPieces(btm, bishop)) {
            victim = bishop;
            can_attack = 1;
          }
        } else {
          if (TotalPieces(btm, attacked)) {
            victim = attacked;
            can_attack = 1;
          }
        }
        if (attacker == pawn) {
          register int victima, scopum;
          register int distance = 2;
          BITBOARD temp;
          if (Rank(From(move)) == rankflip[wtm][RANK2])
            distance++;
          can_attack = 0;
          for (victima = king; victima > pawn; victima--) {
            if (victima == attacker)
              continue;
            temp = Pieces(btm, victima);
            while (temp) {
              scopum = LSB(temp);
              if (FileDistance(From(move), scopum) == 1 &&
                RankDistance(From(move), scopum) <= distance &&
                RankDistance(From(move), scopum) > 1) {
                can_attack = 1;
                break;
              }
              temp &= temp - 1;
            }
            if (can_attack)
              break;
          }
        }
        if (can_attack)
          break;
      }
      if (can_attack && (victim > attacker || victim == king)) {
        MakeMove(tree, ply, move, wtm);
        if (!Check(wtm)) { /* if legal */
          if (victim == king) { /* if check */
            if ((Piece(move) == queen && Distance(KingSQ(Flip(wtm)),
              To(move)) < 2) || DoubleCheck(tree, wtm)) /* dangerous check */
              gain = BIG_SORT_VALUE;
            else /* check */
              gain = BIG_SORT_VALUE / 2;
          }
          else
            gain = Gain(tree, ply, move, wtm); /* SEE gain */
        }
        UnmakeMove(tree, ply, move, wtm);
        if (gain > 0)
          score += gain;
      }
    }
/*
 ************************************************************
 *                                                          *
 *  This is a crude score calculation based on source and   *
 *  target square of the moved piece, and other factors     *
 *  which may be characteristic for the piece in question,  *
 *  such as for example calculation of bonuses for knight   *
 *  and bishop outposts, etc., or a less crude score        *
 *  calculation based on eval before and after the move.    *
 *                                                          *
 ************************************************************
 */
    if (!loss) {
      if (node_type != PV_NODE || score + late_moves_margin <= 0)
        score += CrudeScore(tree, ply, move, wtm);
      else {
        if (!gain && !best_capture) {
          if ((Piece(move) == pawn &&
            rankflip[wtm][Rank(To(move))] >= RANK6) ||
            (Piece(move) != pawn && Piece(move) != king &&
            (Distance(KingSQ(Flip(wtm)), To(move)) <
            Distance(KingSQ(Flip(wtm)), From(move)) ||
            Distance(KingSQ(Flip(wtm)), To(move) < 5) ||
            ((Piece(move) == queen || Piece(move) == rook) &&
            (RankDistance(KingSQ(Flip(wtm)), To(move) < 3) ||
            FileDistance(KingSQ(Flip(wtm)), To(move) < 3))) ||
            (Piece(move) == knight && Distance(KingSQ(Flip(wtm)), To(move)) < 4) ||
            ((Piece(move) == bishop || Piece(move) == queen) &&
            RankDistance(KingSQ(Flip(wtm)), To(move) > 4) &&
            FileDistance(KingSQ(Flip(wtm)), To(move) > 4)))))
            score += LessCrudeScore(tree, alpha, beta, ply, move, wtm);
          else
            score += CrudeScore(tree, ply, move, wtm);
        }
        else
          score += CrudeScore(tree, ply, move, wtm);
      }
    }
    return (score);
  }
  return (0);
}

/* last modified 10/10/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   Gain() is used to analyze non-capture moves already made to see whether   *
 *   or not they appear to be threatening.  For example, if a pawn, which      *
 *   cannot be lost in the immediate exchanges, attacks a rook and a queen,    *
 *   the move wins the rook for the pawn.  Also, say if a pawn, which cannot   *
 *   be lost in the immediate exchanges, is attacking an opponent's rook only, *
 *   some bonus is given, as these moves may lead to forcing lines, so better  *
 *   that they are tried sooner, rather than later.                            *
 *                                                                             *
 *******************************************************************************
 */
int Gain(TREE * RESTRICT tree, int ply, int move, int wtm)
{
  if (!CaptureOrPromote(move) && Piece(move) < king)
  {
    register int score = 0, gain, t;
    register int attacked = 0;
    register int btm = Flip(wtm);
    register int target;
    register int attacker = Piece(move);
    register int victim;
    BITBOARD temp, attacks;
/*
 ************************************************************
 *                                                          *
 *  This is a score calculation based on SEE gain.          *
 *                                                          *
 ************************************************************
 */
    for (victim = queen; victim > pawn; victim--) {
      if (victim == attacker || ((attacker == knight && victim == bishop) ||
        attacker == bishop && victim == knight))
        continue;
      temp = Pieces(btm, victim);
      while (temp) {
        gain = 0;
        target = LSB(temp);
        attacks = AttacksTo(tree, target);
        if ((attacks & Pieces(wtm, attacker))) {
          gain = MVV_LVA[victim][attacker];
          if (gain < 0 && (t = Swap(tree, move, wtm)) >= 0)
            gain = t;
          if (gain < 0)
            gain = 0;
        }
        if (gain) {
          attacked++;
          if (attacked > 1) {
            score += gain; /* unavoidable loss of material */
            break;
          }
          else
            score += gain / 4; /* avoidable loss of material */
        }
        temp &= temp - 1;
      }
    }
    return (score);
  }
  return (0);
}

/* last modified 11/25/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   CrudeScore() is used to perform a crude score calculation based on source *
 *   and target square of the moved piece, and other factors which may be      *
 *   characteristic for the piece in question, such as for example calculation *
 *   of bonuses for knight and bishop outposts, etc.                           *
 *                                                                             *
 *******************************************************************************
 */
int CrudeScore(TREE * RESTRICT tree, int ply, int move, int wtm)
{
  register int btm = Flip(wtm);
  register int source = From(move), target = To(move);
  register int crude_score = 0;
  register int score_mg = 0, score_eg = 0;
  register int file, pawnsq;
  register int phase = Max(2 * eg_mat, TotalPieces(white, occupied) +
    TotalPieces(black, occupied));
  switch (Piece(move)) {
    case pawn:
      score_mg += pval[mg][wtm][target] - pval[mg][wtm][source];
      score_eg += pval[eg][wtm][target] - pval[eg][wtm][source];
      break;
    case knight:
      score_mg += nval[mg][wtm][target] - nval[mg][wtm][source];
      score_eg += nval[eg][wtm][target] - nval[eg][wtm][source];
      if (!(mask_no_pattacks[btm][target] & Pawns(btm))) {
        if (knight_outpost[wtm][target]) {
          score_eg += knight_outpost[wtm][target];
          score_mg += knight_outpost[wtm][target];
          if (knight_outpost[wtm][target] &&
            pawn_attacks[btm][target] & Pawns(wtm)) {
            score_eg += knight_outpost[wtm][target] / 2;
            score_mg += knight_outpost[wtm][target] / 2;
            if (!Knights(btm) && !(Color(target) & Bishops(btm))) {
            score_eg += knight_outpost[wtm][target];
            score_mg += knight_outpost[wtm][target];
            }
          }
        }
      }
      if (!(mask_no_pattacks[btm][source] & Pawns(btm))) {
        if (knight_outpost[wtm][source]) {
        score_eg -= knight_outpost[wtm][source];
        score_mg -= knight_outpost[wtm][source];
        if (knight_outpost[wtm][source] &&
          pawn_attacks[btm][source] & Pawns(wtm)) {
            score_eg -= knight_outpost[wtm][source] / 2;
            score_mg -= knight_outpost[wtm][source] / 2;
            if (!Knights(btm) && !(Color(source) & Bishops(btm))) {
            score_eg -= knight_outpost[wtm][source];
            score_mg -= knight_outpost[wtm][source];
            }
          }
        }
      }
    break;
    case bishop:
      score_mg += bval[mg][wtm][target] - bval[mg][wtm][source];
      score_eg += bval[eg][wtm][target] - bval[eg][wtm][source];
      if (!(mask_no_pattacks[btm][target] & Pawns(btm))) {
        if (bishop_outpost[wtm][target]) {
        score_eg += bishop_outpost[wtm][target];
        score_mg += bishop_outpost[wtm][target];
        if (bishop_outpost[wtm][target] &&
          pawn_attacks[btm][target] & Pawns(wtm)) {
            score_eg += bishop_outpost[wtm][target] / 2;
            score_mg += bishop_outpost[wtm][target] / 2;
            if (!Knights(btm) && !(Color(target) & Bishops(btm))) {
            score_eg += bishop_outpost[wtm][target];
            score_mg += bishop_outpost[wtm][target];
            }
          }
        }
      }
      if (!(mask_no_pattacks[btm][source] & Pawns(btm))) {
        if (bishop_outpost[wtm][source]) {
        score_eg -= bishop_outpost[wtm][source];
        score_mg -= bishop_outpost[wtm][source];
        if (bishop_outpost[wtm][source] &&
          pawn_attacks[btm][source] & Pawns(wtm)) {
            score_eg -= bishop_outpost[wtm][source] / 2;
            score_mg -= bishop_outpost[wtm][source] / 2;
            if (!Knights(btm) && !(Color(source) & Bishops(btm))) {
            score_eg -= bishop_outpost[wtm][source];
            score_mg -= bishop_outpost[wtm][source];
            }
          }
        }
      }
      if (target == sqflip[wtm][A7] && SetMask(sqflip[wtm][B6]) & Pawns(btm)) {
        score_eg -= bishop_trapped;
        score_mg -= bishop_trapped;
      } else if (target == sqflip[wtm][B8] &&
        SetMask(sqflip[wtm][C7]) & Pawns(btm)) {
        score_eg -= bishop_trapped;
        score_mg -= bishop_trapped;
      } else if (target == sqflip[wtm][H7] &&
        SetMask(sqflip[wtm][G6]) & Pawns(btm)) {
        score_eg -= bishop_trapped;
        score_mg -= bishop_trapped;
      } else if (target == sqflip[wtm][G8] &&
        SetMask(sqflip[wtm][F7]) & Pawns(btm)) {
        score_eg -= bishop_trapped;
        score_mg -= bishop_trapped;
      }
      if (source == sqflip[wtm][A7] && SetMask(sqflip[wtm][B6]) & Pawns(btm)) {
        score_eg += bishop_trapped;
        score_mg += bishop_trapped;
      } else if (source == sqflip[wtm][B8] &&
        SetMask(sqflip[wtm][C7]) & Pawns(btm)) {
        score_eg += bishop_trapped;
        score_mg += bishop_trapped;
      } else if (source == sqflip[wtm][H7] &&
        SetMask(sqflip[wtm][G6]) & Pawns(btm)) {
        score_eg += bishop_trapped;
        score_mg += bishop_trapped;
      } else if (source == sqflip[wtm][G8] &&
        SetMask(sqflip[wtm][F7]) & Pawns(btm)) {
        score_eg += bishop_trapped;
        score_mg += bishop_trapped;
      }
    break;
    case rook:
      if (File(target) != File(source)) {
        file = File(target);
        if (!(file_mask[file] & Pawns(wtm))) {
        if (!(file_mask[file] & Pawns(btm))) {
          score_mg += rook_open_file[mg];
          score_eg += rook_open_file[eg];
        } else {
          score_mg += rook_half_open_file[mg];
          score_eg += rook_half_open_file[eg];
          }
        }
        file = File(source);
        if (!(file_mask[file] & Pawns(wtm))) {
        if (!(file_mask[file] & Pawns(btm))) {
          score_mg -= rook_open_file[mg];
          score_eg -= rook_open_file[eg];
        } else {
          score_mg -= rook_half_open_file[mg];
          score_eg -= rook_half_open_file[eg];
          }
        }
      }
      if (Rank(target) == rankflip[wtm][RANK7] && Rank(source) != rankflip[wtm][RANK7]) {
        if (Rank(KingSQ(btm)) == rankflip[wtm][RANK8]
          || Pawns(btm) & rank_mask[rankflip[wtm][RANK7]]) {
        score_mg += rook_on_7th[mg];
        score_eg += rook_on_7th[eg];
        }
      }
      if (Rank(target) != rankflip[wtm][RANK7] && Rank(source) == rankflip[wtm][RANK7]) {
        if (Rank(KingSQ(btm)) == rankflip[wtm][RANK8]
          || Pawns(btm) & rank_mask[rankflip[wtm][RANK7]]) {
        score_mg -= rook_on_7th[mg];
        score_eg -= rook_on_7th[eg];
        }
      }
      file = File(target);
      if (1 << file & tree->pawn_score.passed[wtm]) {
        pawnsq = MSB(Pawns(wtm) & file_mask[file]);
         if (MSB(AttacksFile(target) & plus8dir[target]) == pawnsq) {
        score_mg += rook_behind_passed_pawn[mg][Rank(target)];
        score_eg += rook_behind_passed_pawn[eg][Rank(target)];
        }
      }
      file = File(source);
      if (1 << file & tree->pawn_score.passed[wtm]) {
        pawnsq = MSB(Pawns(wtm) & file_mask[file]);
         if (MSB(AttacksFile(source) & plus8dir[source]) == pawnsq) {
        score_mg -= rook_behind_passed_pawn[mg][Rank(source)];
        score_eg -= rook_behind_passed_pawn[eg][Rank(source)];
        }
      }
      break;
    case queen:
      score_mg += qval[mg][wtm][target] - qval[mg][wtm][source];
      score_eg += qval[eg][wtm][target] - qval[eg][wtm][source];
      break;
    case king:
      if (tree->all_pawns) {
        register int temp;
        temp = 0;
        if (tree->all_pawns & mask_efgh && tree->all_pawns & mask_abcd)
          temp += kval_n[wtm][target];
        else if (tree->all_pawns & mask_efgh)
          temp += kval_k[wtm][target];
        else
          temp += kval_q[wtm][target];
        score_eg += temp;
        temp = 0;
        if (tree->all_pawns & mask_efgh && tree->all_pawns & mask_abcd)
          temp += kval_n[wtm][source];
        else if (tree->all_pawns & mask_efgh)
          temp += kval_k[wtm][source];
        else
          temp += kval_q[wtm][source];
        score_eg -= temp;
      }
      score_mg += kval_n[wtm][target] - kval_n[wtm][source];
      score_eg += kval_n[wtm][target] - kval_n[wtm][source];
      break;
  }
  crude_score = score_eg + (score_mg - score_eg) * (phase - 2 * eg_mat) / Kp;
  return (crude_score);
}

/* last modified 12/07/21 */
/*
 *******************************************************************************
 *                                                                             *
 *   LessCrudeScore() is used to perform a less crude score calculation based  *
 *   on Evaluate().  Robert's ad hoc test showed that it is better not to use  *
 *   EvaluateNNUE() if NNUE_setup is set.                                      *
 *                                                                             *
 *******************************************************************************
 */
int LessCrudeScore(TREE * RESTRICT tree, int alpha, int beta, int ply,
  int move, int wtm)
{
  register int less_crude_score = 0;
  register int value = Evaluate(tree, ply, wtm, alpha, beta);
  MakeMove(tree, ply, move, wtm);
  if (!Check(wtm)) { /* if legal */
    less_crude_score = Evaluate(tree, ply, wtm, alpha, beta) - value;
  }
  UnmakeMove(tree, ply, move, wtm);
  return (less_crude_score);
}

